@@ -1,36 +1,9 @@
-;* --------------------------------------------------------------------*/
-;*    Copyright (c) 1992-1998 by Manuel Serrano. All rights reserved.  */
-;*                                                                     */
-;*                                     ,--^,                           */
-;*                               _ ___/ /|/                            */
-;*                           ,;'( )__, ) '                             */
-;*                          ;;  //   L__.                              */
-;*                          '   \   /  '                               */
-;*                               ^   ^                                 */
-;*                                                                     */
-;*                                                                     */
-;*    This program is distributed in the hope that it will be useful.  */
-;*    Use and copying of this software and preparation of derivative   */
-;*    works based upon this software are permitted, so long as the     */
-;*    following conditions are met:                                    */
-;*           o credit to the authors is acknowledged following         */
-;*             current academic behaviour                              */
-;*           o no fees or compensation are charged for use, copies,    */
-;*             or access to this software                              */
-;*           o this copyright notice is included intact.               */
-;*      This software is made available AS IS, and no warranty is made */
-;*      about the software or its performance.                         */
-;*                                                                     */
-;*      Bug descriptions, use reports, comments or suggestions are     */
-;*      welcome. Send them to                                          */
-;*        Manuel Serrano -- Manuel.Serrano@unice.fr                    */
-;*-------------------------------------------------------------------- */
 ;*=====================================================================*/
-;*    serrano/prgm/project/bigloo/cigloo0.4/Translate/type.scm         */
+;*    serrano/prgm/project/bigloo/cigloo/Translate/type.scm            */
 ;*    -------------------------------------------------------------    */
 ;*    Author      :  Manuel Serrano                                    */
 ;*    Creation    :  Tue Nov 28 10:33:46 1995                          */
-;*    Last change :  Fri Sep  6 11:04:23 1996 (serrano)                */
+;*    Last change :  Mon Jul 31 11:25:08 2006 (serrano)                */
 ;*    -------------------------------------------------------------    */
 ;*    The type translation                                             */
 ;*=====================================================================*/
@@ -58,16 +31,10 @@
 	    (replace-$        <string> <string>)
 	    ($-in-name?       <string>)))
 
-;*---------------------------------------------------------------------*/
-;*    get-hash-number ...                                              */
-;*---------------------------------------------------------------------*/
-(define (get-hash-number o::string)
-   (string->0..2^x-1 o 12))
-
 ;*---------------------------------------------------------------------*/
 ;*    *Tenv*                                                           */
 ;*---------------------------------------------------------------------*/
-(define *Tenv* (make-hash-table 4096 get-hash-number type-id string=? 1024))
+(define *Tenv* (make-hashtable))
 
 ;*---------------------------------------------------------------------*/
 ;*    *Tlist* ...                                                      */
@@ -81,14 +48,14 @@
 ;*    find-type ...                                                    */
 ;*---------------------------------------------------------------------*/
 (define (find-type id)
-   (get-hash id *Tenv*))
+   (hashtable-get *Tenv* id))
 
 ;*---------------------------------------------------------------------*/
 ;*    bind-type! ...                                                   */
 ;*---------------------------------------------------------------------*/
 (define (bind-type! type)
    (set! *Tlist* (cons type *Tlist*))
-   (put-hash! type *Tenv*))
+   (hashtable-put! *Tenv* (type-id type) type))
       
 ;*---------------------------------------------------------------------*/
 ;*    reset-type-list! ...                                             */
@@ -222,7 +189,7 @@
 ;*    make-function-type-name ...                                      */
 ;*---------------------------------------------------------------------*/
 (define (make-function-type-name to from-name)
-   (compose-type-name to (string-append "$(" from-name ")")))
+   (compose-type-name to (string-append "($(" from-name "))")))
 
 ;*---------------------------------------------------------------------*/
 ;*    make-... ...                                                     */
@@ -261,6 +228,8 @@
    (cond
       ((and (simple-t? type) (string=? (simple-t-id type) "char"))
        (simple->type 'char* "char *"))
+      ((and (simple-t? type) (string=? (simple-t-id type) "FILE"))
+       (simple->type 'file* "FILE *"))
       ((function-t? type)
        (make-ptr-to-function type))
       (else
@@ -295,8 +264,11 @@
 ;*    make-array ...                                                   */
 ;*---------------------------------------------------------------------*/
 (define (make-array type constant-expr)
-   (let* ((expr-string (expr->scheme-id constant-expr))
-	  (type-id     (string-append (type-id type) "-array-" expr-string))
+   (let* ((expr-string
+	   (if (null? constant-expr) "" (expr->scheme-id constant-expr)))
+	  (type-id     (string-append (type-id type) "-array"
+				      (if (null? constant-expr) ""
+					  (string-append "-" expr-string))))
 	  (array-type  (find-type type-id)))
       (if (type? array-type)
 	  array-type
@@ -469,24 +441,52 @@
 	  (warning "cigloo"
 		   "incomplete enum type -- "
 		   (enum-t-c-name enum)))
-      (if (member (type-c-name enum) *no-type*)
-	  'nothing
-	  (begin
-	     (fprin *oport*
-		    "   (type "
-		    (type-id enum)
-		    " (enum")
-	     (for-each (lambda (field)
-			  (fprin *oport*
-				 " ("
-				 (ident-id field)
-				 " \""
-				 (ident-id field)
-				 "\")"))
-		       (enum-t-fields enum))
-	     (fprint *oport* ") \""
-		     (type-c-name enum)
-		     "\")"))))
+      (cond
+	 ((member (type-c-name enum) *no-type*)
+	  'nothing)
+	 (*int-enum*
+	  (fprint *oport*
+		  "   (type "
+		  (type-id enum)
+		  " int "
+		  "\""
+		  (type-c-name enum)
+		  "\")")
+	  (for-each (lambda (field)
+		       (fprint *oport*
+			       "   (macro "
+			       (ident-id field)
+			       "::int \""
+			       (ident-id field)
+			       "\")"))
+		    (enum-t-fields enum)))
+	 (else
+	  (fprin *oport*
+		 "   (type "
+		 (type-id enum)
+		 " (enum")
+	  (for-each (lambda (field)
+		       (fprin *oport*
+			      " ("
+			      (ident-id field)
+			      " \""
+			      (ident-id field)
+			      "\")"))
+		    (enum-t-fields enum))
+	  (fprint *oport* ") \""
+		  (type-c-name enum)
+		  "\")")
+	  (if *enum-macros*
+	      (for-each (lambda (field)
+			   (fprint *oport*
+				   "   (macro "
+				   (ident-id field)
+				   "::"
+				   (type-id enum)
+				   " \""
+				   (ident-id field)
+				   "\")"))
+			(enum-t-fields enum))))))
    (define (translate-array-t array)
       (if (member (type-c-name array) *no-type*)
 	  'nothing
