@@ -1,36 +1,9 @@
-;* --------------------------------------------------------------------*/
-;*    Copyright (c) 1992-1998 by Manuel Serrano. All rights reserved.  */
-;*                                                                     */
-;*                                     ,--^,                           */
-;*                               _ ___/ /|/                            */
-;*                           ,;'( )__, ) '                             */
-;*                          ;;  //   L__.                              */
-;*                          '   \   /  '                               */
-;*                               ^   ^                                 */
-;*                                                                     */
-;*                                                                     */
-;*    This program is distributed in the hope that it will be useful.  */
-;*    Use and copying of this software and preparation of derivative   */
-;*    works based upon this software are permitted, so long as the     */
-;*    following conditions are met:                                    */
-;*           o credit to the authors is acknowledged following         */
-;*             current academic behaviour                              */
-;*           o no fees or compensation are charged for use, copies,    */
-;*             or access to this software                              */
-;*           o this copyright notice is included intact.               */
-;*      This software is made available AS IS, and no warranty is made */
-;*      about the software or its performance.                         */
-;*                                                                     */
-;*      Bug descriptions, use reports, comments or suggestions are     */
-;*      welcome. Send them to                                          */
-;*        Manuel Serrano -- Manuel.Serrano@unice.fr                    */
-;*-------------------------------------------------------------------- */
 ;*=====================================================================*/
-;*    serrano/prgm/project/bigloo/comptime1.9c/Ast/alphatize.scm       */
+;*    serrano/prgm/project/bigloo/comptime/Ast/alphatize.scm           */
 ;*    -------------------------------------------------------------    */
 ;*    Author      :  Manuel Serrano                                    */
 ;*    Creation    :  Fri Jan  6 11:09:14 1995                          */
-;*    Last change :  Mon Dec 22 07:59:16 1997 (serrano)                */
+;*    Last change :  Fri Feb 17 14:11:59 2006 (serrano)                */
 ;*    -------------------------------------------------------------    */
 ;*    The substitution tools module                                    */
 ;*=====================================================================*/
@@ -81,12 +54,20 @@
 (define *location* #f)
 
 ;*---------------------------------------------------------------------*/
-;*    get-location ...                                                 */
+;*    get-inline-location ...                                          */
+;*    -------------------------------------------------------------    */
+;*    This function controls how location are propagate inside inline  */
+;*    body. The default behaviour is to propagate the location of the  */
+;*    caller. To change this, simply inverse the order of the tests.   */
 ;*---------------------------------------------------------------------*/
-(define (get-location node)
-   (if (location? (node-loc node))
-       (node-loc node)
-       *location*))
+(define (get-inline-location node)
+   (cond
+      ((location? *location*)
+       *location*)
+      ((location? (node-loc node))
+       (node-loc node))
+      (else
+       #f)))
 
 ;*---------------------------------------------------------------------*/
 ;*    do-alphatize ...                                                 */
@@ -98,7 +79,7 @@
 ;*---------------------------------------------------------------------*/
 (define-method (do-alphatize node::atom)
    (duplicate::atom node
-      (loc (get-location node))))
+      (loc (get-inline-location node))))
 
 ;*---------------------------------------------------------------------*/
 ;*    do-alphatize ::var ...                                           */
@@ -109,23 +90,25 @@
       (cond
 	 ((eq? alpha #unspecified)
 	  (use-variable! var (node-loc node) 'value)
-	  (duplicate::var node (loc (get-location node))))
+	  (duplicate::var node (loc (get-inline-location node))))
 	 ((variable? alpha)
 	  (use-variable! alpha (node-loc node) 'value)
 	  (if (fun? (variable-value alpha))
 	      (instantiate::closure
-		 (loc (get-location node))
+		 (loc (get-inline-location node))
 		 (type (variable-type alpha))
 		 (variable alpha))
 	      (duplicate::var node
-		 (loc (get-location node))
+		 (loc (get-inline-location node))
 		 (variable alpha))))
 	 ((atom? alpha)
 	  (duplicate::atom alpha))
+	 ((kwote? alpha)
+	  (duplicate::kwote alpha))
 	 (else
 	  (internal-error "alphatize"
-			  "Illegal alphatization"
-			  (shape node))))))
+			  "Illegal alphatization (var)"
+			  (cons (shape node) (shape alpha)))))))
 
 ;*---------------------------------------------------------------------*/
 ;*    do-alphatize ::closure ...                                       */
@@ -136,29 +119,29 @@
       (cond
 	 ((eq? alpha #unspecified)
 	  (use-variable! var (node-loc node) 'value)
-	  (duplicate::closure node (loc (get-location node))))
+	  (duplicate::closure node (loc (get-inline-location node))))
 	 ((variable? alpha)
 	  (use-variable! alpha (node-loc node) 'value)
 	  (duplicate::closure node
-	     (loc (get-location node))
+	     (loc (get-inline-location node))
 	     (variable alpha)))
 	 (else
 	  (internal-error "alphatize"
-			  "Illegal alphatization"
+			  "Illegal alphatization (closure)"
 			  (shape node))))))
 
 ;*---------------------------------------------------------------------*/
 ;*    do-alphatize ::kwote ...                                         */
 ;*---------------------------------------------------------------------*/
 (define-method (do-alphatize node::kwote)
-   (duplicate::kwote node (loc (get-location node))))
+   (duplicate::kwote node (loc (get-inline-location node))))
        
 ;*---------------------------------------------------------------------*/
 ;*    do-alphatize ::sequence ...                                      */
 ;*---------------------------------------------------------------------*/
 (define-method (do-alphatize node::sequence)
    (duplicate::sequence node
-      (loc (get-location node))
+      (loc (get-inline-location node))
       (nodes (map do-alphatize (sequence-nodes node)))))
 
 ;*---------------------------------------------------------------------*/
@@ -168,7 +151,7 @@
    ;; we have to enforce here a variable and not a closure (that
    ;; why the duplicate::var of the fun field).
    (duplicate::app node
-      (loc (get-location node))
+      (loc (get-inline-location node))
       (fun (let ((var (do-alphatize (app-fun node))))
 	      (if (closure? var)
 		  (duplicate::var var)
@@ -176,19 +159,21 @@
       (args (map do-alphatize (app-args node)))))
 
 ;*---------------------------------------------------------------------*/
-;*    do-alphatize ::app-ly ...                                         */
+;*    do-alphatize ::app-ly ...                                        */
 ;*---------------------------------------------------------------------*/
 (define-method (do-alphatize node::app-ly)
    (let ((fun (do-alphatize (app-ly-fun node)))
 	 (arg (do-alphatize (app-ly-arg node))))
-      (if (closure? fun)
+      (if (and (closure? fun)
+	       (not (global-optional? (var-variable fun)))
+	       (not (global-key? (var-variable fun))))
 	  (known-app-ly->node '()
-			     (get-location node)
-			     (duplicate::var fun)
-			     arg
-			     'value)
+			      (get-inline-location node)
+			      (duplicate::var fun)
+			      arg
+			      'value)
 	  (duplicate::app-ly node
-	     (loc (get-location node))
+	     (loc (get-inline-location node))
 	     (fun fun)
 	     (arg arg)))))
 
@@ -202,18 +187,10 @@
    (let ((fun  (do-alphatize (funcall-fun node)))
 	 (args (map do-alphatize (funcall-args node))))
       (if (closure? fun)
-	  (if (correct-arity-app? (var-variable fun)
-				  (cdr args))
-	      (make-app-node '()
-			     (get-location node)
-			     (duplicate::var fun)
-			     (cdr args))
-	      (user-error/location (get-location node)
-				   "Illegal application"
-				   "wrong number of argument(s)"
-				   (shape node)))
+	  (sexp->node `(,(duplicate::var fun) ,@(cdr args))
+		      '() (node-loc node) 'app)
 	  (duplicate::funcall node
-	     (loc (get-location node))
+	     (loc (get-inline-location node))
 	     (fun fun)
 	     (args args)))))
 
@@ -222,15 +199,79 @@
 ;*---------------------------------------------------------------------*/
 (define-method (do-alphatize node::pragma)
    (duplicate::pragma node
-      (loc (get-location node))
-      (args (map do-alphatize (pragma-args node)))))
+      (loc (get-inline-location node))
+      (expr* (map do-alphatize (pragma-expr* node)))))
+
+;*---------------------------------------------------------------------*/
+;*    do-alphatize ::getfield ...                                      */
+;*---------------------------------------------------------------------*/
+(define-method (do-alphatize node::getfield)
+   (duplicate::getfield node
+      (loc (get-inline-location node))
+      (expr* (map do-alphatize (getfield-expr* node)))))
+
+;*---------------------------------------------------------------------*/
+;*    do-alphatize ::setfield ...                                      */
+;*---------------------------------------------------------------------*/
+(define-method (do-alphatize node::setfield)
+   (duplicate::setfield node
+      (loc (get-inline-location node))
+      (expr* (map do-alphatize (setfield-expr* node)))))
+
+;*---------------------------------------------------------------------*/
+;*    do-alphatize ::new ...                                           */
+;*---------------------------------------------------------------------*/
+(define-method (do-alphatize node::new)
+   (duplicate::new node
+      (loc (get-inline-location node))
+      (expr* (map do-alphatize (new-expr* node))) ))
+
+;*---------------------------------------------------------------------*/
+;*    do-alphatize ::vlength ...                                       */
+;*---------------------------------------------------------------------*/
+(define-method (do-alphatize node::vlength)
+   (duplicate::vlength node
+      (loc (get-inline-location node))
+      (expr* (map do-alphatize (vlength-expr* node)))))
+
+;*---------------------------------------------------------------------*/
+;*    do-alphatize ::vref ...                                          */
+;*---------------------------------------------------------------------*/
+(define-method (do-alphatize node::vref)
+   (duplicate::vref node
+      (loc (get-inline-location node))
+      (expr* (map do-alphatize (vref-expr* node)))))
+
+;*---------------------------------------------------------------------*/
+;*    do-alphatize ::vset! ...                                         */
+;*---------------------------------------------------------------------*/
+(define-method (do-alphatize node::vset!)
+   (duplicate::vset! node
+      (loc (get-inline-location node))
+      (expr* (map do-alphatize (vset!-expr* node)))))
+
+;*---------------------------------------------------------------------*/
+;*    do-alphatize ::valloc ...                                        */
+;*---------------------------------------------------------------------*/
+(define-method (do-alphatize node::valloc)
+   (duplicate::valloc node
+      (loc (get-inline-location node))
+      (expr* (map do-alphatize (valloc-expr* node)))))
+
+;*---------------------------------------------------------------------*/
+;*    do-alphatize ::isa ...                                           */
+;*---------------------------------------------------------------------*/
+(define-method (do-alphatize node::isa)
+   (duplicate::isa node
+      (expr* (list (do-alphatize (car (isa-expr* node)))))
+      (loc (get-inline-location node))))
 
 ;*---------------------------------------------------------------------*/
 ;*    do-alphatize ::cast ...                                          */
 ;*---------------------------------------------------------------------*/
 (define-method (do-alphatize node::cast)
    (duplicate::cast node
-      (loc (get-location node))
+      (loc (get-inline-location node))
       (arg (do-alphatize (cast-arg node)))))
 
 ;*---------------------------------------------------------------------*/
@@ -244,20 +285,20 @@
 	 ((eq? alpha #unspecified)
 	  (use-variable! var (node-loc node) 'set!)
 	  (duplicate::setq node
-	     (loc (get-location node))
-	     (var (duplicate::var v (loc (get-location node))))
+	     (loc (get-inline-location node))
+	     (var (duplicate::var v (loc (get-inline-location node))))
 	     (value (do-alphatize (setq-value node)))))
 	 ((variable? alpha)
 	  (use-variable! alpha (node-loc node) 'set!)
 	  (duplicate::setq node
-	     (loc (get-location node))
+	     (loc (get-inline-location node))
 	     (var (duplicate::var v
-		     (loc (get-location node))
+		     (loc (get-inline-location node))
 		     (variable alpha)))
 	     (value (do-alphatize (setq-value node)))))
 	 (else
 	  (internal-error "alphatize"
-			  "Illegal alphatization"
+			  "Illegal alphatization (setq)"
 			  (shape node))))))
 
 ;*---------------------------------------------------------------------*/
@@ -265,7 +306,7 @@
 ;*---------------------------------------------------------------------*/
 (define-method (do-alphatize node::conditional)
    (duplicate::conditional node
-      (loc (get-location node))
+      (loc (get-inline-location node))
       (test (do-alphatize (conditional-test node)))
       (true (do-alphatize (conditional-true node)))
       (false (do-alphatize (conditional-false node)))))
@@ -275,7 +316,7 @@
 ;*---------------------------------------------------------------------*/
 (define-method (do-alphatize node::fail)
    (duplicate::fail node
-      (loc (get-location node))
+      (loc (get-inline-location node))
       (proc (do-alphatize (fail-proc node)))
       (msg  (do-alphatize (fail-msg node)))
       (obj  (do-alphatize (fail-obj node)))))
@@ -285,7 +326,7 @@
 ;*---------------------------------------------------------------------*/
 (define-method (do-alphatize node::select)
    (duplicate::select node
-      (loc (get-location node))
+      (loc (get-inline-location node))
       (test (do-alphatize (select-test node)))
       (clauses (map (lambda (clause)
 		       (cons (car clause)
@@ -297,7 +338,7 @@
 ;*---------------------------------------------------------------------*/
 (define-method (do-alphatize node::make-box)
    (duplicate::make-box node
-      (loc (get-location node))
+      (loc (get-inline-location node))
       (value (do-alphatize (make-box-value node)))))
 
 ;*---------------------------------------------------------------------*/
@@ -305,7 +346,7 @@
 ;*---------------------------------------------------------------------*/
 (define-method (do-alphatize node::box-ref)
    (duplicate::box-ref node
-      (loc (get-location node))
+      (loc (get-inline-location node))
       (var (do-alphatize (box-ref-var node)))))
 
 ;*---------------------------------------------------------------------*/
@@ -313,7 +354,7 @@
 ;*---------------------------------------------------------------------*/
 (define-method (do-alphatize node::box-set!)
    (duplicate::box-set! node
-      (loc (get-location node))
+      (loc (get-inline-location node))
       (var (do-alphatize (box-set!-var node)))
       (value (do-alphatize (box-set!-value node)))))
 
@@ -337,20 +378,21 @@
 			  (old-body (sfun-body old-sfun))
 			  (new-body (alphatize (append old-locals old-args)
 					       (append new-locals new-args)
-					       (get-location node)
+					       (get-inline-location node)
 					       old-body))
 			  (new-sfun (duplicate::sfun old-sfun
 				       (args new-args)
 				       (body new-body))))
+		      (local-user?-set! new (local-user? old))
 		      (local-value-set! new new-sfun)))
 		old-locals
 		new-locals)
       (duplicate::let-fun node
-	 (loc (get-location node))
+	 (loc (get-inline-location node))
 	 (locals new-locals)
 	 (body (alphatize old-locals
 			  new-locals
-			  (get-location node)
+			  (get-inline-location node)
 			  (let-fun-body node))))))
 
 ;*---------------------------------------------------------------------*/
@@ -363,18 +405,22 @@
 				;; all local variables must be allocated
 				;; using the `make-local-svar' form
 				;; (for the key attribution).
-				(make-local-svar (local-id l) (local-type l)))
+				(let ((var (make-local-svar (local-id l)
+							    (local-type l))))
+				   (local-user?-set! var (local-user? l))
+				   (local-access-set! var (local-access l))
+				   var))
 			     old-locals))
 	  (new-bindings (map (lambda (binding new-local)
 				(cons new-local (do-alphatize (cdr binding))))
 			     (let-var-bindings node)
 			     new-locals)))
       (duplicate::let-var node
-	 (loc (get-location node))
+	 (loc (get-inline-location node))
 	 (bindings new-bindings)
 	 (body (alphatize old-locals
 			  new-locals
-			  (get-location node)
+			  (get-inline-location node)
 			  (let-var-body node))))))
 
 ;*---------------------------------------------------------------------*/
@@ -391,14 +437,15 @@
 					   (handler alpha-hdlg))))
 								 
 	  (old-body   (set-ex-it-body node)))
+      (local-user?-set! new-var (local-user? old-var))
       (duplicate::set-ex-it node
-	 (loc (get-location node))
+	 (loc (get-inline-location node))
 	 (var (duplicate::var (set-ex-it-var node)
-		 (loc (get-location node))
+		 (loc (get-inline-location node))
 		 (variable new-var)))
 	 (body (alphatize (list old-var)
 			  (list new-var)
-			  (get-location node)
+			  (get-inline-location node)
 			  old-body)))))
 
 ;*---------------------------------------------------------------------*/
@@ -406,7 +453,7 @@
 ;*---------------------------------------------------------------------*/
 (define-method (do-alphatize node::jump-ex-it)
    (duplicate::jump-ex-it node
-      (loc (get-location node))
+      (loc (get-inline-location node))
       (exit (do-alphatize (jump-ex-it-exit node)))
       (value (do-alphatize (jump-ex-it-value node)))))
 
