@@ -1,36 +1,9 @@
-;* --------------------------------------------------------------------*/
-;*    Copyright (c) 1992-1998 by Manuel Serrano. All rights reserved.  */
-;*                                                                     */
-;*                                     ,--^,                           */
-;*                               _ ___/ /|/                            */
-;*                           ,;'( )__, ) '                             */
-;*                          ;;  //   L__.                              */
-;*                          '   \   /  '                               */
-;*                               ^   ^                                 */
-;*                                                                     */
-;*                                                                     */
-;*    This program is distributed in the hope that it will be useful.  */
-;*    Use and copying of this software and preparation of derivative   */
-;*    works based upon this software are permitted, so long as the     */
-;*    following conditions are met:                                    */
-;*           o credit to the authors is acknowledged following         */
-;*             current academic behaviour                              */
-;*           o no fees or compensation are charged for use, copies,    */
-;*             or access to this software                              */
-;*           o this copyright notice is included intact.               */
-;*      This software is made available AS IS, and no warranty is made */
-;*      about the software or its performance.                         */
-;*                                                                     */
-;*      Bug descriptions, use reports, comments or suggestions are     */
-;*      welcome. Send them to                                          */
-;*        Manuel Serrano -- Manuel.Serrano@unice.fr                    */
-;*-------------------------------------------------------------------- */
 ;*=====================================================================*/
 ;*    serrano/prgm/project/bigloo/comptime/Ast/sexp.scm                */
 ;*    -------------------------------------------------------------    */
 ;*    Author      :  Manuel Serrano                                    */
 ;*    Creation    :  Fri May 31 15:05:39 1996                          */
-;*    Last change :  Fri Jan 30 10:24:51 1998 (serrano)                */
+;*    Last change :  Wed Nov  9 14:54:06 2005 (serrano)                */
 ;*    -------------------------------------------------------------    */
 ;*    We build an `ast node' from a `sexp'                             */
 ;*---------------------------------------------------------------------*/
@@ -40,7 +13,8 @@
 ;*---------------------------------------------------------------------*/
 (module ast_sexp
 
-   (include "Tools/trace.sch")
+   (include "Tools/trace.sch"
+	    "Tools/location.sch")
    
    (import  tools_error
 	    tools_shape
@@ -52,6 +26,8 @@
 	    type_env
 	    type_cache
 	    type_typeof
+	    engine_param
+	    backend_backend
 	    ast_ident
 	    ast_env
 	    ast_var
@@ -62,21 +38,31 @@
 	    ast_let
 	    ast_exit
 	    ast_app
-	    ast_apply)
+	    ast_apply
+	    ast_private
+	    effect_feffect)
    
-   (export  (top-level-sexp->node::node    <sexp> ::obj)
-	    (sexp->node::node              ::obj <obj> ::obj ::symbol)
+   (export  (top-level-sexp->node::node <sexp> ::obj)
+	    (sexp->node::node ::obj <obj> ::obj ::symbol)
+	    (sexp*->node::pair-nil ::pair-nil <obj> ::obj ::symbol)
 	    (define-primop-ref->node::node ::global ::node)
-	    (define-primop->node::node     ::global)
-	    (location->node::node          ::global)
-	    (error-sexp->node::node        ::bstring ::obj ::obj)
-	    (use-variable!                 ::variable ::obj ::symbol)))
+	    (define-primop->node::node ::global)
+	    (location->node::node ::global)
+	    (error-sexp->node::node ::bstring ::obj ::obj)
+	    (use-variable! ::variable ::obj ::symbol)
+	    (make-anonymous-name::symbol loc . pref)))
 
 ;*---------------------------------------------------------------------*/
 ;*    top-level-sexp->node ...                                         */
 ;*---------------------------------------------------------------------*/
 (define (top-level-sexp->node exp loc)
-   (sexp->node exp '() loc 'value))
+   (bind-exit (skip)
+      (with-exception-handler
+	 (lambda (e)
+	    (exception-notify e)
+	    (skip #unspecified))
+	 (lambda ()
+	    (sexp->node exp '() loc 'value)))))
 
 ;*---------------------------------------------------------------------*/
 ;*    *sites* ...                                                      */
@@ -93,7 +79,9 @@
 ;*---------------------------------------------------------------------*/
 (define (sexp->node exp stack loc site)
    [assert (site) (memq site *sites*)]
-   (trace (ast 2) "sexp->node: " exp #\Newline)
+   (trace (ast 2) "sexp->node(" loc "): " (shape exp))
+   (trace (ast 3) " site: " site)
+   (trace (ast 2) #\Newline)
    (match-case exp
 ;*--- () --------------------------------------------------------------*/
       (()
@@ -122,7 +110,7 @@
 		   (object? atom))
 	       (error-sexp->node "Illegal atom" exp loc))  
 	      (else
-	       (instantiate::atom
+ 	       (instantiate::atom
 		  (loc loc)
 		  (type (typeof-atom atom))
 		  (value atom)))))
@@ -150,10 +138,16 @@
 		 (else
 		  (use-variable! global loc site)
 		  (if (and (not (eq? site 'app)) (fun? (global-value global)))
-		      (instantiate::closure
-			 (loc loc)
-			 (type (global-type global))
-			 (variable global))
+		      (if (and (cfun? (global-value global))
+			       (not (backend-foreign-closure (the-backend)))
+			       (not (eq? site 'apply)))
+			  (error-sexp->node
+			   (format "This backend (~a) does not support external first class value functions" (backend-name (the-backend)))
+			   exp loc)
+			  (instantiate::closure
+			     (loc loc)
+			     (type (global-type global))
+			     (variable global)))
 		      (instantiate::var 
 			 (loc loc)
 			 (type (global-type global))
@@ -163,7 +157,7 @@
        (let ((loc (find-location/loc exp loc)))
 	  (match-case exp
 	     ((@ (and (? symbol?) ?name) (and (? symbol?) ?module))
-	      (let ((global (find-global name module)))
+	      (let ((global (find-global/module name module)))
 		 (cond
 		    ((not (global? global))
 		     (error-sexp->node "Unbound variable" exp loc))
@@ -199,8 +193,7 @@
 		  (instantiate::kwote (loc loc)
 				      (type *_*)
 				      (value value)))
-		 ((or (integer? value)
-		      (real? value)
+		 ((or (number? value)
 		      (string? value)
 		      (cnst? value)
 		      ;; I won't put in the compiler that characters
@@ -235,22 +228,32 @@
 	       (sexp->node exp stack loc site))
               (else
 	       (if (or (symbol? si) (atom? si))
-		   (let* ((loc     (find-location/loc exp loc))
-			  (l-si    (find-location/loc si loc))
-			  (l-alors (find-location/loc alors loc))
-			  (l-sinon (find-location/loc sinon loc))
+		   (let* ((oloc    (find-location/loc exp loc))
+			  (cdloc   (find-location/loc (cdr exp) loc))
+			  (cddloc  (find-location/loc (cddr exp) loc))
+			  (cdddloc (find-location/loc (cdddr exp) loc))
+			  (l-si    (find-location/loc si cdloc))
+			  (l-alors (find-location/loc alors cddloc))
+			  (l-sinon (find-location/loc sinon cdddloc))
 			  (si      (sexp->node si stack l-si 'value))
 			  (alors   (sexp->node alors stack l-alors 'value))
-			  (sinon   (sexp->node sinon stack l-sinon 'value)))
+			  (sinon   (sexp->node sinon stack l-sinon 'value))
+			  (loc     (let ((loc (find-location/loc
+					       exp
+					       (node-loc si))))
+				      (if (location? loc)
+					  loc
+					  oloc))))
 		      (instantiate::conditional (loc loc)
 						(type *_*)
 						(test si)
 						(true alors)
 						(false sinon)))
-		   (let ((test (gensym 'test)))
+		   (let ((test (mark-symbol-non-user! (gensym 'test))))
 		      (replace! exp
 				`(let ((,(symbol-append test '::bool) ,si))
-				    (if ,test ,alors ,sinon)))
+				    ,(epairify-rec `(if ,test ,alors ,sinon)
+						   exp)))
 		      (sexp->node exp stack loc site))))))
           ((?- ?si ?alors)
            (set-cdr! (cddr exp) (list #unspecified))
@@ -261,16 +264,24 @@
       ((set! . ?-)
        (match-case exp
           ((?- ?var ?val)
-	   (let* ((loc     (find-location/loc exp loc))
-		  (val-loc (find-location/loc val loc))
-		  (val     (sexp->node val stack val-loc 'value)))
-	      (let ((ast (sexp->node var stack loc 'set!)))
+	   (let* ((loc (find-location/loc exp loc))
+		  (cdloc (find-location/loc (cdr exp) loc))
+		  (cddloc (find-location/loc (cddr exp) loc))
+		  (val-loc (find-location/loc val cddloc))
+		  (val (sexp->node val stack val-loc 'value)))
+	      (let ((ast (sexp->node var stack cdloc 'set!)))
 		 (if (var? ast)
-		     (instantiate::setq
-			(loc   loc)
-			(type  *unspec*)
-			(var   ast)
-			(value val))
+		     (with-access::var ast (variable)
+			(if (and (global? variable)
+				 (global-read-only? variable))
+			    (error-sexp->node "Read-only variable"
+					      exp
+					      cdloc)
+			    (instantiate::setq
+			       (loc   loc)
+			       (type  *unspec*)
+			       (var   ast)
+			       (value val))))
 		     (error-sexp->node "illegal `set!' expression"
 				       exp
 				       loc)))))
@@ -284,10 +295,12 @@
        ;; considered as a mutation of the defined variable.
        (match-case exp
           ((?- ?var ?val)
-	   (let* ((loc     (find-location/loc exp loc))
-		  (val-loc (find-location/loc val loc))
-		  (val     (sexp->node val stack val-loc 'value)))
-	      (let ((ast (sexp->node var stack loc 'value)))
+	   (let* ((loc (find-location/loc exp loc))
+		  (cdloc (find-location/loc (cdr exp) loc))
+		  (cddloc (find-location/loc (cddr exp) loc))
+		  (val-loc (find-location/loc val cddloc))
+		  (val (sexp->node val stack val-loc 'value)))
+	      (let ((ast (sexp->node var stack cdloc 'value)))
 		 (if (and (var? ast) (global? (var-variable ast)))
 		     (begin
 			(global-src-set! (var-variable ast) val)
@@ -305,9 +318,9 @@
 			     (find-location/loc exp loc)))))
 ;*--- a pattern to improve pattern-matching compilation ---------------*/
       ((((or let letrec labels) ?- ?body) . ?args)
-       (let ((let-part (car exp)))
-          (set-car! (cddr let-part) `(,body ,@args))
-          (sexp->node let-part stack loc site)))
+       (let* ((let-part (car exp))
+	      (nexp `(,(car let-part) ,(cadr let-part) (,body ,@args))))
+	  (sexp->node nexp stack loc site)))
 ;*--- let & letrec ----------------------------------------------------*/
       (((or let letrec) . ?-)
        (let->node exp stack loc 'value))
@@ -370,9 +383,9 @@
 ;*--- lambda ----------------------------------------------------------*/
       ((lambda . ?-)
        (match-case exp
-          ((?- ?args . ?body)
-           (let ((loc (find-location/loc exp loc))
-                 (fun (gensym 'lambda)))
+          ((?- ?args . ?body) 
+           (let ((loc  (find-location/loc exp loc))
+		 (fun  (make-anonymous-name loc)))
               (sexp->node `(labels ((,fun ,args ,(normalize-progn body))) ,fun)
 			  stack
 			  loc
@@ -383,18 +396,31 @@
 			     (find-location/loc exp loc)))))
 ;*--- pragma ----------------------------------------------------------*/
       ((pragma . ?-)
-       (pragma/type->node #f *unspec* exp stack loc site))
-;*--- pragma ----------------------------------------------------------*/
+       (pragma/type->node #f #f *unspec* exp stack loc site))
+;*--- pragma/effect ---------------------------------------------------*/
+      ((pragma/effect ?effect . ?rest)
+       (pragma/type->node #f
+			  (parse-effect effect)
+			  *unspec* `(pragma ,@rest) stack loc site))
+;*--- free-pragma -----------------------------------------------------*/
       ((free-pragma . ?-)
-       (pragma/type->node #t *unspec* exp stack loc site))
+       (pragma/type->node #t #f *unspec* exp stack loc site))
+;*--- pragma/effect ---------------------------------------------------*/
+      ((free-pragma/effect ?effect . ?rest)
+       (pragma/type->node #t
+			  (parse-effect effect)
+			  *unspec* `(pragma ,@rest) stack loc site))
 ;*--- failure ---------------------------------------------------------*/
       ((failure . ?-)
        (match-case exp
           ((?- ?proc ?msg ?obj)
 	   (let* ((loc      (find-location/loc exp loc))
-		  (loc-proc (find-location/loc proc loc))
-		  (loc-msg  (find-location/loc msg loc))
-		  (loc-obj  (find-location/loc obj loc))
+		  (cdloc    (find-location/loc (cdr exp) loc))
+		  (cddloc   (find-location/loc (cddr exp) loc))
+		  (cdddloc  (find-location/loc (cdddr exp) loc))
+		  (loc-proc (find-location/loc proc cdloc))
+		  (loc-msg  (find-location/loc msg cddloc))
+		  (loc-obj  (find-location/loc obj cdddloc))
 		  (proc     (sexp->node proc stack loc-proc 'value))
 		  (msg      (sexp->node msg stack loc-msg 'value))
 		  (obj      (sexp->node obj stack loc-obj 'value)))
@@ -409,39 +435,49 @@
 			     (find-location/loc exp loc)))))
 ;*--- case ------------------------------------------------------------*/
       ((case . ?-)
+       ;; former versions of the compiler used to make side effect
+       ;; one the list that hold the clauses. This is a bad idea.
+       ;; Because of macro expansion it could be that the very same
+       ;; case form is seen twice by the compiler. If we compile case
+       ;; by the means of side effects, the second time the compiler
+       ;; will see the case form it will think of it as incorrect.
        (match-case exp
           ((?- ?test . ?clauses)
            (let* ((loc  (find-location/loc exp loc))
+		  (cdloc (find-location/loc (cdr exp) loc))
+		  (cddloc (find-location/loc (cdr exp) loc))
                   (test (sexp->node test
 				    stack
-				    (find-location/loc test loc)
+				    (find-location/loc test cdloc)
 				    'value)))
-	      (let loop ((cls clauses))
+	      (let loop ((cls clauses)
+			 (nclauses '()))
 		 (if (null? cls)
 		     (instantiate::select
 			(loc loc)
 			(type *_*)
 			(test test)
 			(item-type (typeof-atom (car (car (car clauses)))))
-			(clauses clauses))
-		     (let ((clause (car cls)))
-			;; we build the ast for the then part of the clause
-			(set-cdr! clause
-				  (sexp->node (normalize-progn (cdr clause))
-					      stack
-					      (find-location/loc clause loc)
-					      'value))
+			(clauses (reverse! nclauses)))
+		     (let* ((clause (car cls))
+			    (body   (sexp->node (normalize-progn (cdr clause))
+						stack
+						(find-location/loc clause cddloc)
+						'value))
+			    (nclause (cons (car clause) body)))
 			;; we check that it is not an illegal `else' clause
 			(if (and (eq? (car clause) 'else)
 				 (not (null? (cdr cls))))
 			    (error-sexp->node "Illegal `case' form"
-					      (find-location/loc exp loc)
-					      exp)
-			    (loop (cdr cls))))))))
+					      exp
+					      (find-location/loc exp loc))
+			    (loop (cdr cls)
+				  (cons (epairify nclause clause)
+					nclauses))))))))
           (else
 	   (error-sexp->node "Illegal `case' form"
-			     (find-location/loc exp loc)
-			     exp))))
+			     exp
+			     (find-location/loc exp loc)))))
 ;*--- set-exit --------------------------------------------------------*/
       ((set-exit . ?-)
        (set-exit->node exp stack loc site))
@@ -449,8 +485,15 @@
       ((jump-exit . ?-)
        (jump-exit->node exp stack loc site))
 ;*--- apply -----------------------------------------------------------*/
-      ((apply . ?-)
+      ((apply ?- ?-)
        (applycation->node exp stack loc site))
+;*--- private ---------------------------------------------------------*/
+      ((#unspecified)
+       (error-sexp->node "Illegal `application' form"
+			 exp
+			 (find-location/loc exp loc)))
+      ((? private-sexp?)
+       (private->node exp stack loc site))
 ;*--- app -------------------------------------------------------------*/
       (else
        ;; this expression can be a function call or a typed pragma
@@ -458,25 +501,40 @@
        ;; we compile a function call. This check is required by the
        ;; form (pragma::??? ...) (because we can't add a branch in the
        ;; match-case to check the node `pragma::???').
-       (let ((caller (car exp)))
+       (let ((caller (car exp))
+	     (loc    (find-location/loc exp loc)))
 	  (if (symbol? caller)
 	      ;; it might be a typed special forms (such as pragma or lambda)
-	      (let* ((pid  (parse-id caller))
+	      (let* ((pid  (parse-id caller loc))
 		     (id   (car pid))
 		     (type (cdr pid)))
 		 (case id
 		    ((pragma)
-		     (pragma/type->node #f type	exp stack loc site))
+		     (pragma/type->node #f #f type exp stack loc site))
+		    ((pragma/effect)
+		     (if (not (pair? (cdr exp)))
+			 (error-sexp->node "Illegal `pragma/effect' form"
+					   exp
+					   (find-location/loc exp loc))
+			 (pragma/type->node #f (parse-effect (cadr exp))
+					    type `((car exp) ,@(cddr exp))
+					    stack loc site)))
 		    ((free-pragma)
-		     (pragma/type->node #t type exp stack loc site))
+		     (pragma/type->node #t #f type exp stack loc site))
+		    ((free-pragma/effect)
+		     (if (not (pair? (cdr exp)))
+			 (error-sexp->node "Illegal `free-pragma/effect' form"
+					   exp
+					   (find-location/loc exp loc))
+			 (pragma/type->node #t (parse-effect (cadr exp))
+					    type `((car exp) ,@(cddr exp))
+					    stack loc site)))
 		    ((lambda)
 		     (match-case exp
 			((?- ?args . ?body)
 			 (let* ((loc  (find-location/loc exp loc))
-				(fun  (gensym 'lambda))
-				(tfun (symbol-append fun
-						     4dots
-						     (type-id type))))
+				(fun  (mark-symbol-non-user! (gensym 'lambda)))
+				(tfun (make-typed-ident fun (type-id type))))
 			    (sexp->node
 			     `(labels ((,tfun ,args ,(normalize-progn body)))
 				 ,fun)
@@ -486,7 +544,7 @@
 			(else
 			 (error-sexp->node "Illegal `lambda' form"
 					   exp
-					   (find-location/loc exp loc)))))
+					   loc))))
 		    (else
 		     (application->node exp stack loc site))))
 	      (application->node exp stack loc site))))))
@@ -494,19 +552,24 @@
 ;*---------------------------------------------------------------------*/
 ;*    sexp*->node ...                                                  */
 ;*---------------------------------------------------------------------*/
-(define (sexp*->node exp* stack loc site)
-   (let loop ((exps     exp*)
-	      (res      '()))
+(define (sexp*->node::pair-nil exp*::pair-nil stack loc site)
+   (let loop ((exps exp*)
+	      (res  '())
+	      (loc  loc))
       (cond
 	 ((null? exps)
 	  (error-sexp->node "Illegal empty sequence" exps loc))
 	 ((null? (cdr exps))
-	  (let ((loc (find-location/loc (car exps) loc)))
-	     (reverse! (cons (sexp->node (car exps) stack loc site) res))))
+	  (let ((loc (find-location/loc (car exps)
+					(find-location/loc exps loc)))
+		(nsite (if (eq? site 'app) 'value site)))
+	     (reverse! (cons (sexp->node (car exps) stack loc nsite) res))))
 	 (else
-	  (let ((loc (find-location/loc (car exps) loc)))
+	  (let ((loc (find-location/loc (car exps)
+					(find-location/loc exps loc))))
 	     (loop (cdr exps)
-		   (cons (sexp->node (car exps) stack loc 'value) res)))))))
+		   (cons (sexp->node (car exps) stack loc 'value) res)
+		   loc))))))
 
 ;*---------------------------------------------------------------------*/
 ;*    A 1-line cache lookup machinery                                  */
@@ -561,7 +624,7 @@
 ;*    define-primop-ref->node ...                                      */
 ;*---------------------------------------------------------------------*/
 (define (define-primop-ref->node global ref)
-   (let ((fun (sexp->node 'define-primop-ref! '() #f 'app)))
+   (let ((fun (sexp->node '(@ define-primop-ref! __evenv) '() #f 'app)))
       (if (var? fun)
 	  (instantiate::app (type *_*)
 			    (fun fun)
@@ -577,7 +640,7 @@
 ;*---------------------------------------------------------------------*/
 (define (define-primop->node global)
    (use-variable! global #f 'value)
-   (let ((fun (sexp->node 'define-primop! '() #f 'app)))
+   (let ((fun (sexp->node '(@ define-primop! __evenv) '() #f 'app)))
       (if (var? fun)
 	  (instantiate::app (type *_*)
 			    (fun fun)
@@ -606,3 +669,28 @@
 			    (args (list (sexp->node global '() #f 'value))))
 	  fun)))
    
+;*---------------------------------------------------------------------*/
+;*    make-anonymous-name ...                                          */
+;*---------------------------------------------------------------------*/
+(define (make-anonymous-name::symbol loc . pref)
+   (let ((pref (if (and (pair? pref) (string? (car pref)))
+		   (car pref)
+		   "anonymous")))
+      (if (location? loc)
+	  (let ((file (location-fname loc))
+		(line (location-lnum loc))
+		(base (symbol->string (gensym (string-append "<" pref ":")))))
+	     (cond
+		((or (and (>=fx *bdb-debug* 1)
+			  (memq 'bdb (backend-debug-support (the-backend))))
+		     (>=fx *compiler-debug* 1))
+		 (let ((file (if (substring=? file "./" 2)
+				 (substring file 2 (string-length file))
+				 file)))
+		    (string->symbol
+		     (string-append base ":"
+				    file ":" (number->string line)
+				    ">"))))
+		(else
+		 (string->symbol (string-append base ">")))))
+	  (symbol-append (gensym (string-append "<" pref "-")) '>))))
