@@ -1,36 +1,10 @@
-;* --------------------------------------------------------------------*/
-;*    Copyright (c) 1992-1998 by Manuel Serrano. All rights reserved.  */
-;*                                                                     */
-;*                                     ,--^,                           */
-;*                               _ ___/ /|/                            */
-;*                           ,;'( )__, ) '                             */
-;*                          ;;  //   L__.                              */
-;*                          '   \   /  '                               */
-;*                               ^   ^                                 */
-;*                                                                     */
-;*                                                                     */
-;*    This program is distributed in the hope that it will be useful.  */
-;*    Use and copying of this software and preparation of derivative   */
-;*    works based upon this software are permitted, so long as the     */
-;*    following conditions are met:                                    */
-;*           o credit to the authors is acknowledged following         */
-;*             current academic behaviour                              */
-;*           o no fees or compensation are charged for use, copies,    */
-;*             or access to this software                              */
-;*           o this copyright notice is included intact.               */
-;*      This software is made available AS IS, and no warranty is made */
-;*      about the software or its performance.                         */
-;*                                                                     */
-;*      Bug descriptions, use reports, comments or suggestions are     */
-;*      welcome. Send them to                                          */
-;*        Manuel Serrano -- Manuel.Serrano@unice.fr                    */
-;*-------------------------------------------------------------------- */
 ;*=====================================================================*/
-;*    serrano/prgm/project/bigloo/comptime1.9b/Cfa/closure.scm         */
+;*    serrano/prgm/project/bigloo/comptime/Cfa/closure.scm             */
 ;*    -------------------------------------------------------------    */
 ;*    Author      :  Manuel Serrano                                    */
 ;*    Creation    :  Thu Jun 27 11:35:13 1996                          */
-;*    Last change :  Wed Apr 16 12:22:32 1997 (serrano)                */
+;*    Last change :  Thu Mar 23 13:53:24 2006 (serrano)                */
+;*    Copyright   :  1996-2006 Manuel Serrano, see LICENSE file        */
 ;*    -------------------------------------------------------------    */
 ;*    The closure optimization described in:                           */
 ;*                                                                     */
@@ -59,16 +33,18 @@
 	    ast_node
 	    ast_env
 	    cfa_info
+	    cfa_info2
 	    cfa_cfa
 	    cfa_approx
 	    cfa_set
-	    cfa_info
-	    cfa_stack)
-   (export  (closure-optimization!)
+	    cfa_ltype)
+   (export  (closure-optimization! ::pair-nil)
 	    (closure-optimization?)
 	    (add-procedure-ref!  ::node)
 	    (add-make-procedure! ::node)
-	    (add-funcall!        ::node)))
+	    (add-funcall! ::node)
+	    (approx-procedure-el?::bool ::approx)
+	    (approx-procedure-el1?::bool ::approx)))
 
 ;*---------------------------------------------------------------------*/
 ;*    closure-optimization? ...                                        */
@@ -79,7 +55,7 @@
 ;*---------------------------------------------------------------------*/
 ;*    closure-optimization! ...                                        */
 ;*---------------------------------------------------------------------*/
-(define (closure-optimization!)
+(define (closure-optimization! globals)
    (if (closure-optimization?)
        (begin
 	  (verbose 1 "   . Light closures" #\newline)
@@ -101,9 +77,9 @@
 		    (if (not (or lost?
 				 (not X-T?)
 				 (<fx (sfun-arity (global-value fun)) 0)
-				 (not (integer? size))
+				 (not (fixnum? size))
 				 ;; procedure with no free variable
-				 ;; can be leight neither extra-light
+				 ;; cannot be light neither extra-light
 				 ;; because they are already optimized
 				 ;; by the constant compilation.
 				 (=fx size 0)))
@@ -130,7 +106,7 @@
 	  ;; then, we have to scan, all funcall and procedure-ref
 	  ;; and procedure-set!, procedure? in order to change them according
 	  ;; to procedure's classifications and free variables types.
-	  (light-closure!)
+	  (light-closure! globals)
 	  ;; we don't need the cache anymore
 	  (stop-cache))))
 
@@ -247,10 +223,12 @@
 ;*---------------------------------------------------------------------*/
 ;*    light-closure! ...                                               */
 ;*---------------------------------------------------------------------*/
-(define (light-closure!)
+(define (light-closure! globals)
    (light-funcall!)
    (light-access!)
-   (light-make-procedure!))
+   (light-make-procedure!)
+   ;; CARE: *saw* (fomerly invoked only with *saw*)
+   (light-type! globals))
 
 ;*---------------------------------------------------------------------*/
 ;*    light-make-procedure! ...                                        */
@@ -277,11 +255,9 @@
 	       (else
 		(var-variable-set! fun *make-el-procedure*))))
 	 (set! args (cddr args))
-	 (if (stack-optimization?)
-	     (stack! app)
-	     app)))
+	 app))
    (define (make-light-procedure-app app)
-      (with-access::make-procedure-app app (fun args stackable?)
+      (with-access::make-procedure-app app (fun args)
 	 (let* ((size (get-node-atom-value (caddr args)))
 		(ffun (var-variable (car args)))
 		(sfun (variable-value ffun)))
@@ -292,9 +268,7 @@
 				  'sefun))
 	    (var-variable-set! fun *make-l-procedure*)
 	    (set-cdr! args (cddr args))
-	    (if (stack-optimization?)
-		(stack! app)
-		app))))
+	    app)))
    ;; we change the procedure allocation sites
    (for-each (lambda (app)
 		(with-access::make-procedure-app app (X T)
@@ -421,6 +395,26 @@
    (if (closure-optimization?)
        (set! *procedure-ref-list* (cons ast *procedure-ref-list*))))
 
+;*---------------------------------------------------------------------*/
+;*    approx-procedure-el? ...                                         */
+;*---------------------------------------------------------------------*/
+(define (approx-procedure-el? approx)
+   (let ((type (approx-type approx))
+	 (alloc-list (set->list (approx-allocs approx))))
+      (and (eq? type *procedure*)
+	   (pair? alloc-list)
+	   (null? (cdr alloc-list))
+	   (make-procedure-app? (car alloc-list))
+	   (make-procedure-app-X (car alloc-list)))))
+
+;*---------------------------------------------------------------------*/
+;*    approx-procedure-el1? ...                                        */
+;*---------------------------------------------------------------------*/
+(define (approx-procedure-el1? approx)
+   (let ((alloc-list (set->list (approx-allocs approx))))
+      (with-access::make-procedure-app (car alloc-list) (fun)
+	 (eq? (var-variable fun) *make-el-procedure-1*))))
+   
 ;*---------------------------------------------------------------------*/
 ;*    A small cache                                                    */
 ;*---------------------------------------------------------------------*/
@@ -440,17 +434,17 @@
 ;*    start-cache ...                                                  */
 ;*---------------------------------------------------------------------*/
 (define (start-cache)
-   (set! *procedure-ref* (find-global 'procedure-ref 'foreign))
-   (set! *procedure-set!* (find-global 'procedure-set! 'foreign))
-   (set! *procedure-l-ref* (find-global 'procedure-l-ref 'foreign))
-   (set! *procedure-l-set!* (find-global 'procedure-l-set! 'foreign))
-   (set! *procedure-el-ref* (find-global 'procedure-el-ref 'foreign))
-   (set! *procedure-el-set!* (find-global 'procedure-el-set! 'foreign))
-   (set! *procedure-1-el-ref* (find-global 'procedure-1-el-ref 'foreign))
-   (set! *procedure-1-el-set!* (find-global 'procedure-1-el-set! 'foreign))
-   (set! *make-el-procedure* (find-global 'make-el-procedure 'foreign))
-   (set! *make-el-procedure-1* (find-global 'make-el-procedure-1 'foreign))
-   (set! *make-l-procedure* (find-global 'make-l-procedure 'foreign)))
+   (set! *procedure-ref* (find-global/module 'procedure-ref 'foreign))
+   (set! *procedure-set!* (find-global/module 'procedure-set! 'foreign))
+   (set! *procedure-l-ref* (find-global/module 'procedure-l-ref 'foreign))
+   (set! *procedure-l-set!* (find-global/module 'procedure-l-set! 'foreign))
+   (set! *procedure-el-ref* (find-global/module 'procedure-el-ref 'foreign))
+   (set! *procedure-el-set!* (find-global/module 'procedure-el-set! 'foreign))
+   (set! *procedure-1-el-ref* (find-global/module 'procedure-1-el-ref 'foreign))
+   (set! *procedure-1-el-set!* (find-global/module 'procedure-1-el-set! 'foreign))
+   (set! *make-el-procedure* (find-global/module 'make-el-procedure 'foreign))
+   (set! *make-el-procedure-1* (find-global/module 'make-el-procedure-1 'foreign))
+   (set! *make-l-procedure* (find-global/module 'make-l-procedure 'foreign)))
 
 ;*---------------------------------------------------------------------*/
 ;*    stop-cache ...                                                   */
