@@ -1,36 +1,10 @@
-;* --------------------------------------------------------------------*/
-;*    Copyright (c) 1992-1998 by Manuel Serrano. All rights reserved.  */
-;*                                                                     */
-;*                                     ,--^,                           */
-;*                               _ ___/ /|/                            */
-;*                           ,;'( )__, ) '                             */
-;*                          ;;  //   L__.                              */
-;*                          '   \   /  '                               */
-;*                               ^   ^                                 */
-;*                                                                     */
-;*                                                                     */
-;*    This program is distributed in the hope that it will be useful.  */
-;*    Use and copying of this software and preparation of derivative   */
-;*    works based upon this software are permitted, so long as the     */
-;*    following conditions are met:                                    */
-;*           o credit to the authors is acknowledged following         */
-;*             current academic behaviour                              */
-;*           o no fees or compensation are charged for use, copies,    */
-;*             or access to this software                              */
-;*           o this copyright notice is included intact.               */
-;*      This software is made available AS IS, and no warranty is made */
-;*      about the software or its performance.                         */
-;*                                                                     */
-;*      Bug descriptions, use reports, comments or suggestions are     */
-;*      welcome. Send them to                                          */
-;*        Manuel Serrano -- Manuel.Serrano@unice.fr                    */
-;*-------------------------------------------------------------------- */
 ;*=====================================================================*/
-;*    serrano/prgm/project/bigloo/comptime1.9/Cfa/setup.scm            */
+;*    serrano/prgm/project/bigloo/comptime/Cfa/setup.scm               */
 ;*    -------------------------------------------------------------    */
 ;*    Author      :  Manuel Serrano                                    */
 ;*    Creation    :  Tue Jun 25 14:08:53 1996                          */
-;*    Last change :  Mon Mar 10 15:49:26 1997 (serrano)                */
+;*    Last change :  Sat Dec 17 09:29:59 2005 (serrano)                */
+;*    Copyright   :  1996-2005 Manuel Serrano, see LICENSE file        */
 ;*    -------------------------------------------------------------    */
 ;*    We setup the ast for the Cfa.                                    */
 ;*=====================================================================*/
@@ -49,6 +23,8 @@
 	    ast_var
 	    ast_node
 	    cfa_info
+	    cfa_info2
+	    cfa_info3
 	    cfa_approx)
    (export  (set-initial-approx! globals)
 	    (generic node-setup! ::node)
@@ -59,6 +35,8 @@
 ;*---------------------------------------------------------------------*/
 (define (set-initial-approx! globals)
    (for-each (lambda (global)
+		(trace (cfa 2) "set-initial-approx!: " (shape global)
+		       #\Newline)
 		(let ((fun (global-value global)))
 		   (fun-setup! fun global)
 		   (for-each (lambda (local)
@@ -222,6 +200,7 @@
 ;*    node-setup! ::app ...                                            */
 ;*---------------------------------------------------------------------*/
 (define-method (node-setup! node::app)
+   (trace (cfa 2) "Je setup une app: " (shape node) #\Newline)
    (with-access::app node (fun args)
       (node-setup*! args)
       (let ((variable (var-variable fun)))
@@ -287,12 +266,60 @@
 ;*    node-setup! ::pragma ...                                         */
 ;*---------------------------------------------------------------------*/
 (define-method (node-setup! node::pragma)
-   (with-access::pragma node (args type)
-      (node-setup*! args)
+   (with-access::pragma node (expr* type)
+      (node-setup*! expr*)
       (widen!::pragma/Cinfo node
 	 (approx (make-type-approx type)))
       (approx-set-top! (pragma/Cinfo-approx node))))
 
+;*---------------------------------------------------------------------*/
+;*    node-setup! ::getfield ...                                       */
+;*---------------------------------------------------------------------*/
+(define-method (node-setup! node::getfield)
+   (with-access::getfield node (expr* type)
+      (node-setup*! expr*)
+      (widen!::getfield/Cinfo node
+	 (approx (make-type-approx type)))
+      (approx-set-top! (getfield/Cinfo-approx node))))
+
+;*---------------------------------------------------------------------*/
+;*    node-setup! ::setfield ...                                       */
+;*---------------------------------------------------------------------*/
+(define-method (node-setup! node::setfield)
+   (with-access::setfield node (expr* type)
+      (node-setup*! expr*)
+      (widen!::setfield/Cinfo node
+	 (approx (make-type-approx type)))
+      (approx-set-top! (setfield/Cinfo-approx node))))
+
+;*---------------------------------------------------------------------*/
+;*    node-setup! ::new ...                                            */
+;*---------------------------------------------------------------------*/
+(define-method (node-setup! node::new)
+   (with-access::new node (expr* type)
+      (node-setup*! expr*)
+      (widen!::new/Cinfo node
+	 (approx (make-type-approx type)))
+      (approx-set-top! (new/Cinfo-approx node))))
+
+;*---------------------------------------------------------------------*/
+;*    node-setup! ::isa ...                                            */
+;*---------------------------------------------------------------------*/
+(define-method (node-setup! node::isa)
+   (with-access::isa node (type expr*)
+      (node-setup*! expr*)
+      (widen!::isa/Cinfo node
+	 (approx (make-type-approx type)))
+      (approx-set-top! (isa/Cinfo-approx node))))
+
+;*---------------------------------------------------------------------*/
+;*    node-setup! ::cast-null ...                                      */
+;*---------------------------------------------------------------------*/
+(define-method (node-setup! node::cast-null)
+   (with-access::cast-null node (type)
+      (widen!::cast-null/Cinfo node
+	 (approx (make-type-approx type)))))
+
 ;*---------------------------------------------------------------------*/
 ;*    node-setup! ::cast ...                                           */
 ;*---------------------------------------------------------------------*/
@@ -304,6 +331,7 @@
 ;*    node-setup! ::setq ...                                           */
 ;*---------------------------------------------------------------------*/
 (define-method (node-setup! node::setq)
+   (trace (cfa 2) "Je setup un setq: " (shape node) #\Newline)
    (with-access::setq node (var value)
       (node-setup! value)
       (node-setup! var)
@@ -366,7 +394,7 @@
 ;*---------------------------------------------------------------------*/
 (define-method (node-setup! node::let-var)
    (with-access::let-var node (body bindings)
-      (trace (cfa 3) "let-var settup: " #\Newline)
+      (trace (cfa 3) "let-var settup: " (shape node) #\Newline)
       (for-each (lambda (binding)
 		   (let ((var (car binding))
 			 (val (cdr binding)))
@@ -374,15 +402,18 @@
 		      ;; if the variable is read-only we set it a binding
 		      ;; value to improve the approximations which require
 		      ;; offset (such as make-procedure, procedure-ref)
+		      (node-setup! val)
 		      (widen!::reshaped-local var
 			 (binding-value (if (eq? (local-access var) 'read)
 					    val
 					    #f)))
-		      (node-setup! val)
+		      (trace (cfa 3) "Je vais descendre dans val: " (shape val)
+			     " " (find-runtime-type val)
+			     #\Newline)
 		      (trace (cfa 3) "  " (shape var) " ... " #\Newline)))
 		bindings)
       (node-setup! body)))
-
+ 
 ;*---------------------------------------------------------------------*/
 ;*    node-setup! ::set-ex-it ...                                      */
 ;*---------------------------------------------------------------------*/
