@@ -1,40 +1,14 @@
-;* --------------------------------------------------------------------*/
-;*    Copyright (c) 1992-1998 by Manuel Serrano. All rights reserved.  */
-;*                                                                     */
-;*                                     ,--^,                           */
-;*                               _ ___/ /|/                            */
-;*                           ,;'( )__, ) '                             */
-;*                          ;;  //   L__.                              */
-;*                          '   \   /  '                               */
-;*                               ^   ^                                 */
-;*                                                                     */
-;*                                                                     */
-;*    This program is distributed in the hope that it will be useful.  */
-;*    Use and copying of this software and preparation of derivative   */
-;*    works based upon this software are permitted, so long as the     */
-;*    following conditions are met:                                    */
-;*           o credit to the authors is acknowledged following         */
-;*             current academic behaviour                              */
-;*           o no fees or compensation are charged for use, copies,    */
-;*             or access to this software                              */
-;*           o this copyright notice is included intact.               */
-;*      This software is made available AS IS, and no warranty is made */
-;*      about the software or its performance.                         */
-;*                                                                     */
-;*      Bug descriptions, use reports, comments or suggestions are     */
-;*      welcome. Send them to                                          */
-;*        Manuel Serrano -- Manuel.Serrano@unice.fr                    */
-;*-------------------------------------------------------------------- */
 ;*=====================================================================*/
-;*    serrano/prgm/project/bigloo/comptime1.9b/Cfa/specialize.scm      */
+;*    serrano/prgm/project/bigloo/comptime/Cfa/specialize.scm          */
 ;*    -------------------------------------------------------------    */
 ;*    Author      :  SERRANO Manuel                                    */
 ;*    Creation    :  Fri Apr 11 13:18:21 1997                          */
-;*    Last change :  Tue Apr 15 08:07:05 1997 (serrano)                */
+;*    Last change :  Mon Feb 28 17:05:08 2005 (serrano)                */
+;*    Copyright   :  1997-2005 Manuel Serrano, see LICENSE file        */
 ;*    -------------------------------------------------------------    */
 ;*    This module implements an optimization asked by John Gerard      */
 ;*    Malecki <johnm@vlibs.com>. What is does is, for each generic     */
-;*    operation (e.g. +, max, ...) is a specialized operation exists,  */
+;*    operation (e.g. +, max, ...) if a specialized operation exists,  */
 ;*    regarding Cfa type informations, the generic operation is        */
 ;*    replaced by the specific one.                                    */
 ;*=====================================================================*/
@@ -47,10 +21,11 @@
    (import  engine_param
 	    type_type
 	    type_cache
+	    type_typeof
 	    tools_shape
 	    tools_speek
 	    tools_error
-	    coerce_typeof
+	    backend_backend
 	    ast_var
 	    ast_node
 	    ast_env
@@ -58,7 +33,112 @@
 	    inline_walk)
    (static  (wide-class specialized-global::global
 	       (fix read-only)))
-   (export  (specialize! globals)))
+   (export  (specialize! globals)
+	    (arithmetic-operator? ::global)
+	    (arithmetic-spec-types ::global)))
+
+;*---------------------------------------------------------------------*/
+;*    arithmetic-operator? ...                                         */
+;*---------------------------------------------------------------------*/
+(define (arithmetic-operator? global)
+   (with-access::global global (id module)
+      (let ((cell (assq id *specializations*)))
+	 (and (pair? cell)
+	      (eq? (cadr (cadr cell)) module)))))
+	 
+;*---------------------------------------------------------------------*/
+;*    arithmetic-spec-types ...                                        */
+;*    -------------------------------------------------------------    */
+;*    In order to find on which types an arithmetic operator is        */
+;*    specialized, we inspect which modules defines its specialized    */
+;*    versions.                                                        */
+;*---------------------------------------------------------------------*/
+(define (arithmetic-spec-types global)
+   (with-access::global global (id module)
+      (let ((cell (assq id *specializations*)))
+	 (if (not (pair? cell))
+	     '()
+	     (let loop ((spec (cddr cell))
+			(types '()))
+		(cond
+		   ((null? spec)
+		    (if (null? types)
+			(internal-error "arithmetic-spec-types"
+					"Null specification for"
+					(shape global))
+			types))
+		   ((eq? (cadr (car spec)) '__r4_numbers_6_5_fixnum)
+		    (loop (cdr spec) (cons *long* types)))
+		   ((eq? (cadr (car spec)) '__r4_numbers_6_5_flonum)
+		    (loop (cdr spec) (cons *real* types)))
+		   (else
+		    (loop (cdr spec) types))))))))
+	 
+;*---------------------------------------------------------------------*/
+;*    *specializations* ...                                            */
+;*---------------------------------------------------------------------*/
+(define *specializations*
+   '((c-eq? (c-eq? foreign)
+	    (=fx __r4_numbers_6_5_fixnum)
+	    (=fl __r4_numbers_6_5_flonum))
+     (2= (2= __r4_numbers_6_5)
+	 (=fx __r4_numbers_6_5_fixnum)
+	 (=fl __r4_numbers_6_5_flonum))
+     (2< (2< __r4_numbers_6_5)
+	 (<fx __r4_numbers_6_5_fixnum)
+	 (<fl __r4_numbers_6_5_flonum))
+     (2> (2> __r4_numbers_6_5)
+	 (>fx __r4_numbers_6_5_fixnum)
+	 (>fl __r4_numbers_6_5_flonum))
+     (2<= (2<= __r4_numbers_6_5)
+	  (<=fx __r4_numbers_6_5_fixnum)
+	  (<=fl __r4_numbers_6_5_flonum))
+     (2>= (2>= __r4_numbers_6_5)
+	  (>=fx __r4_numbers_6_5_fixnum)
+	  (>=fl __r4_numbers_6_5_flonum))
+     (zero? (zero? __r4_numbers_6_5)
+	    (zerofx? __r4_numbers_6_5_fixnum)
+	    (zerofl? __r4_numbers_6_5_flonum))
+     (positive? (positive? __r4_numbers_6_5)
+		(positivefx? __r4_numbers_6_5_fixnum)
+		(positivefl? __r4_numbers_6_5_flonum))
+     (negative? (negative? __r4_numbers_6_5)
+		(negativefx? __r4_numbers_6_5_fixnum)
+		(negativefl? __r4_numbers_6_5_flonum))
+     (2+ (2+ __r4_numbers_6_5)
+	 (+fx __r4_numbers_6_5_fixnum)
+	 (+fl __r4_numbers_6_5_flonum))
+     (2- (2- __r4_numbers_6_5)
+	 (-fx __r4_numbers_6_5_fixnum)
+	 (-fl __r4_numbers_6_5_flonum))
+     (2* (2* __r4_numbers_6_5)
+	 (*fx __r4_numbers_6_5_fixnum)
+	 (*fl __r4_numbers_6_5_flonum))
+     (2/ (2/ __r4_numbers_6_5)
+	 (/fl __r4_numbers_6_5_flonum))
+     (abs (abs __r4_numbers_6_5)
+	  (absfx __r4_numbers_6_5_fixnum)
+	  (absfl __r4_numbers_6_5_flonum))
+     (log (log __r4_numbers_6_5)
+	  (logfl __r4_numbers_6_5_flonum))
+     (exp (exp __r4_numbers_6_5)
+	  (expfl __r4_numbers_6_5_flonum))
+     (sin (sin __r4_numbers_6_5)
+	  (sinfl __r4_numbers_6_5_flonum))
+     (cos (cos __r4_numbers_6_5)
+	  (cosfl __r4_numbers_6_5_flonum))
+     (tan (tan __r4_numbers_6_5)
+	  (tanfl __r4_numbers_6_5_flonum))
+     (atan (atan __r4_numbers_6_5)
+	   (atanfl __r4_numbers_6_5_flonum))
+     (asin (asin __r4_numbers_6_5)
+	   (asinfl __r4_numbers_6_5_flonum))
+     (acos (acos __r4_numbers_6_5)
+	   (acosfl __r4_numbers_6_5_flonum))
+     (sqrt (sqrt __r4_numbers_6_5)
+	   (sqrtfl __r4_numbers_6_5_flonum))))
+
+(define *c-eq?* #unspecified)
 
 ;*---------------------------------------------------------------------*/
 ;*    specialize-optimization? ...                                     */
@@ -75,20 +155,30 @@
 		(if (>fx (cdr type.num) 0)
 		    (verbose 2 "        (-> " (shape (car type.num))
 			     ": " (cdr type.num) #")\n")))
-	     *specialized-types*))
+	     *specialized-types*)
+   (for-each (lambda (type.num)
+		(if (>fx (cdr type.num) 0)
+		    (verbose 2 "        (eq? " (shape (car type.num))
+			     ": " (cdr type.num) #")\n")))
+	     *specialized-eq-types*))
 
 ;*---------------------------------------------------------------------*/
 ;*    *specialize* ...                                                 */
 ;*---------------------------------------------------------------------*/
-(define *specialize*        '())
-(define *specialized-types* '())
+(define *specialize*                  '())
+(define *specialized-types*           '())
+(define *specialized-eq-types*        '())
+(define *boxed-eq?*                   #unspecified)
 
 ;*---------------------------------------------------------------------*/
 ;*    install-specialized-type! ...                                    */
 ;*---------------------------------------------------------------------*/
 (define (install-specialized-type! type)
    (if (not (pair? (assq type *specialized-types*)))
-       (set! *specialized-types* (cons (cons type 0) *specialized-types*))))
+       (begin
+	  (set! *specialized-types* (cons (cons type 0) *specialized-types*))
+	  (set! *boxed-eq?* (find-global 'c-boxed-eq? 'foreign))
+	  (set! *c-eq?* (find-global 'c-eq? 'foreign)))))
 
 ;*---------------------------------------------------------------------*/
 ;*    add-specialized-type! ...                                        */
@@ -101,11 +191,23 @@
 			  (shape type))
 	  (set-cdr! cell (+fx 1 (cdr cell))))))
 
+;*---------------------------------------------------------------------*/
+;*    add-specialized-eq-type! ...                                     */
+;*---------------------------------------------------------------------*/
+(define (add-specialized-eq-type! type)
+   (let ((cell (assq type *specialized-eq-types*)))
+      (if (not (pair? cell))
+	  (set! *specialized-eq-types* (cons (cons type 1)
+					     *specialized-eq-types*))
+	  (set-cdr! cell (+fx 1 (cdr cell))))))
+
 ;*---------------------------------------------------------------------*/
 ;*    install-specialize! ...                                          */
 ;*---------------------------------------------------------------------*/
-(define (install-specialize! gen-id gen-mod . fixes)
-   (let ((generic (find-global gen-id gen-mod)))
+(define (install-specialize! gen . fixes)
+   (let* ((gen-id (car gen))
+	  (gen-mod (cadr gen))
+	  (generic (find-global/module gen-id gen-mod)))
       (if (global? generic)
 	  (let loop ((fixes fixes)
 		     (res   '()))
@@ -114,28 +216,43 @@
 		    (widen!::specialized-global generic
 		       (fix res))
 		    (set! *specialize* (cons generic *specialize*)))
-		 (let* ((id     (car fixes))
-			(mod    (cadr fixes))
-			(global (find-global id mod)))
-		    (if (and (global? global)
-			     (sfun? (global-value global))
-			     (pair? (sfun-args (global-value global))))
-			(let ((type (local-type
-				     (car
-				      (sfun-args
-				       (global-value
-					global))))))
-			   (install-specialized-type! type)
-			   (loop (cddr fixes)
-				 (cons (cons type global) res)))
+		 (let* ((fix    (car fixes))
+			(id     (car fix))
+			(mod    (cadr fix))
+			(global (find-global/module id mod)))
+		    (if (global? global)
+			(let ((val (global-value global)))
+			   (cond
+			      ((and (sfun? val) (pair? (sfun-args val)))
+			       (let ((type (let ((v (car
+						     (sfun-args
+						      val))))
+					      (cond
+						 ((type? v)
+						  v)
+						 ((local? v)
+						  (local-type v))
+						 (else
+						  (internal-error
+						   "install-specialize"
+						   "Illegal argument"
+						   (global-id global)))))))
+				  (install-specialized-type! type)
+				  (loop (cdr fixes)
+					(cons (cons type global) res))))
+			      ((and (cfun? val) (pair? (cfun-args-type val)))
+			       (let ((type (car (cfun-args-type val))))
+				  (install-specialized-type! type)
+				  (loop (cdr fixes)
+					(cons (cons type global) res))))))
 			(begin
 			   (warning "install-specialize!"
 				    "Can't find global"
 				    " -- "
 				    (cons id mod))
-			   (loop (cddr fixes) res))))))
+			   (loop (cdr fixes) res))))))
 	  (warning "install-specialize!"
-		   "Can't fund global"
+		   "Can't find global"
 		   " -- "
 		   (cons gen-id gen-mod)))))
 
@@ -144,68 +261,18 @@
 ;*---------------------------------------------------------------------*/
 (define (uninstall-specializes!)
    (set! *specialized-types* '())
-   (for-each shrink! *specialize*))
-   
+   (for-each (lambda (o) (shrink! o)) *specialize*)
+   (set! *specialize* '()))
+
 ;*---------------------------------------------------------------------*/
 ;*    specialize! ...                                                  */
 ;*---------------------------------------------------------------------*/
 (define (specialize! globals)
    (if (specialize-optimization?)
        (begin
-	  (install-specialize! '2= '__r4_numbers_6_5
-			       '=fx '__r4_numbers_6_5_fixnum
-			       '=fl '__r4_numbers_6_5_flonum)
-	  (install-specialize! '2< '__r4_numbers_6_5
-			       '<fx '__r4_numbers_6_5_fixnum
-			       '<fl '__r4_numbers_6_5_flonum)
-	  (install-specialize! '2> '__r4_numbers_6_5
-			       '>fx '__r4_numbers_6_5_fixnum
-			       '>fl '__r4_numbers_6_5_flonum)
-	  (install-specialize! '2<= '__r4_numbers_6_5
-			       '<=fx '__r4_numbers_6_5_fixnum
-			       '<=fl '__r4_numbers_6_5_flonum)
-	  (install-specialize! '2>= '__r4_numbers_6_5
-			       '>=fx '__r4_numbers_6_5_fixnum
-			       '>=fl '__r4_numbers_6_5_flonum)
-	  (install-specialize! 'zero? '__r4_numbers_6_5
-			       'zerofx? '__r4_numbers_6_5_fixnum
-			       'zerofl? '__r4_numbers_6_5_flonum)
-	  (install-specialize! 'positive? '__r4_numbers_6_5
-			       'positivefx? '__r4_numbers_6_5_fixnum
-			       'positivefl? '__r4_numbers_6_5_flonum)
-	  (install-specialize! 'negative? '__r4_numbers_6_5
-			       'negativefx? '__r4_numbers_6_5_fixnum
-			       'negativefl? '__r4_numbers_6_5_flonum)
-	  (install-specialize! '2+ '__r4_numbers_6_5
-			       '+fx '__r4_numbers_6_5_fixnum
-			       '+fl '__r4_numbers_6_5_flonum)
-	  (install-specialize! '2- '__r4_numbers_6_5
-			       '-fx '__r4_numbers_6_5_fixnum
-			       '-fl '__r4_numbers_6_5_flonum)
-	  (install-specialize! '2* '__r4_numbers_6_5
-			       '*fx '__r4_numbers_6_5_fixnum
-			       '*fl '__r4_numbers_6_5_flonum)
-	  (install-specialize! '2/ '__r4_numbers_6_5
-			       '/fl '__r4_numbers_6_5_flonum)
-	  (install-specialize! 'abs '__r4_numbers_6_5
-			       'absfx '__r4_numbers_6_5_fixnum
-			       'absfl '__r4_numbers_6_5_flonum)
-	  (install-specialize! 'log '__r4_numbers_6_5
-			       'logfl '__r4_numbers_6_5_flonum)
-	  (install-specialize! 'exp '__r4_numbers_6_5
-			       'expfl '__r4_numbers_6_5_flonum)
-	  (install-specialize! 'sin '__r4_numbers_6_5
-			       'sinfl '__r4_numbers_6_5_flonum)
-	  (install-specialize! 'cos '__r4_numbers_6_5
-			       'cosfl '__r4_numbers_6_5_flonum)
-	  (install-specialize! 'tan '__r4_numbers_6_5
-			       'tanfl '__r4_numbers_6_5_flonum)
-	  (install-specialize! 'asin '__r4_numbers_6_5
-			       'asinfl '__r4_numbers_6_5_flonum)
-	  (install-specialize! 'acos '__r4_numbers_6_5
-			       'acosfl '__r4_numbers_6_5_flonum)
-	  (install-specialize! 'sqrt '__r4_numbers_6_5
-			       'sqrtfl '__r4_numbers_6_5_flonum)
+	  (for-each (lambda (spec)
+		       (apply install-specialize! (cdr spec)))
+		    *specializations*)
 	  (patch-tree! globals)
 	  (show-specialize)
 	  (uninstall-specializes!)))
@@ -280,11 +347,11 @@
       node))
 
 ;*---------------------------------------------------------------------*/
-;*    patch! ::pragma ...                                              */
+;*    patch! ::extern ...                                              */
 ;*---------------------------------------------------------------------*/
-(define-method (patch! node::pragma)
-   (with-access::pragma node (args type)
-      (patch*! args)
+(define-method (patch! node::extern)
+   (with-access::extern node (expr* type)
+      (patch*! expr*)
       node))
 
 ;*---------------------------------------------------------------------*/
@@ -418,20 +485,25 @@
       (patch*! args)
       (set! fun (patch! fun))
       (let ((v (var-variable fun)))
-	 (if (specialized-global? v)
-	     (specialize-app! node)
-	     node))))
+	 (cond
+	    ((specialized-global? v)
+	     (specialize-app! node))
+	    (else
+	     node)))))
 
 ;*---------------------------------------------------------------------*/
 ;*    specialize-app! ...                                              */
 ;*---------------------------------------------------------------------*/
-(define (specialize-app!::node node::app)
+(define (specialize-app!::app node::app)
    (with-access::app node (fun args)
       (if (null? args)
 	  node
-	  ;; we check if the type of all the argument
+	  ;; we check if the type of all the arguments
 	  ;; is either fixnum or flonum
 	  (let* ((type (typeof (car args)))
+		 (type (if (or (eq? type *char*) (eq? type *int*))
+			   *long*
+			   type))
 		 (glo  (var-variable fun))
 		 (spec (assq type (specialized-global-fix glo))))
 	     (if (pair? spec)
@@ -446,6 +518,55 @@
 			(loop (cdr args)))
 		       (else
 			;; sorry, it fails
-			node)))
-		 node)))))
-			
+			(specialize-eq? node))))
+		 (specialize-eq? node))))))
+
+;*---------------------------------------------------------------------*/
+;*    specialize-eq? ...                                               */
+;*    -------------------------------------------------------------    */
+;*    For the JVM back-end it is important to use a kind of non        */
+;*    fully-polymorph definition of EQ? as soon as one of the          */
+;*    argument is known not to be an integer.                          */
+;*---------------------------------------------------------------------*/
+(define (specialize-eq? node)
+   (with-access::app node (fun args)
+      (define (boxed-eq! node type)
+	 (if (global? *boxed-eq?*)
+	     (begin
+		(var-variable-set! fun *boxed-eq?*)
+		(add-specialized-eq-type! type))))
+      (if (and (eq? (var-variable fun) *c-eq?*)
+	       (backend-typed-eq (the-backend)))
+	  ;; here we are...
+	  (let ((t1 (typeof (car args)))
+		(t2 (typeof (cadr args))))
+	     (cond
+		((and (eq? t1 *obj*) (eq? t2 *obj*))
+		 #unspecified)
+		((eq? t1 *bint*)
+		 (if (eq? t2 *bint*)
+		     ;; we specialize for an boxed integers eq?
+		     (boxed-eq! node *bint*)))
+		((eq? t2 *bint*)
+		 #unspecified)
+		((eq? t1 *breal*)
+		 (if (eq? t2 *breal*)
+		     ;; we specialize for an boxed reals eq?
+		     (boxed-eq! node *breal*)))
+		((eq? t2 *breal*)
+		 #unspecified)
+		((eq? t2 *bint*)
+		 #unspecified)
+		((eq? t1 *bchar*)
+		 (if (eq? t2 *bchar*)
+		     ;; we specialize for an boxed chars eq?
+		     (boxed-eq! node *bchar*)))
+		((eq? t2 *bchar*)
+		 #unspecified)
+		((and (bigloo-type? t1) (bigloo-type? t2))
+		 ;; we specialize with a boxed eq?
+		 (boxed-eq! node (if (eq? t1 *obj*) t2 t1)))
+		(else
+		 #unspecified))
+	     node)
+	  node)))
