@@ -1,36 +1,10 @@
-;* --------------------------------------------------------------------*/
-;*    Copyright (c) 1992-1998 by Manuel Serrano. All rights reserved.  */
-;*                                                                     */
-;*                                     ,--^,                           */
-;*                               _ ___/ /|/                            */
-;*                           ,;'( )__, ) '                             */
-;*                          ;;  //   L__.                              */
-;*                          '   \   /  '                               */
-;*                               ^   ^                                 */
-;*                                                                     */
-;*                                                                     */
-;*    This program is distributed in the hope that it will be useful.  */
-;*    Use and copying of this software and preparation of derivative   */
-;*    works based upon this software are permitted, so long as the     */
-;*    following conditions are met:                                    */
-;*           o credit to the authors is acknowledged following         */
-;*             current academic behaviour                              */
-;*           o no fees or compensation are charged for use, copies,    */
-;*             or access to this software                              */
-;*           o this copyright notice is included intact.               */
-;*      This software is made available AS IS, and no warranty is made */
-;*      about the software or its performance.                         */
-;*                                                                     */
-;*      Bug descriptions, use reports, comments or suggestions are     */
-;*      welcome. Send them to                                          */
-;*        Manuel Serrano -- Manuel.Serrano@unice.fr                    */
-;*-------------------------------------------------------------------- */
 ;*=====================================================================*/
-;*    serrano/prgm/project/bigloo/comptime1.9c/Cfa/tvector.scm         */
+;*    serrano/prgm/project/bigloo/comptime/Cfa/tvector.scm             */
 ;*    -------------------------------------------------------------    */
 ;*    Author      :  Manuel Serrano                                    */
 ;*    Creation    :  Wed Apr  5 18:47:23 1995                          */
-;*    Last change :  Tue Aug  5 10:48:54 1997 (serrano)                */
+;*    Last change :  Wed Feb 22 18:22:39 2006 (serrano)                */
+;*    Copyright   :  1995-2006 Manuel Serrano, see LICENSE file        */
 ;*    -------------------------------------------------------------    */
 ;*    The `vector->tvector' optimization.                              */
 ;*=====================================================================*/
@@ -58,12 +32,12 @@
 	    ast_sexp
 	    ast_env
 	    cfa_info
+	    cfa_info2
+	    cfa_info3
 	    cfa_cfa
 	    cfa_approx
 	    cfa_set
-	    cfa_info
 	    cfa_type
-	    cfa_stack
 	    globalize_walk
 	    inline_inline
 	    inline_walk)
@@ -72,7 +46,7 @@
 	    (vector->tvector! globals)
 	    (add-make-vector! ::node)
 	    (tvector-optimization?)
-	    (generic get-vector-item-type::type ::app)))
+	    (generic get-vector-item-type::type ::node)))
 
 ;*---------------------------------------------------------------------*/
 ;*    tvector-optimization? ...                                        */
@@ -144,7 +118,8 @@
 	  (let ((g (find-global 'vector?)))
 	     (if (global? g)
 		 (let ((f (global-value g)))
-		    (local-type-set! (car (sfun-args f)) *obj*)))))))
+		    (local-type-set! (car (sfun-args f)) *obj*)))))
+       #unspecified))
     
 ;*---------------------------------------------------------------------*/
 ;*    vector->tvector! ...                                             */
@@ -168,7 +143,9 @@
 			   "additional-body: " (shape add-tree) #\Newline)
 		    (patch-tree! globals)
 		    add-tree)
-		 '())))
+		 (begin
+		    (patch-tree! globals)
+		    '()))))
        '()))
 
 ;*---------------------------------------------------------------------*/
@@ -190,7 +167,7 @@
 ;*    optimized.                                                       */
 ;*---------------------------------------------------------------------*/
 (define (get-tvectors)
-   (let loop ((apps      *make-vector-list*)
+   (let loop ((apps *make-vector-list*)
 	      (tvectors '()))
       (if (null? apps)
 	  tvectors
@@ -207,7 +184,7 @@
 ;*---------------------------------------------------------------------*/
 ;*    get-vector-item-type ...                                         */
 ;*---------------------------------------------------------------------*/
-(define-generic (get-vector-item-type app::app))
+(define-generic (get-vector-item-type node::node))
  
 ;*---------------------------------------------------------------------*/
 ;*    get-vector-item-type ::make-vector-app ...                       */
@@ -219,14 +196,14 @@
 	  (approx-type value-approx))))
 
 ;*---------------------------------------------------------------------*/
-;*    get-vector-item-type ::create-vector-app ...                     */
+;*    get-vector-item-type ::valloc/Cinfo+optim ...                    */
 ;*---------------------------------------------------------------------*/
-(define-method (get-vector-item-type app::create-vector-app)
-   (with-access::create-vector-app app (value-approx seen?)
+(define-method (get-vector-item-type node::valloc/Cinfo+optim)
+   (with-access::valloc/Cinfo+optim node (value-approx seen?)
       (if (not seen?)
 	  *vector*
 	  (approx-type value-approx))))
-
+   
 ;*---------------------------------------------------------------------*/
 ;*    show-tvector ...                                                 */
 ;*---------------------------------------------------------------------*/
@@ -311,9 +288,12 @@
       (let* ((approx (cfa! node))
 	     (tv     (get-approx-type approx)))
 	 (if (tvec? tv)
-	     (let ((knode (shrink! knode)))
-		(duplicate::kwote knode (value (a-tvector tv value))))
-	     (shrink! knode)))))
+	     (let* ((knode (shrink! knode))
+		    (n (duplicate::kwote knode (value (a-tvector tv value)))))
+		(widen!::kwote/Cinfo n
+		   (approx approx)))
+	     (widen!::kwote/Cinfo (shrink! knode)
+		(approx approx))))))
 		    
 ;*---------------------------------------------------------------------*/
 ;*    patch! ::var ...                                                 */
@@ -354,11 +334,11 @@
       node))
 
 ;*---------------------------------------------------------------------*/
-;*    patch! ::pragma ...                                              */
+;*    patch! ::extern ...                                              */
 ;*---------------------------------------------------------------------*/
-(define-method (patch! node::pragma)
-   (with-access::pragma node (args type)
-      (patch*! args)
+(define-method (patch! node::extern)
+   (with-access::extern node (expr* type)
+      (patch*! expr*)
       node))
 
 ;*---------------------------------------------------------------------*/
@@ -495,8 +475,6 @@
 	 (if (global? v)
 	     (if (cfun? (variable-value v))
 		 (case (global-id v)
-		    ((c-vector-length)
-		     (patch-vector-length! node))
 		    ((c-vector?)
 		     (patch-vector?! node))
 		    (else
@@ -508,20 +486,20 @@
 	     node))))
 
 ;*---------------------------------------------------------------------*/
-;*    patch-vector-length! ...                                         */
+;*    patch! ::vlength ...                                             */
 ;*---------------------------------------------------------------------*/
-(define (patch-vector-length! node::app)
-   (with-access::vector-set!-app node (args loc)
-      (patch*! args)
-      (let* ((approx (cfa! (car args)))
+(define-method (patch! node::vlength)
+   (with-access::vlength/Cinfo node (expr* loc tvector?)
+      (patch*! expr*)
+      (let* ((approx (cfa! (car expr*)))
 	     (tv     (get-approx-type approx)))
-	 (if (tvec? tv)
+	 (if (and (tvec? tv) (not tvector?))
 	     (let* ((length-tv (symbol-append (type-id tv) '-length))
-		    (new-node  (sexp->node `(,length-tv ,@args)
+		    (new-node  (sexp->node `(,length-tv ,(car expr*))
 					   '()
 					   loc
 					   'value)))
-		(node-type-set! new-node *long*)
+		(node-type-set! new-node *int*)
 		(inline-node new-node 1 '()))
 	     node))))
 
@@ -529,7 +507,7 @@
 ;*    patch-vector?! ...                                               */
 ;*---------------------------------------------------------------------*/
 (define (patch-vector?! node::app)
-   (with-access::make-vector-app node (args loc)
+   (with-access::app node (args loc)
       (patch*! args)
       (let* ((approx (cfa! (car args)))
 	     (type   (approx-type approx)))
@@ -544,7 +522,7 @@
 ;*    patch-vector->list! ...                                          */
 ;*---------------------------------------------------------------------*/
 (define (patch-vector->list! node::app)
-   (with-access::vector-set!-app node (args loc)
+   (with-access::app node (args loc)
       (patch*! args)
       (let* ((approx (cfa! (car args)))
 	     (tv     (get-approx-type approx)))
@@ -562,7 +540,7 @@
 ;*    patch! ::make-vector-app ...                                     */
 ;*---------------------------------------------------------------------*/
 (define-method (patch! node::make-vector-app)
-   (with-access::make-vector-app node (value-approx fun args loc stackable?)
+   (with-access::make-vector-app node (value-approx fun args loc)
       (patch*! args)
       (let* ((type (approx-type value-approx))
 	     (tv   (type-tvector type)))
@@ -573,64 +551,56 @@
 					   loc
 					   'value)))
 		(node-type-set! new-node tv)
-		(inline-node (if (stack-optimization?)
-				 (node-heap->stack! new-node stackable?)
-				 new-node)
-			     1
-			     '()))
+		(inline-node new-node 1 '()))
 	     node))))
 
 ;*---------------------------------------------------------------------*/
-;*    patch! ::create-vector-app ...                                   */
+;*    patch! ::valloc/Cinfo+optim ...                                  */
 ;*---------------------------------------------------------------------*/
-(define-method (patch! node::create-vector-app)
-   (with-access::create-vector-app node (value-approx fun args loc stackable?)
-      (patch*! args)
+(define-method (patch! node::valloc/Cinfo+optim)
+   (with-access::valloc/Cinfo+optim node (value-approx expr* loc)
+      (patch*! expr*)
       (let* ((type (approx-type value-approx))
 	     (tv   (type-tvector type)))
 	 (if (type? tv)
 	     (let* ((create-tv (symbol-append 'allocate- (type-id tv)))
-		    (new-node  (sexp->node `(,create-tv ,@args)
+		    (new-node  (sexp->node `(,create-tv ,@expr*)
 					   '()
 					   loc
 					   'value)))
 		(node-type-set! new-node tv)
-		(inline-node (if (stack-optimization?)
-				 (node-heap->stack! new-node stackable?)
-				 new-node)
-			     1
-			     '()))
+		(inline-node new-node 1 '()))
 	     node))))
 	    
 ;*---------------------------------------------------------------------*/
-;*    patch! ::vector-ref-app ...                                      */
+;*    patch! ::vref/Cinfo ...                                          */
 ;*---------------------------------------------------------------------*/
-(define-method (patch! node::vector-ref-app)
-   (with-access::vector-ref-app node (fun args loc)
-      (patch*! args)
-      (let* ((vec-approx (cfa! (car args)))
+(define-method (patch! node::vref/Cinfo)
+   (with-access::vref/Cinfo node (expr* loc tvector?)
+      (patch*! expr*)
+      (let* ((vec-approx (cfa! (car expr*)))
 	     (tv         (get-approx-type vec-approx)))
-	 (if (not (tvec? tv))
+	 (if (or tvector? (not (tvec? tv)))
 	     node
 	     (let* ((tv-ref   (symbol-append (type-id tv) '-ref))
-		    (new-node (sexp->node `(,tv-ref ,@args)
-					  '()
+		    (new-node (sexp->node `(,tv-ref ,@expr*)
+ 					  '()
 					  loc
 					  'value)))
 		(inline-node new-node 1 '()))))))
 
 ;*---------------------------------------------------------------------*/
-;*    patch! ::vector-set!-app ...                                     */
+;*    patch! ::vset!/Cinfo ...                                         */
 ;*---------------------------------------------------------------------*/
-(define-method (patch! node::vector-set!-app)
-   (with-access::vector-set!-app node (fun args loc)
-      (patch*! args)
-      (let* ((vec-approx (cfa! (car args)))
+(define-method (patch! node::vset!/Cinfo)
+   (with-access::vset!/Cinfo node (expr* loc tvector? ftype)
+      (patch*! expr*)
+      (let* ((vec-approx (cfa! (car expr*)))
 	     (tv         (get-approx-type vec-approx)))
-	 (if (not (tvec? tv))
+	 (if (or tvector? (not (tvec? tv)))
 	     node
 	     (let* ((tv-set!  (symbol-append (type-id tv) '-set!))
-		    (new-node (sexp->node `(,tv-set! ,@args)
+		    (new-node (sexp->node `(,tv-set! ,@expr*)
 					  '()
 					  loc
 					  'value)))
