@@ -1,36 +1,10 @@
-;* --------------------------------------------------------------------*/
-;*    Copyright (c) 1992-1998 by Manuel Serrano. All rights reserved.  */
-;*                                                                     */
-;*                                     ,--^,                           */
-;*                               _ ___/ /|/                            */
-;*                           ,;'( )__, ) '                             */
-;*                          ;;  //   L__.                              */
-;*                          '   \   /  '                               */
-;*                               ^   ^                                 */
-;*                                                                     */
-;*                                                                     */
-;*    This program is distributed in the hope that it will be useful.  */
-;*    Use and copying of this software and preparation of derivative   */
-;*    works based upon this software are permitted, so long as the     */
-;*    following conditions are met:                                    */
-;*           o credit to the authors is acknowledged following         */
-;*             current academic behaviour                              */
-;*           o no fees or compensation are charged for use, copies,    */
-;*             or access to this software                              */
-;*           o this copyright notice is included intact.               */
-;*      This software is made available AS IS, and no warranty is made */
-;*      about the software or its performance.                         */
-;*                                                                     */
-;*      Bug descriptions, use reports, comments or suggestions are     */
-;*      welcome. Send them to                                          */
-;*        Manuel Serrano -- Manuel.Serrano@unice.fr                    */
-;*-------------------------------------------------------------------- */
 ;*=====================================================================*/
-;*    serrano/prgm/project/bigloo/comptime1.9b/Cgen/cgen.scm           */
+;*    serrano/prgm/project/bigloo/comptime/Cgen/cgen.scm               */
 ;*    -------------------------------------------------------------    */
 ;*    Author      :  Manuel Serrano                                    */
 ;*    Creation    :  Tue Jul  2 13:17:04 1996                          */
-;*    Last change :  Thu Apr  3 14:17:19 1997 (serrano)                */
+;*    Last change :  Tue Oct 25 07:12:44 2005 (serrano)                */
+;*    Copyright   :  1996-2005 Manuel Serrano, see LICENSE file        */
 ;*    -------------------------------------------------------------    */
 ;*    The C production code.                                           */
 ;*=====================================================================*/
@@ -39,36 +13,43 @@
 ;*    The module                                                       */
 ;*---------------------------------------------------------------------*/
 (module cgen_cgen
-   (include "Tools/trace.sch")
+   (include "Tools/trace.sch"
+	    "Tools/location.sch")
    (import  tools_error
 	    tools_shape
+	    engine_param
 	    type_type
 	    type_tools
 	    type_cache
+	    type_typeof
+	    object_class
+	    object_slots
 	    ast_var
 	    ast_node
 	    ast_local
 	    effect_effect
-	    coerce_typeof
-	    cgen_emit
 	    cgen_cop
 	    cgen_emit-cop
-	    cgen_prototype
-	    cgen_capp)
-   (export  (cgen ::global)
+	    cgen_capp
+	    backend_c_emit
+	    backend_c_prototype
+	    backend_backend
+	    backend_cplib)
+   (export  (cgen-function ::global)
 	    (node-setq::setq variable::variable value::node)
 	    (generic node->cop::cop ::node ::procedure)
 	    (make-local-svar/name::local ::symbol ::type)
+	    (bdb-let-var::cop ::cop loc)
 	    *the-global*
 	    *return-kont*
 	    *id-kont*
-	    *block-kont*
+	    (block-kont cop loc)
 	    *stop-kont*))
-  
+
 ;*---------------------------------------------------------------------*/
-;*    cgen ...                                                         */
+;*    cgen-function ...                                                */
 ;*---------------------------------------------------------------------*/
-(define (cgen global::global)
+(define (cgen-function global::global)
    (enter-function (global-id global))
    (trace cgen "cgen global: " (shape global) #\Newline)
    (trace (cgen 2) "*void-kont*: " *void-kont* #\Newline)
@@ -78,17 +59,47 @@
       [assert (sh) (string? (global-name global))])
    (let* ((sfun (widen!::sfun/C (global-value global)
 		   (label (instantiate::clabel
+			     (loc (sfun-loc (global-value global)))
 			     (name (global-name global))))
 		   (integrated #t)))
+	  (loc  (sfun-loc sfun))
 	  (cop  (node->cop (sfun-body sfun)
 			   (if (eq? (global-type global) *void*)
 			       *void-kont*
 			       *return-kont*))))
+      (reset-bdb-loc!)
+      (newline *c-port*)
+      (newline *c-port*)
+      (display "/* " *c-port*)
+      (display (shape global) *c-port*)
+      (display " */" *c-port*)
+      ;; we have to emit a dummy location otherwise gdb get confused
+      ;; with the function arguments! Thus all functions looks like
+      ;; starting at the module definition site instead of there correct
+      ;; definition site
+      (emit-bdb-loc #f)
+      ;; we have to reset the loc because we must not emit location
+      ;; before the first c expression otherwise gdb get confused
+      (reset-bdb-loc!)
       (clabel-body-set! (sfun/C-label sfun) cop)
-      (sfun->c sfun global)
-      (emit-cop (*block-kont* (sfun/C-label sfun)))
-      (newline *c-port*)
-      (newline *c-port*)
+      (global->c global)
+      (let ((cop (block-kont (sfun/C-label sfun) loc)))
+	 ;; we define a local variable that acts as a temporary variable
+	 (display "{ AN_OBJECT;" *c-port*)
+	 ;; when compiling for debugging, we have to insert a dummy
+	 ;; statement otherwise gdb get confused
+	 (if (and (> *bdb-debug* 0) (location? loc))
+	     (begin
+		(emit-bdb-loc loc)
+		(display "{ obj_t ___ = BUNSPEC; } /* bdb dummy init stmt */"
+			 *c-port*)))
+	 ;; we now may emit the body
+	 (emit-cop cop)
+	 ;; emit the current location before the closing bracket
+	 (emit-bdb-loc (get-current-bdb-loc))
+	 ;; and then clause the function body
+	 (fprint *c-port* "}"))
+      (no-bdb-newline)
       (leave-function)))
   
 ;*---------------------------------------------------------------------*/
@@ -99,24 +110,31 @@
 (define *the-global* #unspecified)
 
 ;*---------------------------------------------------------------------*/
-;*    sfun->c ...                                                      */
+;*    global->c ...                                                    */
 ;*---------------------------------------------------------------------*/
-(define (sfun->c sfun variable)
-   (with-access::variable variable (type id name)
-      (display (make-typed-declaration type name) *c-port*)
-      (if (null? (sfun-args sfun))
-	  (fprint *c-port* "()")
-	  (begin
-	     (display #\( *c-port*)
-	     (let loop ((args (sfun-args sfun)))
-		(if (null? (cdr args))
-		    (with-access::local (car args) (name id type)
-		       (display (make-typed-declaration type name) *c-port*)
-		       (fprint *c-port* #\)))
-		    (with-access::local (car args) (name id type)
-		       (display (make-typed-declaration type name) *c-port*)
-		       (display ", " *c-port*)
-		       (loop (cdr args)))))))))
+(define (global->c g::global)
+   (with-access::global g (type id name import value)
+      (when (eq? import 'export) (display "BGL_EXPORTED_DEF " *c-port*))
+      (display (make-typed-declaration
+		type
+		(string-append
+		 name
+		 (if (null? (sfun-args value))
+		     "()"
+		     (string-append
+		      "("
+		      (let loop ((args (sfun-args value)))
+			 (if (null? (cdr args))
+			     (with-access::local (car args) (name id type)
+				(string-append
+				 (make-typed-declaration type name) ")"))
+			     (with-access::local (car args) (name id type)
+				(string-append
+				 (make-typed-declaration type name)
+				 ", "
+				 (loop (cdr args))))))))))
+	       *c-port*))
+   (no-bdb-newline))
 
 ;*---------------------------------------------------------------------*/
 ;*    *return-kont* ...                                                */
@@ -124,9 +142,11 @@
 (define *return-kont*
    (lambda (cop)
       (instantiate::creturn
+	 (loc   (cop-loc cop))
 	 (value (cond
 		   ((csetq? cop)
 		    (instantiate::csequence
+		       (loc    (cop-loc cop))
 		       (c-exp? #t)
 		       (cops (list cop (instantiate::catom
 					  (value #unspecified))))))
@@ -153,15 +173,16 @@
       (instantiate::stop (value cop))))
 
 ;*---------------------------------------------------------------------*/
-;*    *block-kont* ...                                                 */
+;*    block-kont ...                                                   */
 ;*---------------------------------------------------------------------*/
-(define *block-kont*
-   (lambda (cop)
-      (cond
-	 ((block? cop)
-	  cop)
-	 (else
-	  (instantiate::block (body cop))))))
+(define (block-kont cop loc)
+   (cond
+      ((block? cop)
+       cop)
+      (else
+       (instantiate::block
+	  (body cop)
+	  (loc  loc)))))
 
 ;*---------------------------------------------------------------------*/
 ;*    *fail-kont* ...                                                  */
@@ -176,7 +197,7 @@
 ;*---------------------------------------------------------------------*/
 ;*    make-setq-kont ...                                               */
 ;*---------------------------------------------------------------------*/
-(define (make-setq-kont var kont)
+(define (make-setq-kont var loc kont)
    (lambda (cop)
       (if (cfail? cop)
 	  cop
@@ -185,12 +206,15 @@
 		   (value (cond
 			     ((csetq? cop)
 			      (instantiate::csequence
+				 (loc (cop-loc cop))
 				 (c-exp? #t)
 				 (cops (list cop
 					       (instantiate::catom
+						  (loc (cop-loc cop))
 						  (value #unspecified))))))
 			     (else
-			      cop))))))))
+			      cop)))
+		   (loc loc))))))
 			  
 ;*---------------------------------------------------------------------*/
 ;*    node->cop ...                                                    */
@@ -201,16 +225,20 @@
 ;*    node->cop ...                                                    */
 ;*---------------------------------------------------------------------*/
 (define-method (node->cop node::atom kont)
-   (trace (cgen 3) "(node->cop node::atom kont): " (shape node) #\Newline
+   (trace (cgen 3)
+	  "(node->cop node::atom kont): " (shape node) #\Newline
 	  "  kont: " kont #\Newline)
-   (with-access::atom node (value)
-      (kont (instantiate::catom (value value)))))
+   (with-access::atom node (value loc)
+      (kont (instantiate::catom
+	       (value value)
+	       (loc   loc)))))
 
 ;*---------------------------------------------------------------------*/
 ;*    node->cop ...                                                    */
 ;*---------------------------------------------------------------------*/
 (define-method (node->cop node::kwote kont)
-   (trace (cgen 3) "(node->cop node::kwote kont): " (shape node) #\Newline
+   (trace (cgen 3)
+	  "(node->cop node::kwote kont): " (shape node) #\Newline
 	  "  kont: " kont #\Newline)
    (internal-error "node->cop" "Unexpected `kwote' node" (shape node)))
 
@@ -218,16 +246,23 @@
 ;*    node->cop ...                                                    */
 ;*---------------------------------------------------------------------*/
 (define-method (node->cop node::var kont)
-   (trace (cgen 3) "(node->cop node::var kont): " (shape node) #\Newline
+   (trace (cgen 3)
+	  "(node->cop node::var kont): " (shape node) #\Newline
 	  "  kont: " kont #\Newline)
-   (with-access::var node (variable)
-      (kont (instantiate::varc (variable variable)))))
+   (trace (cgen 4) "(node->cop node::var kont): " (shape node) #\Newline
+	  "  var-variable-name: " (variable-name (var-variable node))
+	  #\Newline)
+   (with-access::var node (variable loc)
+      (kont (instantiate::varc
+	       (variable variable)
+	       (loc      loc)))))
 
 ;*---------------------------------------------------------------------*/
 ;*    node->cop ::closure ...                                          */
 ;*---------------------------------------------------------------------*/
 (define-method (node->cop node::closure kont)
-   (trace (cgen 3) "(node->cop node::closure kont): " (shape node) #\Newline
+   (trace (cgen 3)
+	  "(node->cop node::closure kont): " (shape node) #\Newline
 	  "  kont: " kont #\Newline)
    (internal-error "node->cop" "Unexpected `closure' node" (shape node)))
 
@@ -235,13 +270,15 @@
 ;*    node->cop ::sequence ...                                         */
 ;*---------------------------------------------------------------------*/
 (define-method (node->cop node::sequence kont)
-   (trace (cgen 3) "(node->cop node::sequence kont): " (shape node) #\Newline
+   (trace (cgen 3)
+	  "(node->cop node::sequence kont): " (shape node) #\Newline
 	  "  kont: " kont #\Newline)
-   (with-access::sequence node (nodes)
+   (with-access::sequence node (nodes loc)
       (let ((exp nodes))
 	 (cond
 	    ((null? exp)
-	     (kont (instantiate::nop)))
+	     (kont (instantiate::nop
+		      (loc loc))))
 	    ((null? (cdr exp))
 	     (let ((cop (node->cop (car exp) kont)))
 		(instantiate::stop (value cop))))
@@ -249,166 +286,204 @@
 	     (let loop ((exp exp)
 			(new '()))
 		(if (null? (cdr exp))
-		    (instantiate::csequence
-		       (cops (reverse! (cons (node->cop (car exp) kont) new))))
+		    (let ((cop (node->cop (car exp) kont)))
+		       (instantiate::csequence
+			  (loc  (cop-loc cop))
+			  (cops (reverse! (cons cop new)))))
 		    (begin
 		       (if (not (side-effect? (car exp)))
 			   (loop (cdr exp) new)
 			   (loop (cdr exp)
 				 (cons (node->cop (car exp) *stop-kont*)
 				       new)))))))))))
+
+;*---------------------------------------------------------------------*/
+;*    extern->cop ...                                                  */
+;*---------------------------------------------------------------------*/
+(define (extern->cop format::bstring args-safe node::extern kont)
+   (trace (cgen 3)
+	  "(extern->cop node::extern kont): " (shape node) #\Newline
+	  "  kont: " kont #\Newline)
+   (with-access::extern node (expr* loc)
+      (node-args->cop expr*
+		      args-safe
+		      loc
+		      (lambda (new-args)
+			 (kont (instantiate::cpragma
+				  (loc loc)
+				  (format format)
+				  (args new-args)))))))
    
 ;*---------------------------------------------------------------------*/
 ;*    node->cop ::pragma ...                                           */
 ;*---------------------------------------------------------------------*/
 (define-method (node->cop node::pragma kont)
-   (trace (cgen 3) "(node->cop node::pragma kont): " (shape node) #\Newline
+   (trace (cgen 3)
+	  "(node->cop node::pragma kont): " (shape node) #\Newline
 	  "  kont: " kont #\Newline)
-   (with-access::pragma node (format args)
-      (let loop ((old-actuals  args)
-		 (new-actuals  '())
-		 (aux          (make-local-svar/name 'aux *obj*))
-		 (auxs         '())
-		 (exps         '()))
-	 (if (null? old-actuals)
-	     (if (null? auxs)
-		 (kont (instantiate::cpragma
-			  (format format)
-			  (args (reverse! new-actuals))))
-		 (instantiate::block
-		    (body (instantiate::csequence
-			     (cops
-			      (list
-			       (instantiate::local-var (vars auxs))
-			       (instantiate::csequence (cops exps))
-			       (kont (instantiate::cpragma
-					(format format)
-					(args (reverse! new-actuals))))))))))
-	     (let ((cop (node->cop (node-setq aux (car old-actuals))
-				   *id-kont*)))
-		(if (and (csetq? cop)
-			 (eq? (varc-variable (csetq-var cop)) aux)
-			 (or (catom? (csetq-value cop))
-			     (varc? (csetq-value cop))
-			     (cpragma? (csetq-value cop))))
-		    (loop (cdr old-actuals)
-			  (cons (csetq-value cop) new-actuals)
-			  aux
-			  auxs
-			  exps)
-		    (begin
-		       (local-type-set! aux (typeof (car old-actuals)))
-		       (loop (cdr old-actuals)
-			     (cons (instantiate::varc (variable aux))
-				   new-actuals)
-			     (make-local-svar/name 'aux *obj*)
-			     (cons aux auxs)
-			     (cons cop exps)))))))))
+   (with-access::pragma node (format)
+      (extern->cop format #f node kont)))
 
+;*---------------------------------------------------------------------*/
+;*    node->cop ::getfield ...                                         */
+;*---------------------------------------------------------------------*/
+(define-method (node->cop node::getfield kont)
+   (trace (cgen 3)
+	  "(node->cop node::getfield kont): " (shape node) #\Newline
+	  "  kont: " kont #\Newline)
+   (with-access::getfield node (fname otype)
+      (let ((fmt (string-append "(((" (type-name otype) ")CREF($1))->"
+				fname ")")))
+	 (extern->cop fmt #t node kont))))
+   
+;*---------------------------------------------------------------------*/
+;*    node->cop ::setfield ...                                         */
+;*---------------------------------------------------------------------*/
+(define-method (node->cop node::setfield kont)
+   (trace (cgen 3)
+	  "(node->cop node::setfield kont): " (shape node) #\Newline
+	  "  kont: " kont #\Newline)
+   (with-access::setfield node (fname ftype otype)
+      (let ((fmt (string-append "((((" (type-name otype) ")CREF($1))->"
+				fname ") = ((" (type-name ftype)
+				")$2), BUNSPEC)")))
+	 (extern->cop fmt #t node kont))))
+
+;*---------------------------------------------------------------------*/
+;*    node->cop ::vlength ...                                          */
+;*---------------------------------------------------------------------*/
+(define-method (node->cop node::vlength kont)
+   (with-access::vlength node (c-format)
+      (trace (cgen 3)
+	     "(node->cop node::vlength kont): " (shape node) #\Newline
+	     "  kont: " kont " " c-format #\Newline)
+      (extern->cop c-format #t node kont)))
+   
+;*---------------------------------------------------------------------*/
+;*    node->cop ::vref ...                                             */
+;*---------------------------------------------------------------------*/
+(define-method (node->cop node::vref kont)
+   (with-access::vref node (c-format)
+      (trace (cgen 3)
+	     "(node->cop node::vref kont): " (shape node) #\Newline
+	     "  kont: " kont " " c-format #\Newline)
+      (extern->cop c-format #t node kont)))
+   
+;*---------------------------------------------------------------------*/
+;*    node->cop ::vset! ...                                            */
+;*---------------------------------------------------------------------*/
+(define-method (node->cop node::vset! kont)
+   (with-access::vset! node (c-format)
+      (trace (cgen 3)
+	     "(node->cop node::vset! kont): " (shape node) #\Newline
+	     "  kont: " kont " " c-format #\Newline)
+      (extern->cop c-format #t node kont)))
+   
+;*---------------------------------------------------------------------*/
+;*    node->cop ::valloc ...                                           */
+;*---------------------------------------------------------------------*/
+(define-method (node->cop node::valloc kont)
+   (with-access::valloc node (c-heap-format)
+      (trace (cgen 3)
+	     "(node->cop node::vallocate kont): " (shape node) #\Newline
+	     "  kont: " kont " " c-heap-format " " #\Newline) 
+      (extern->cop c-heap-format #t node kont)))
+   
 ;*---------------------------------------------------------------------*/
 ;*    node->cop ::cast ...                                             */
 ;*---------------------------------------------------------------------*/
 (define-method (node->cop node::cast kont)
-   (trace (cgen 3) "(node->cop node::cast kont): " (shape node) #\Newline
+   (trace (cgen 3)
+	  "(node->cop node::cast kont): " (shape node) #\Newline
 	  "  kont: " kont #\Newline)
-   (with-access::cast node (arg type)
-      (kont (instantiate::ccast
-	       (type type)
-	       (arg (node->cop arg *id-kont*))))))
+   (with-access::cast node (arg type loc)
+      (node-args->cop (list arg)
+		      #t
+		      loc
+		      (lambda (new-args)
+			 (kont (instantiate::ccast
+				  (type type)
+				  (loc loc)
+				  (arg (car new-args))))))))
 
 ;*---------------------------------------------------------------------*/
 ;*    node->cop ::setq ...                                             */
 ;*---------------------------------------------------------------------*/
 (define-method (node->cop node::setq kont)
-   (trace (cgen 3) "(node->cop node::setq kont): " (shape node) #\Newline
+   (trace (cgen 3)
+	  "(node->cop node::setq kont): " (shape node) #\Newline
 	  "  kont: " kont #\Newline)
-   (with-access::setq node (value)
+   (with-access::setq node (value loc)
       (let ((var (var-variable (setq-var node))))
 	 (if (and (var? value) (eq? var (var-variable value)))
-	     (kont (*void-kont* (instantiate::catom (value #unspecified))))
-	     (node->cop value (make-setq-kont var kont))))))
+	     (kont (*void-kont* (instantiate::catom
+				   (loc loc)
+				   (value #unspecified))))
+	     (node->cop value (make-setq-kont var loc kont))))))
 
 ;*---------------------------------------------------------------------*/
 ;*    node->cop ::conditional ...                                      */
 ;*---------------------------------------------------------------------*/
 (define-method (node->cop node::conditional kont)
-   (trace (cgen 3) "(node->cop node::conditional kont): " (shape node) #\Newline
+   (trace (cgen 3)
+	  "(node->cop node::conditional kont): " (shape node) #\Newline
+	  "  loc: " (node-loc node) #\Newline
 	  "  kont: " kont #\Newline)
-   (with-access::conditional node (test true false)
+   (with-access::conditional node (test true false loc)
       (let* ((aux   (make-local-svar/name 'test *bool*))
 	     (ctest (node->cop (node-setq aux test) *id-kont*)))
 	 (if (and (csetq? ctest) (eq? (varc-variable (csetq-var ctest)) aux))
 	     (instantiate::cif
 		(test (csetq-value ctest))
-		(true (*block-kont* (node->cop true kont)))
-		(false (*block-kont* (node->cop false kont))))
+		(true (block-kont (node->cop true kont) loc))
+		(false (block-kont (node->cop false kont) loc))
+		(loc   loc))
 	     (instantiate::block
+		(loc loc)
 		(body
 		 (instantiate::csequence
+		    (loc loc)
 		    (cops
 		     (list
-		      (instantiate::local-var (vars (list aux)))
+		      (instantiate::local-var
+			 (vars (list aux))
+			 (loc  loc))
 		      ctest
 		      (instantiate::cif
-			 (test (instantiate::varc (variable aux)))
-			 (false (*block-kont* (node->cop false kont)))
-			 (true (*block-kont* (node->cop true kont)))))))))))))
+			 (test (instantiate::varc
+				  (variable aux)
+				  (loc loc)))
+			 (false (block-kont (node->cop false kont) loc))
+			 (true (block-kont (node->cop true kont) loc))
+			 (loc  loc)))))))))))
 
 ;*---------------------------------------------------------------------*/
 ;*    node->cop ::fail ...                                             */
 ;*---------------------------------------------------------------------*/
 (define-method (node->cop node::fail kont)
-   (trace (cgen 3) "(node->cop node::fail kont): " (shape node) #\Newline
+   (trace (cgen 3)
+	  "(node->cop node::fail kont): " (shape node) #\Newline
 	  "  kont: " kont #\Newline)
-   (with-access::fail node (proc msg obj)
-      (let loop ((old-actuals  (list proc msg obj))
-		 (new-actuals  '())
-		 (aux          (make-local-svar/name 'aux *obj*))
-		 (auxs         '())
-		 (exps         '()))
-	 (if (null? old-actuals)
-	     (if (null? auxs)
-		 (*fail-kont* (instantiate::cfail
-				 (proc (caddr new-actuals))
-				 (msg  (cadr new-actuals))
-				 (obj  (car new-actuals))))
-		 (instantiate::block
-		    (body (instantiate::csequence
-			     (cops (list (instantiate::local-var
-					    (vars auxs))
-					 (instantiate::csequence
-					    (cops exps))
-					 (*fail-kont*
-					  (instantiate::cfail
-					     (proc (caddr new-actuals))
-					     (msg (cadr new-actuals))
-					     (obj (car new-actuals))))))))))
-	     (let ((cop (node->cop (node-setq aux (car old-actuals))
-				   *id-kont*)))
-		(if (and (csetq? cop)
-			 (eq? (varc-variable (csetq-var cop)) aux))
-		    ;; the local is useless, we ignore it
-		    (loop (cdr old-actuals)
-			  (cons (csetq-value cop) new-actuals)
-			  aux
-			  auxs
-			  exps)
-		    (begin
-		       (local-type-set! aux *obj*)
-		       (loop (cdr old-actuals)
-			     (cons (instantiate::varc (variable aux))
-				   new-actuals)
-			     (make-local-svar/name 'aux *obj*)
-			     (cons aux auxs)
-			     (cons cop exps)))))))))
+   (with-access::fail node (proc msg obj loc)
+      (node-args->cop (list proc msg obj)
+		      #f
+		      loc
+		      (lambda (new-args)
+			 (*fail-kont*
+			  (instantiate::cfail
+			     (loc loc)
+			     (proc (car new-args))
+			     (msg (cadr new-args))
+			     (obj (caddr new-args))))))))
 
 ;*---------------------------------------------------------------------*/
 ;*    node->cop ::select ...                                           */
 ;*---------------------------------------------------------------------*/
 (define-method (node->cop node::select kont)
-   (trace (cgen 3) "(node->cop node::select kont): " (shape node) #\Newline
+   (trace (cgen 3)
+	  "(node->cop node::select kont): " (shape node) #\Newline
 	  "  kont: " kont #\Newline)
-   (with-access::select node (clauses test item-type)
+   (with-access::select node (clauses test item-type loc)
       (for-each (lambda (clause)
 		   (set-cdr! clause (node->cop (cdr clause) kont)))
 		clauses)
@@ -417,15 +492,21 @@
 	    (if (and (csetq? cop)
 		     (eq? (varc-variable (csetq-var cop)) aux))
 		(instantiate::cswitch
+		   (loc  loc)
 		   (test (csetq-value cop))
 		   (clauses clauses))
 		(instantiate::block
+		   (loc loc)
 		   (body (instantiate::csequence
+			    (loc loc)
 			    (cops (list (instantiate::local-var
+					   (loc  loc)
 					   (vars (list aux)))
 					cop
 					(instantiate::cswitch
+					   (loc  loc)
 					   (test (instantiate::varc
+						    (loc loc)
 						    (variable aux)))
 					   (clauses clauses))))))))))))
 
@@ -433,9 +514,10 @@
 ;*    node->cop ::let-fun ...                                          */
 ;*---------------------------------------------------------------------*/
 (define-method (node->cop node::let-fun kont)
-   (trace (cgen 3) "(node->cop node::let-fun kont): " (shape node) #\Newline
+   (trace (cgen 3)
+	  "(node->cop node::let-fun kont): " (shape node) #\Newline
 	  "  kont: " kont #\Newline)
-   (with-access::let-fun node (body locals)
+   (with-access::let-fun node (body locals loc)
       ;; local function are open-coded of their first call site.
       ;; So, the compilation of `let-fun' construction is just
       ;; the declaration of all local functions' formals (and
@@ -444,16 +526,23 @@
       (let loop ((locals      locals)
 		 (all-formals '()))
 	 (if (null? locals)
-	     (*block-kont*
-	      (instantiate::csequence
-		 (cops (list (instantiate::local-var (vars all-formals))
-			       (node->cop body kont)))))
+	     (block-kont
+	      (bdb-let-var
+	       (instantiate::csequence
+		  (loc loc)
+		  (cops (list (instantiate::local-var
+				 (loc  loc)
+				 (vars all-formals))
+			      (node->cop body kont))))
+	       loc)
+	      #f)
 	     (let ((local (car locals)))
 		(set-variable-name! local)
-		(let* ((fun   (widen!::sfun/C (local-value local)
-				 (label (instantiate::clabel
-					   (name (local-name local))))
-				 (integrated #f)))
+		(let* ((fun (widen!::sfun/C (local-value local)
+			       (label (instantiate::clabel
+					 (loc  (sfun-loc (local-value local)))
+					 (name (local-name local))))
+			       (integrated #f)))
 		       (formals (sfun-args fun)))
 		   (for-each set-variable-name! formals)
 		   (loop (cdr locals) (append formals all-formals))))))))
@@ -462,50 +551,78 @@
 ;*    node->cop ::let-var ...                                          */
 ;*---------------------------------------------------------------------*/
 (define-method (node->cop node::let-var kont)
-   (trace (cgen 3) "(node->cop node::let-var kont): " (shape node) #\Newline
+   (trace (cgen 3)
+	  "(node->cop node::let-var kont): " (shape node) #\Newline
 	  "  kont: " kont #\Newline)
-   (with-access::let-var node (body bindings)
-      (let ((decls (instantiate::local-var (vars (map car bindings))))
+   (with-access::let-var node (body bindings loc)
+      (let ((decls (instantiate::local-var
+		      (loc loc)
+		      (vars (map car bindings))))
 	    (sets  (map (lambda (x)
 			   (set-variable-name! (car x))
 			   (node->cop (node-setq (car x) (cdr x)) *stop-kont*))
 			bindings))
 	    (body  (let ((cop (node->cop body kont)))
-		      (instantiate::stop (value cop)))))
-	 (*block-kont* (instantiate::csequence
-			  (cops (cons decls (append sets (list body)))))))))
- 
+		      (instantiate::stop
+			 (value cop)))))
+	 (block-kont
+	  (bdb-let-var
+	   (instantiate::csequence
+	      (loc loc)
+	      (cops (cons decls (append sets (list body)))))
+	   loc)
+	  loc))))
+
+;*---------------------------------------------------------------------*/
+;*    bdb-let-var ...                                                  */
+;*---------------------------------------------------------------------*/
+(define (bdb-let-var cop loc)
+   (if (and (>fx *bdb-debug* 0) (location? loc))
+       (instantiate::bdb-block
+	  (body cop)
+	  (loc  loc))
+       cop))
+   
 ;*---------------------------------------------------------------------*/
 ;*    node->cop ::set-ex-it ...                                        */
 ;*---------------------------------------------------------------------*/
 (define-method (node->cop node::set-ex-it kont)
-   (trace (cgen 3) "(node->cop node::set-ex-it kont): " (shape node) #\Newline
+   (trace (cgen 3)
+	  "(node->cop node::set-ex-it kont): " (shape node) #\Newline
 	  "  kont: " kont #\Newline)
-   (with-access::set-ex-it node (var body)
+   (with-access::set-ex-it node (var body loc)
       (let ((exit (var-variable var)))
 	 (set-variable-name! exit)
 	 (instantiate::csequence
+	    (loc loc)
 	    (cops (list
 		   (instantiate::cpragma
+		      (loc    loc)
 		      (format "jmp_buf jmpbuf")
 		      (args '()))
-		   (instantiate::local-var (vars (list (var-variable var))))
+		   (instantiate::local-var
+		      (loc loc)
+		      (vars (list (var-variable var))))
 		   (instantiate::cset-ex-it
-		      (exit (instantiate::varc (variable exit)))
+		      (loc loc)
+		      (exit (instantiate::varc
+			       (loc loc)
+			       (variable exit)))
 		      (jump-value (node->cop (instantiate::pragma
-						(loc #f)
+						(loc loc)
 						(type *_*)
 						(format "_exit_value_")
-						(args '()))
+						(expr* '()))
 					     kont))
 		      (body (instantiate::csequence
+			       (loc loc)
 			       (cops
 				(list
 				 (node->cop
 				  (node-setq
 				   exit
 				   (instantiate::pragma
-				      (loc #f)
+				      (loc loc)
 				      (type *_*)
 				      (format
 				       (string-append
@@ -513,7 +630,7 @@
 					(string-sans-$
 					 (type-name (local-type exit)))
 					")jmpbuf"))
-				      (args '())))
+				      (expr* '())))
 				  *id-kont*)
 				 (node->cop body kont))))))))))))
 
@@ -521,9 +638,10 @@
 ;*    node->cop ::jump-ex-it ...                                       */
 ;*---------------------------------------------------------------------*/
 (define-method (node->cop node::jump-ex-it kont)
-   (trace (cgen 3) "(node->cop node::jump-ex-it kont): " (shape node) #\Newline
+   (trace (cgen 3)
+	  "(node->cop node::jump-ex-it kont): " (shape node) #\Newline
 	  "  kont: " kont #\Newline)
-   (with-access::jump-ex-it node (exit value)
+   (with-access::jump-ex-it node (exit value loc)
       (let* ((vaux  (make-local-svar/name 'aux *obj*))
 	     (vcop  (node->cop (node-setq vaux value) *id-kont*))
 	     (exit  exit)
@@ -534,43 +652,61 @@
 		  (csetq? ecop) (eq? (varc-variable (csetq-var ecop)) eaux))
 	     (*exit-kont*
 	      (instantiate::cjump-ex-it
+		 (loc loc)
 		 (exit (csetq-value ecop))
 		 (value (csetq-value vcop)))))
 	    ((and (csetq? vcop) (eq? (varc-variable (csetq-var vcop)) vaux))
 	     (instantiate::block
+		(loc loc)
 		(body (instantiate::csequence
+			 (loc loc)
 			 (cops (list (instantiate::local-var
-					  (vars (list eaux)))
+					(loc loc)
+					(vars (list eaux)))
 				       (instantiate::csequence
+					  (loc loc)
 					  (cops (list ecop)))
 				       (*exit-kont*
 					(instantiate::cjump-ex-it
+					   (loc loc)
 					   (exit (instantiate::varc
+						    (loc loc)
 						    (variable eaux)))
 					   (value (csetq-value vcop))))))))))
 	    ((and (csetq? ecop) (eq? (varc-variable (csetq-var ecop)) eaux))
 	     (instantiate::block
+		(loc loc)
 		(body (instantiate::csequence
+			 (loc loc)
 			 (cops (list (instantiate::local-var
-					  (vars (list vaux)))
+					(loc loc)
+					(vars (list vaux)))
 				     (instantiate::csequence
+					(loc loc)
 					(cops (list vcop)))
 				     (*exit-kont*
 				      (instantiate::cjump-ex-it
+					 (loc loc)
 					 (exit (csetq-value ecop))
 					 (value (instantiate::varc
+						   (loc loc)
 						   (variable vaux)))))))))))
 	    (else
 	     (instantiate::block
+		(loc loc)
 		(body (instantiate::csequence
+			 (loc loc)
 			 (cops
 			  (list
 			   (instantiate::local-var
+			      (loc loc)
 			      (vars (list eaux vaux)))
 			   (instantiate::csequence
+			      (loc loc)
 			      (cops (list ecop vcop)))
 			   (*exit-kont*
 			    (instantiate::cjump-ex-it
+			       (loc loc)
 			       (exit (instantiate::varc
 					(variable eaux)))
 			       (value (instantiate::varc
@@ -580,44 +716,59 @@
 ;*    node->cop ::make-box ...                                         */
 ;*---------------------------------------------------------------------*/
 (define-method (node->cop node::make-box kont)
-   (trace (cgen 3) "(node->cop node::make-box kont): " (shape node) #\Newline
+   (trace (cgen 3)
+	  "(node->cop node::make-box kont): " (shape node) #\Newline
 	  "  kont: " kont #\Newline)
-   (with-access::make-box node (value)
+   (with-access::make-box node (value loc)
       (if (or (var? value) (atom? value) (kwote? value))
 	  (node->cop value
-		     (lambda (v) (kont (instantiate::cmake-box (value v)))))
-	  (let* ((aux  (make-local-svar/name 'cell-val *obj*))
+		     (lambda (v) (kont (instantiate::cmake-box
+					  (loc loc)
+					  (value v)))))
+	  (let* ((aux  (make-local-svar/name 'cellval *obj*))
 		 (cval (node->cop (node-setq aux value) *id-kont*)))
 	     (instantiate::block
+		(loc loc)
 		(body (instantiate::csequence
+			 (loc loc)
 			 (cops (list
-				(instantiate::local-var (vars (list aux)))
+				(instantiate::local-var
+				   (loc loc)
+				   (vars (list aux)))
 				cval
 				(kont
 				 (instantiate::cmake-box
+				    (loc loc)
 				    (value (instantiate::varc
+					      (loc loc)
 					      (variable aux))))))))))))))
 
 ;*---------------------------------------------------------------------*/
 ;*    node->cop ::box-ref ...                                          */
 ;*---------------------------------------------------------------------*/
 (define-method (node->cop node::box-ref kont)
-   (trace (cgen 3) "(node->cop node::box-ref kont): " (shape node) #\Newline
+   (trace (cgen 3)
+	  "(node->cop node::box-ref kont): " (shape node) #\Newline
 	  "  kont: " kont #\Newline)
-   (with-access::box-ref node (var)
-      (kont (node->cop var (lambda (v) (instantiate::cbox-ref (var v)))))))
+   (with-access::box-ref node (var loc)
+      (kont (node->cop var (lambda (v) (instantiate::cbox-ref
+					  (loc loc)
+					  (var v)))))))
 
 ;*---------------------------------------------------------------------*/
 ;*    node->cop ::box-set! ...                                         */
 ;*---------------------------------------------------------------------*/
 (define-method (node->cop node::box-set! kont)
-   (trace (cgen 3) "(node->cop node::box-set! kont): " (shape node) #\Newline
+   (trace (cgen 3)
+	  "(node->cop node::box-set! kont): " (shape node) #\Newline
 	  "  kont: " kont #\Newline)
-   (with-access::box-set! node (var value)
+   (with-access::box-set! node (var value loc)
       (let ((v (var-variable var)))
 	 (node->cop value
 		    (lambda (vl) (kont (instantiate::cbox-set!
+					  (loc loc)
 					  (var (instantiate::varc
+						  (loc loc)
 						  (variable v)))
 					  (value vl))))))))
 
@@ -626,7 +777,7 @@
 ;*---------------------------------------------------------------------*/
 (define (node-setq::setq variable::variable value::node)
    (instantiate::setq
-      (loc #f)
+      (loc (node-loc value))
       (type *unspec*)
       (var (instantiate::var
 	      (loc #f)
@@ -639,5 +790,61 @@
 ;*---------------------------------------------------------------------*/
 (define (make-local-svar/name::local id::symbol type::type)
    (let ((local (make-local-svar id type)))
-      (set-variable-name! local)
+      (if (not (string? (local-name local)))
+	  (error "make-local-svar/name" "Illegal local name" local))
       local))
+
+;*---------------------------------------------------------------------*/
+;*    no-bdb-newline ...                                               */
+;*    -------------------------------------------------------------    */
+;*    Emit a newline only if not compiling for bdb.                    */
+;*---------------------------------------------------------------------*/
+(define (no-bdb-newline)
+   (if (eq? *bdb-debug* 0)
+       (newline *c-port*)))
+
+;*---------------------------------------------------------------------*/
+;*    node-args->cop ...                                               */
+;*---------------------------------------------------------------------*/
+(define (node-args->cop args args-safe loc kont)
+   (let loop ((old-actuals  args)
+	      (new-actuals  '())
+	      (aux          (make-local-svar/name 'aux *obj*))
+	      (auxs         '())
+	      (exps         '()))
+      (if (null? old-actuals)
+	  (if (null? auxs)
+	      (kont (reverse! new-actuals))
+	      (instantiate::block
+		 (body (instantiate::csequence
+			  (loc loc)
+			  (cops (list
+				 (instantiate::local-var
+				    (vars auxs)
+				    (loc  loc))
+				 (instantiate::csequence (cops exps))
+				 (kont (reverse! new-actuals))))))))
+	  (let ((cop (node->cop (node-setq aux (car old-actuals)) *id-kont*)))
+	     (if (and (csetq? cop)
+		      (eq? (varc-variable (csetq-var cop)) aux)
+		      (or args-safe
+			  (catom? (csetq-value cop))
+			  (varc? (csetq-value cop))
+			  (cpragma? (csetq-value cop))))
+		 (loop (cdr old-actuals)
+		       (cons (csetq-value cop) new-actuals)
+		       aux
+		       auxs
+		       exps)
+		 (begin
+		    (local-type-set! aux (typeof (car old-actuals)))
+		    (loop (cdr old-actuals)
+			  (cons (instantiate::varc
+				   (variable aux)
+				   (loc      loc))
+				new-actuals)
+			  (make-local-svar/name 'aux *obj*)
+			  (cons aux auxs)
+			  (cons cop exps))))))))
+   
+   
