@@ -1,36 +1,10 @@
-;* --------------------------------------------------------------------*/
-;*    Copyright (c) 1992-1998 by Manuel Serrano. All rights reserved.  */
-;*                                                                     */
-;*                                     ,--^,                           */
-;*                               _ ___/ /|/                            */
-;*                           ,;'( )__, ) '                             */
-;*                          ;;  //   L__.                              */
-;*                          '   \   /  '                               */
-;*                               ^   ^                                 */
-;*                                                                     */
-;*                                                                     */
-;*    This program is distributed in the hope that it will be useful.  */
-;*    Use and copying of this software and preparation of derivative   */
-;*    works based upon this software are permitted, so long as the     */
-;*    following conditions are met:                                    */
-;*           o credit to the authors is acknowledged following         */
-;*             current academic behaviour                              */
-;*           o no fees or compensation are charged for use, copies,    */
-;*             or access to this software                              */
-;*           o this copyright notice is included intact.               */
-;*      This software is made available AS IS, and no warranty is made */
-;*      about the software or its performance.                         */
-;*                                                                     */
-;*      Bug descriptions, use reports, comments or suggestions are     */
-;*      welcome. Send them to                                          */
-;*        Manuel Serrano -- Manuel.Serrano@unice.fr                    */
-;*-------------------------------------------------------------------- */
 ;*=====================================================================*/
-;*    serrano/prgm/project/bigloo/comptime1.9b/Cgen/emit-cop.scm       */
+;*    serrano/prgm/project/bigloo/comptime/Cgen/emit-cop.scm           */
 ;*    -------------------------------------------------------------    */
 ;*    Author      :  Manuel Serrano                                    */
 ;*    Creation    :  Tue Jul  2 14:39:37 1996                          */
-;*    Last change :  Tue Jun 24 11:30:57 1997 (serrano)                */
+;*    Last change :  Wed Aug 23 14:17:15 2006 (serrano)                */
+;*    Copyright   :  1996-2006 Manuel Serrano, see LICENSE file        */
 ;*    -------------------------------------------------------------    */
 ;*    The emission of cop code.                                        */
 ;*=====================================================================*/
@@ -39,22 +13,32 @@
 ;*    The module                                                       */
 ;*---------------------------------------------------------------------*/
 (module cgen_emit-cop
-   (import type_type
-	   type_tools
-	   type_cache
-	   engine_param
-	   ast_var
-	   ast_node
-	   cgen_emit
-	   cgen_cop)
-   (export (generic emit-cop::bool ::cop)
-	   (untrigraph::bstring ::bstring)))
+   (include "Tools/location.sch"
+	    "Tools/fprint.sch"
+	    "Tools/trace.sch")
+   (import  type_type
+	    type_tools
+	    type_cache
+	    engine_param
+	    ast_var
+	    ast_node
+	    ast_env
+	    backend_c_emit
+	    cgen_cop
+	    (*module-location* module_module))
+   (export  (generic emit-cop::bool ::cop)
+	    (reset-bdb-loc!)
+	    (emit-bdb-loc ::obj)
+	    (get-current-bdb-loc)))
 
 ;*---------------------------------------------------------------------*/
 ;*    emit-cop ...                                                     */
 ;*    -------------------------------------------------------------    */
 ;*    If emit-cop emit an expression with a `;' it returns #f,         */
 ;*    otherwise it returns #t.                                         */
+;*    -------------------------------------------------------------    */
+;*    The general idea of that printer is that no specific printer     */
+;*    ever emit \n because they are emitted by the location printer.   */
 ;*---------------------------------------------------------------------*/
 (define-generic (emit-cop::bool cop::cop))
  
@@ -62,44 +46,53 @@
 ;*    emit-cop ::clabel ...                                            */
 ;*---------------------------------------------------------------------*/
 (define-method (emit-cop cop::clabel)
-   (with-access::clabel cop (used? name body)
+   (with-access::clabel cop (used? name body loc)
       (if used?
 	  (begin
+	     (emit-bdb-loc loc)
 	     (display name *c-port*)
-	     (write-char #\: *c-port*)
-	     (newline *c-port*)))
+	     (write-char #\: *c-port*)))
       (emit-cop body)))
 
 ;*---------------------------------------------------------------------*/
 ;*    emit-cop ::cgoto ...                                             */
 ;*---------------------------------------------------------------------*/
 (define-method (emit-cop cop::cgoto)
-   (with-access::cgoto cop (label)
-      (fprint *c-port* "goto " (clabel-name label) #\;)
+   (with-access::cgoto cop (label loc)
+      (emit-bdb-loc loc)
+      (fprin *c-port* "goto " (clabel-name label) #\;)
       #f))
 
 ;*---------------------------------------------------------------------*/
 ;*    emit-cop ::block ...                                             */
 ;*---------------------------------------------------------------------*/
 (define-method (emit-cop cop::block)
-   (with-access::block cop (body)
+   (with-access::block cop (body loc)
+      (emit-bdb-loc loc)
       (if (block? body)
 	  (emit-cop body)
 	  (begin
-	     (write-char #\{ *c-port*)
-	     (newline *c-port*)
-	     (if (emit-cop body) (fprint *c-port* #\;))
-	     (write-char #\} *c-port*)
-	     (newline *c-port*)
+	     (display #"{ " *c-port*)
+	     (trace cgen (display "/* cop-block */" *c-port*))
+	     (emit-bdb-loc-comment loc)
+	     (if (emit-cop body)
+		 (begin
+		    (emit-bdb-loc (get-current-bdb-loc))
+		    (display "; " *c-port*)
+		    (trace cgen (display "/* cop-block */" *c-port*))))
+	     (display "} " *c-port*)
+	     (trace cgen (display "/* cop-block */" *c-port*))
 	     #f))))
 
 ;*---------------------------------------------------------------------*/
 ;*    emit-cop ::creturn ...                                           */
 ;*---------------------------------------------------------------------*/
 (define-method (emit-cop cop::creturn)
-   (with-access::creturn cop (value)
+   (with-access::creturn cop (value loc)
+      (emit-bdb-loc loc)
       (display "return " *c-port*)
-      (if (emit-cop value) (fprint *c-port* #\;))
+      (if (emit-cop value)
+	  (write-char #\; *c-port*))
       #f))
 
 ;*---------------------------------------------------------------------*/
@@ -129,32 +122,38 @@
 ;*    emit-cop ::cpragma ...                                           */
 ;*---------------------------------------------------------------------*/
 (define-method (emit-cop cop::cpragma)
-   (with-access::cpragma cop (args format)
+   (with-access::cpragma cop (args format loc)
+      (emit-bdb-loc loc)
       (if (null? args)
 	  (display format *c-port*)
 	  (let* ((sport  (open-input-string format))
 		 (args   (list->vector args))
 		 (parser (regular-grammar ()
-			    ((#\$ (+ (>-< #\0 #\9)))
+			    ((: #\$ (+ (in (#\0 #\9))))
 			     (let* ((str   (the-string))
 				    (len   (the-length))
 				    (index (string->number
 					    (substring str 1 len))))
 				(emit-cop (vector-ref args (-fx index 1)))
 				(ignore)))
+			    ("$$"
+			     (display "$" *c-port*)
+			     (ignore))
 			    ((+ (out #\$))
 			     (display (the-string) *c-port*)
 			     (ignore))
 			    (else
-			     (the-failing-char)))))
+			     (the-failure)))))
 	     (read/rp parser sport)
+	     (close-input-port sport)
 	     #t))))
 
 ;*---------------------------------------------------------------------*/
 ;*    emit-cop ::ccast ...                                             */
 ;*---------------------------------------------------------------------*/
 (define-method (emit-cop cop::ccast)
-   (with-access::ccast cop (arg type)
+   (with-access::ccast cop (arg type loc)
+      (emit-bdb-loc loc)
       (display "((" *c-port*)
       (display (type-name type) *c-port*)
       (write-char #\) *c-port*)
@@ -166,34 +165,41 @@
 ;*    emit-cop ::csequence ...                                         */
 ;*---------------------------------------------------------------------*/
 (define-method (emit-cop cop::csequence)
-   (with-access::csequence cop (c-exp? cops)
+   (with-access::csequence cop (c-exp? cops loc)
       (if c-exp?
 	  (begin
 	     (if (null? cops)
 		 (emit-atom-value #unspecified)
 		 (begin
-		    (write-char #\( *c-port*)
+		    (display "( " *c-port*)
+		    (trace cgen (display "/* cop-csequence */" *c-port*))
 		    (let liip ((exp cops))
 		       (if (null? (cdr exp))
 			   (begin
 			      (emit-cop (car exp))
-			      (write-char #\) *c-port*)
+			      (display ") " *c-port*)
+			      (trace cgen "/* cop-csequence */" *c-port*)
 			      #t)
 			   (begin
 			      (emit-cop (car exp))
 			      (if (cfail? (car exp))
 				  (begin
-				     (write-char #\) *c-port*)
+				     (display ") " *c-port*)
+				     (trace cgen  "/* cop-csequence */" *c-port*)
 				     #t)
 				  (begin
-				     (write-char #\, *c-port*)
-				     (newline *c-port*)
+				     (display ", " *c-port*)
+				     (trace cgen "/* cop-csequence */" *c-port*)
 				     (liip (cdr exp))))))))))
 	  (let liip ((exp cops))
 	     (if (null? exp)
 		 #f
 		 (let ((e (car exp)))
-		    (if (emit-cop e) (fprint *c-port* #\;))
+		    (if (emit-cop e)
+			(begin
+			   (display "; " *c-port*)
+			   (trace cgen
+				  (display "/* cop-csequence */" *c-port*))))
 		    (if (cfail? e)
 			(liip '())
 			(liip (cdr exp)))))))))
@@ -202,22 +208,31 @@
 ;*    emit-cop ::nop ...                                               */
 ;*---------------------------------------------------------------------*/
 (define-method (emit-cop cop::nop)
-   (fprint *c-port* #\;)
-   #f)
+   (with-access::nop cop (loc)
+      (display "; " *c-port*)
+      (trace cgen (display "/* cop-nop */" *c-port*))
+      #f))
 
 ;*---------------------------------------------------------------------*/
 ;*    emit-cop ::stop ...                                              */
 ;*---------------------------------------------------------------------*/
 (define-method (emit-cop cop::stop)
-   (with-access::stop cop (value)
-      (if (emit-cop value) (fprint *c-port* #\;))
+   (with-access::stop cop (value loc)
+      (if (emit-cop value)
+	  ;; we don't have to emit a location here because the
+	  ;; location is held by the printed value
+	  (begin
+	     (display "; " *c-port*)
+	     (trace cgen (display "/* cop-stop */" *c-port*))))
       #f))
 
 ;*---------------------------------------------------------------------*/
 ;*    emit-cop ::csetq ...                                             */
 ;*---------------------------------------------------------------------*/
 (define-method (emit-cop cop::csetq)
-   (with-access::csetq cop (var value)
+   (with-access::csetq cop (var value loc)
+      ;; we first emit a location for this node
+      (emit-bdb-loc loc)
       (emit-cop var)
       ;; don't omit to put space sourrounding `=' otherwise
       ;; it could become an ambiguous assignement (e.g. x=-1).
@@ -229,7 +244,8 @@
 ;*    emit-cop ::cif ...                                               */
 ;*---------------------------------------------------------------------*/
 (define-method (emit-cop cop::cif)
-   (with-access::cif cop (test true false)
+   (with-access::cif cop (test true false loc)
+      (emit-bdb-loc loc)
       (display "if(" *c-port*)
       (emit-cop test)
       (write-char #\) *c-port*)
@@ -237,29 +253,40 @@
       (display " else " *c-port*)
       (emit-cop false)))
 
-;*---------------------------------------------------------------------*/
-;*    emit-cop ::branch ...                                            */
-;*---------------------------------------------------------------------*/
-(define-method (emit-cop cop::branch)
-   (with-access::branch cop (test true false end)
-      (emit-cop test)
-      (if (emit-cop true) (fprint *c-port* #\;))
-      (if (emit-cop false) (fprint *c-port* #\;))
-      (emit-cop end)))
-
 ;*---------------------------------------------------------------------*/
 ;*    emit-cop ::local ...                                             */
 ;*---------------------------------------------------------------------*/
 (define-method (emit-cop cop::local-var)
-   (with-access::local-var cop (vars)
+   (with-access::local-var cop (vars loc)
+      (emit-bdb-loc loc)
       (for-each (lambda (local)
-		   (fprint *c-port*
-			   (make-typed-declaration (local-type local)
-						   (local-name local))
-			   #\;))
+		   (fprin *c-port*
+			  (make-typed-declaration (local-type local)
+						  (local-name local))
+			  (if (and (>fx *bdb-debug* 0)
+				   (eq? (type-class (local-type local))
+					'bigloo))
+			      (string-append " = (("
+					     (make-typed-declaration
+					      (local-type local)
+					      "")
+					     ")BUNSPEC)")
+			      "")
+			  #\;))
 		vars)
       #f))
 
+;*---------------------------------------------------------------------*/
+;*    emit-cop ::bdb-block ...                                         */
+;*---------------------------------------------------------------------*/
+(define-method (emit-cop cop::bdb-block)
+   (with-access::bdb-block cop (loc body)
+      (emit-bdb-loc loc)
+      (fprin *c-port* "int bigloo_dummy_bdb; bigloo_dummy_bdb = 0; { ")
+      (emit-cop body)
+      (display "} " *c-port*)
+      (trace cgen (display " /* cop-bdb-block */" *c-port*))))
+   
 ;*---------------------------------------------------------------------*/
 ;*    emit-cop ::cfuncall ...                                          */
 ;*---------------------------------------------------------------------*/
@@ -333,7 +360,7 @@
 			    (loop (cdr actuals)))))))
 	    (emit-stdc-regular-cfuncall (cop)
 	       (begin
-		  (display "(VA_PROCUDUREP( " *c-port*)
+		  (display "(VA_PROCEDUREP( " *c-port*)
 		  (emit-cop (cfuncall-fun cop))
 		  (display " ) ? " *c-port*)
 		  (emit-regular-cfuncall/eoa cop)
@@ -341,6 +368,7 @@
 		  (emit-regular-cfuncall/oeoa cop)
 		  (display " )" *c-port*)
 		  #t)))
+      (emit-bdb-loc (cop-loc cop))
       (case (cfuncall-strength cop)
 	 ((elight)
 	  (emit-extra-light-cfuncall cop))
@@ -355,7 +383,8 @@
 ;*    emit-cop ::capply ...                                            */
 ;*---------------------------------------------------------------------*/
 (define-method (emit-cop cop::capply)
-   (with-access::capply cop (fun arg)
+   (with-access::capply cop (fun arg loc)
+      (emit-bdb-loc loc)
       (display "apply(" *c-port*)
       (emit-cop fun)
       (display ", " *c-port*)
@@ -367,71 +396,112 @@
 ;*    emit-cop ::capp ...                                              */
 ;*---------------------------------------------------------------------*/
 (define-method (emit-cop cop::capp)
-   (labels ((emit-infix-capp ()
-	       (let ((actuals (capp-args cop)))
-		  (write-char #\( *c-port*)
-		  (emit-cop (car actuals))
-		  (emit-cop (capp-fun cop))
-		  (emit-cop (cadr actuals))
-		  (write-char #\) *c-port*)
-		  #t))
-	    (emit-prefix-capp ()
-               (let ((actuals (capp-args cop)))
-		  (emit-cop (capp-fun cop))
-		  (write-char #\( *c-port*)
-		  (if (null? actuals)
-		      (begin
-			 (write-char #\) *c-port*)
-			 #t)
-		      (let loop ((actuals actuals))
-			 (if (null? (cdr actuals))
-			     (begin
-				(emit-cop (car actuals))
-				(write-char #\) *c-port*)
-				#t)
-			     (begin
-				(emit-cop (car actuals))
-				(display ", " *c-port*)
-				(loop (cdr actuals)))))))))
-      (let ((fun (varc-variable (capp-fun cop))))
-	 (if (and (cfun? (global-value fun)) (cfun-infix? (global-value fun)))
-	     (emit-infix-capp)
-	     (emit-prefix-capp)))))
+   (define (emit-infix-capp)
+      (let ((actuals (capp-args cop)))
+	 (write-char #\( *c-port*)
+	 (if (null? actuals)
+	     (emit-cop (capp-fun cop))
+	     (begin
+		(emit-cop (car actuals))
+		(emit-cop (capp-fun cop))
+		(emit-cop (cadr actuals))))
+	 (write-char #\) *c-port*)
+	 #t))
+   (define (emit-prefix-capp)
+      (let ((actuals (capp-args cop)))
+	 (emit-cop (capp-fun cop))
+	 (write-char #\( *c-port*)
+	 (if (null? actuals)
+	     (begin
+		(write-char #\) *c-port*)
+		#t)
+	     (let loop ((actuals actuals))
+		(if (null? (cdr actuals))
+		    (begin
+		       (emit-cop (car actuals))
+		       (write-char #\) *c-port*)
+		       #t)
+		    (begin
+		       (emit-cop (car actuals))
+		       (display ", " *c-port*)
+		       (loop (cdr actuals))))))))
+   (define (emit-prefix-capp-sans-bdb-loc)
+      (let ((o *bdb-debug-no-line-directives?*))
+	 (set! *bdb-debug-no-line-directives?* #t)
+	 (emit-prefix-capp)
+	 (set! *bdb-debug-no-line-directives?* o)))
+   (let ((fun (varc-variable (capp-fun cop)))
+	 (loc (capp-loc cop)))
+      (emit-bdb-loc loc)
+      (cond
+	 ((and (cfun? (global-value fun)) (cfun-infix? (global-value fun)))
+	  (emit-infix-capp))
+	 ((and (cfun? (global-value fun)) (cfun-macro? (global-value fun)))
+	  (emit-prefix-capp-sans-bdb-loc))
+	 (else
+	  (emit-prefix-capp)))))
+
+;*---------------------------------------------------------------------*/
+;*    *bfalse*                                                         */
+;*    -------------------------------------------------------------    */
+;*    A local cache for the C false macro.                             */
+;*---------------------------------------------------------------------*/
+(define *bfalse* #f)
 
 ;*---------------------------------------------------------------------*/
 ;*    emit-cop ::cfail ...                                             */
 ;*---------------------------------------------------------------------*/
 (define-method (emit-cop cop::cfail)
-   (with-access::cfail cop (proc msg obj)
-      (display "FAILURE(" *c-port*)
-      (emit-cop proc)
-      (write-char #\, *c-port*)
-      (emit-cop msg)
-      (write-char #\, *c-port*)
-      (emit-cop obj)
-      (display ");" *c-port*)
+   (with-access::cfail cop (proc msg obj loc)
+      (emit-bdb-loc loc)
+      (if (not *bfalse*)
+	  (set! *bfalse* (find-global/module 'bfalse 'foreign)))
+      (cond
+	 ((and (varc? proc) (eq? (varc-variable proc) *bfalse*)
+	       (varc? msg) (eq? (varc-variable msg) *bfalse*)
+	       (varc? obj) (eq? (varc-variable obj) *bfalse*))
+	  (display "exit( -1 );" *c-port*))
+	 ((<=fx *bdb-debug* 0)
+	  (display "FAILURE(" *c-port*)
+	  (emit-cop proc)
+	  (write-char #\, *c-port*)
+	  (emit-cop msg)
+	  (write-char #\, *c-port*)
+	  (emit-cop obj)
+	  (display ");" *c-port*))
+	 (else
+	  (display "the_failure(" *c-port*)
+	  (emit-cop proc)
+	  (write-char #\, *c-port*)
+	  (emit-cop msg)
+	  (write-char #\, *c-port*)
+	  (emit-cop obj)
+	  (display "), exit( -1 );" *c-port*)))
       #f))
 
 ;*---------------------------------------------------------------------*/
 ;*    emit-cop ::cswitch ...                                           */
 ;*---------------------------------------------------------------------*/
 (define-method (emit-cop cop::cswitch)
-   (with-access::cswitch cop (test clauses)
-      (display "switch " *c-port*)
-      (write-char #\( *c-port*)
+   (with-access::cswitch cop (test clauses loc)
+      (emit-bdb-loc loc)
+      (display "switch( " *c-port*)
       (emit-cop test)
-      (write-char #\) *c-port*)
-      (write-char #\{ *c-port*)
-      (newline *c-port*)
+      (display ") { " *c-port*)
+      (trace cgen (display "/* cop-cswitch */" *c-port*))
       (let loop ((clauses clauses))
 	 (let ((clause (car clauses)))
 	    (if (eq? (car clause) 'else)
-		(begin
+		(let ((loc (cop-loc (cdr clause))))
+		   (emit-bdb-loc loc)
 		   (display "default: " *c-port*)
-		   (newline *c-port*)
-		   (if (emit-cop (cdr clause)) (fprint *c-port* #\;))
-		   (write-char #\} *c-port*)
-		   (newline *c-port*)
+		   (if (emit-cop (cdr clause))
+		       (begin
+			  (display "; " *c-port*)
+			  (trace cgen
+				 (display "/* cswitch default */" *c-port*))))
+		   (display "} " *c-port*)
+		   (trace cgen (display "/* cswitch */" *c-port*))
 		   #f)
 		(begin
 		   (for-each (lambda (t)
@@ -440,15 +510,20 @@
 				(display " : " *c-port*)
 				(newline *c-port*))
 			     (car clause))
-		   (if (emit-cop (cdr clause)) (fprint *c-port* #\;))
-		   (fprint *c-port* "break;")
+		   (if (emit-cop (cdr clause))
+		       (begin
+			  (display "; " *c-port*)
+			  (trace cgen
+				 (display "/* cswitch clause */" *c-port*))))
+		   (display "break;" *c-port*)
 		   (loop (cdr clauses))))))))
 
 ;*---------------------------------------------------------------------*/
 ;*    emit-cop ::cmake-box ...                                         */
 ;*---------------------------------------------------------------------*/
 (define-method (emit-cop cop::cmake-box)
-   (with-access::cmake-box cop (value)
+   (with-access::cmake-box cop (value loc)
+      (emit-bdb-loc loc)
       (display "MAKE_CELL(" *c-port*)
       (emit-cop value)
       (write-char #\) *c-port*)
@@ -458,7 +533,8 @@
 ;*    emit-cop ::cbox-ref ...                                          */
 ;*---------------------------------------------------------------------*/
 (define-method (emit-cop cop::cbox-ref)
-   (with-access::cbox-ref cop (var)
+   (with-access::cbox-ref cop (var loc)
+      (emit-bdb-loc loc)
       (display "CELL_REF(" *c-port*)
       (emit-cop var)
       (write-char #\) *c-port*)
@@ -468,7 +544,8 @@
 ;*    emit-cop ::cbox-set! ...                                         */
 ;*---------------------------------------------------------------------*/
 (define-method (emit-cop cop::cbox-set!)
-   (with-access::cbox-set! cop (var value)
+   (with-access::cbox-set! cop (var value loc)
+      (emit-bdb-loc loc)
       (display "CELL_SET(" *c-port*)
       (emit-cop var)
       (display ", " *c-port*)
@@ -480,23 +557,29 @@
 ;*    emit-cop ::cset-ex-it ...                                        */
 ;*---------------------------------------------------------------------*/
 (define-method (emit-cop cop::cset-ex-it)
-   (with-access::cset-ex-it cop (exit jump-value body)
+   (with-access::cset-ex-it cop (exit jump-value body loc)
+      (emit-bdb-loc loc)
+      (display "{BGL_STORE_TRACE(); " *c-port*)
       (display "if( SET_EXIT(" *c-port*)
       (emit-cop exit)
-      (display ") ) {" *c-port*)
-      (display "RESTORE_TRACE(); " *c-port*)
+      (display ") ) { " *c-port*)
+      (trace cgen (display "/* cop-cset-ex-it */" *c-port*))
+      (display "BGL_RESTORE_TRACE(); " *c-port*)
       (emit-cop jump-value)
+      (emit-bdb-loc loc)
       (display "} else {" *c-port*)
       (emit-cop body)
-      (write-char #\} *c-port*)
-      (newline *c-port*)
+      (emit-bdb-loc loc)
+      (display "}} " *c-port*)
+      (trace cgen (display "/* cop-cset-ex-it */" *c-port*))
       #f))
 
 ;*---------------------------------------------------------------------*/
 ;*    emit-cop ::cjump-ex-it ...                                       */
 ;*---------------------------------------------------------------------*/
 (define-method (emit-cop cop::cjump-ex-it)
-   (with-access::cjump-ex-it cop (exit value)
+   (with-access::cjump-ex-it cop (exit value loc)
+      (emit-bdb-loc loc)
       (display "JUMP_EXIT( " *c-port*)
       (emit-cop exit)
       (write-char #\, *c-port*)
@@ -505,141 +588,66 @@
       #t))
 
 ;*---------------------------------------------------------------------*/
-;*    emit-atom-value ...                                              */
-;*---------------------------------------------------------------------*/
-(define (emit-atom-value value)
-   (cond
-      ((boolean? value)
-       (display "((" *c-port*)
-       (display (string-sans-$ (type-name *bool*)) *c-port*)
-       (display #\) *c-port*)
-       (display (if value 1 0) *c-port*)
-       (display #\) *c-port*))
-      ((null? value)
-       (display "BNIL" *c-port*))
-      ((char? value)
-       (display "((" *c-port*)
-       (display (string-sans-$ (type-name *char*)) *c-port*)
-       (display ")" *c-port*)
-       (if (=fx (char->integer value) 0)
-	   (display "'\\000'" *c-port*)
-	   (begin
-	      (write-char #\' *c-port*)
-	      (if (=fx (char->integer value) 39)
-		  (display "\\''" *c-port*)
-		  (begin
-		     (case value
-			((#\return)
-			 (write-char #\\ *c-port*)
-			 (write-char #\r *c-port*))
-			((#\tab)
-			 (write-char #\\ *c-port*)
-			 (write-char #\t *c-port*))
-			((#\newline)
-			 (write-char #\\ *c-port*)
-			 (write-char #\n *c-port*))
-			((#\\)
-			 (write-char #\\ *c-port*)
-			 (write-char #\\ *c-port*))
-			(else
-			 (write-char value *c-port*)))
-		     (write-char #\' *c-port*)))))
-       (write-char #\) *c-port*))
-      ((eq? value #unspecified)
-       (display "BUNSPEC" *c-port*))
-      ((cnst? value)
-       (display "BCNST(" *c-port*)
-       (display (cnst->integer value) *c-port*)
-       (display #\) *c-port*))
-      ((string? value)
-       (display #\" *c-port*)
-       (display (untrigraph (string-for-read value)) *c-port*)
-       (display #\" *c-port*))
-      ((fixnum? value)
-       (display "((" *c-port*)
-       (display (string-sans-$ (type-name *long*)) *c-port*)
-       (display ")" *c-port*)
-       (display value *c-port*)
-       (display ")" *c-port*))
-      ((flonum? value)
-       (display "((" *c-port*)
-       (display (string-sans-$ (type-name *real*)) *c-port*)
-       (display ")" *c-port*)
-       (display value *c-port*)
-       (display ")" *c-port*))
-      (else
-       (display value *c-port*))))
-
-;*---------------------------------------------------------------------*/
-;*    untrigraph ...                                                   */
+;*    *bdb-loc* ...                                                    */
 ;*    -------------------------------------------------------------    */
-;*    We remove ?? and replace it by \077\077\--- (the octal ascii     */
-;*    code of ?) in order to avoir C trigraph confusions.              */
+;*    The current bdb source location information.                     */
 ;*---------------------------------------------------------------------*/
-(define (untrigraph from)
-   (let* ((len   (string-length from))
-	  (len-3 (-fx len 3)))
-      ;; first we count how many collision we have
-      (let ((nb-col (let loop ((i      0)
-			       (nb-col 0))
-		       (cond
-			  ((>fx i len-3)
-			   nb-col)
-			  ((not (char=? (string-ref from i) #\?))
-			   (loop (+fx i 1) nb-col))
-			  ((not (char=? (string-ref from (+fx i 1)) #\?))
-			   (loop (+fx i 2) nb-col))
-			  ;; yes, we have one
-			  (else
-			   (loop (+fx i 3) (+fx nb-col 1)))))))
-	 (if (=fx nb-col 0)
-	     ;; there is no trigraph clashes
-	     from
-	     ;; there is some, we allocate a new string. Each trigraph
-	     ;; require 4 times its size.
-	     (let ((res   (make-string (+fx len (*fx 3 (*fx nb-col 3)))))
-		   (len-1 (-fx len 1)))
-		(let loop ((r 0)
-			   (w 0))
-		   (cond
-		      ((=fx r len)
-		       res)
-		      ((or (not (char=? (string-ref from r) #\?))
-			   (>fx r len-3))
-		       (string-set! res w (string-ref from r))
-		       (loop (+fx r 1) (+fx w 1)))
-		      ((not (char=? (string-ref from (+fx r 1)) #\?))
-		       (string-set! res w #\?)
-		       (string-set! res (+fx w 1) (string-ref from (+fx r 1)))
-		       (loop (+fx r 2) (+fx w 2)))
-		      (else
-		       ;; this is a trigraph
-		       (string-set! res w #\\)
-		       (string-set! res (+fx w 1) #\0)
-		       (string-set! res (+fx w 2) #\7)
-		       (string-set! res (+fx w 3) #\7)
-		       (string-set! res (+fx w 4) #\\)
-		       (string-set! res (+fx w 5) #\0)
-		       (string-set! res (+fx w 6) #\7)
-		       (string-set! res (+fx w 7) #\7)
-		       (string-set! res (+fx w 8) #\\)
-		       (let ((code (integer->string
-				    (char->integer (string-ref from (+fx r 2)))
-				    8)))
-			  (if (=fx (string-length code) 4)
-			      (begin
-				 (string-set! res (+fx w 9) #\0)
-				 (string-set! res (+fx w 10)
-					      (string-ref code 2))
-				 (string-set! res (+fx w 11)
-					      (string-ref code 3)))
-			      (begin
-				 (string-set! res (+fx w 9)
-					      (string-ref code 2))
-				 (string-set! res (+fx w 10)
-					      (string-ref code 3))
-				 (string-set! res (+fx w 11)
-					      (string-ref code 4)))))
-		       (loop (+fx r 3) (+fx w 12))))))))))
+(define *bdb-loc* #unspecified)
 
+;*---------------------------------------------------------------------*/
+;*    reset-bdb-loc! ...                                               */
+;*---------------------------------------------------------------------*/
+(define (reset-bdb-loc!)
+   (set! *bdb-loc* #unspecified))
 
+;*---------------------------------------------------------------------*/
+;*    get-current-bdb-loc ...                                          */
+;*---------------------------------------------------------------------*/
+(define (get-current-bdb-loc)
+   *bdb-loc*)
+
+;*---------------------------------------------------------------------*/
+;*    emit-bdb-loc ...                                                 */
+;*    -------------------------------------------------------------    */
+;*    This function emits a bdb location information (that is a        */
+;*    C # line information.) This function emits this information      */
+;*    only if we are dumping the C code for a different Scheme source  */
+;*    code line that the previous line dump.                           */
+;*---------------------------------------------------------------------*/
+(define (emit-bdb-loc cur-loc)
+   (cond
+      ((or (not *c-debug-lines-info*) *bdb-debug-no-line-directives?*)
+       (newline *c-port*))
+      ((not (location? cur-loc))
+       (cond
+	  ((location? *bdb-loc*)
+	   (emit-bdb-loc *bdb-loc*))
+	  ((location? *module-location*)
+	   ;; when no location is found, we use the location of
+	   ;; the module clause
+	   (emit-bdb-loc *module-location*))
+	  ((pair? *src-files*)
+	   ;; when no location at all is found, emit a dummy
+	   ;; "first file line" location
+	   (fprint *c-port* #"\n#line " 1 " \"" (car *src-files*) #\"))))
+      (else
+       (let ((cur-fname (location-fname cur-loc))
+	     (cur-line  (location-lnum cur-loc)))
+	  (when (and (integer? cur-line) (string? cur-fname))
+	     (fprint *c-port* #"\n#line " cur-line " \"" cur-fname #\"))
+	  (set! *bdb-loc* cur-loc)))))
+
+;*---------------------------------------------------------------------*/
+;*    emit-bdb-loc-comment ...                                         */
+;*    -------------------------------------------------------------    */
+;*    For debug purposes this function write a location is a C         */
+;*    comment (iff CUR-LOC is a location).                             */
+;*---------------------------------------------------------------------*/
+(define (emit-bdb-loc-comment cur-loc)
+   (if (location? cur-loc)
+       (begin
+	  (display "/* " *c-port*)
+	  (display (location-fname cur-loc) *c-port*)
+	  (display " " *c-port*)
+	  (display (location-lnum cur-loc) *c-port*)
+	  (display " */" *c-port*))))
