@@ -1,36 +1,10 @@
-;* --------------------------------------------------------------------*/
-;*    Copyright (c) 1992-1998 by Manuel Serrano. All rights reserved.  */
-;*                                                                     */
-;*                                     ,--^,                           */
-;*                               _ ___/ /|/                            */
-;*                           ,;'( )__, ) '                             */
-;*                          ;;  //   L__.                              */
-;*                          '   \   /  '                               */
-;*                               ^   ^                                 */
-;*                                                                     */
-;*                                                                     */
-;*    This program is distributed in the hope that it will be useful.  */
-;*    Use and copying of this software and preparation of derivative   */
-;*    works based upon this software are permitted, so long as the     */
-;*    following conditions are met:                                    */
-;*           o credit to the authors is acknowledged following         */
-;*             current academic behaviour                              */
-;*           o no fees or compensation are charged for use, copies,    */
-;*             or access to this software                              */
-;*           o this copyright notice is included intact.               */
-;*      This software is made available AS IS, and no warranty is made */
-;*      about the software or its performance.                         */
-;*                                                                     */
-;*      Bug descriptions, use reports, comments or suggestions are     */
-;*      welcome. Send them to                                          */
-;*        Manuel Serrano -- Manuel.Serrano@unice.fr                    */
-;*-------------------------------------------------------------------- */
 ;*=====================================================================*/
-;*    serrano/prgm/project/bigloo/comptime1.9b/Cnst/alloc.scm          */
+;*    serrano/prgm/project/bigloo/comptime/Cnst/alloc.scm              */
 ;*    -------------------------------------------------------------    */
 ;*    Author      :  Manuel Serrano                                    */
 ;*    Creation    :  Mon Feb  6 13:51:36 1995                          */
-;*    Last change :  Tue Apr  8 13:32:58 1997 (serrano)                */
+;*    Last change :  Mon Jun 19 08:11:18 2006 (serrano)                */
+;*    Copyright   :  1995-2006 Manuel Serrano, see LICENSE file        */
 ;*    -------------------------------------------------------------    */
 ;*    The constant allocations.                                        */
 ;*=====================================================================*/
@@ -42,6 +16,7 @@
    (include "Tvector/tvector.sch")
    (import  engine_param
 	    module_module
+	    backend_backend
 	    tools_shape
 	    tools_error
 	    type_type
@@ -55,7 +30,8 @@
 	    ast_glo-def
 	    coerce_coerce
 	    cnst_cache
-	    cnst_node)
+	    cnst_node
+	    ast_ident)
    (export  (start-cnst-alloc!)
 	    (stop-cnst-alloc!)
 	    (get-cnst-offset)
@@ -63,14 +39,18 @@
 	    (get-cnst-sexp)
 	    (get-cnst-table)
 	    (cnst-table-id)
-	    (cnst-alloc-string::node    ::bstring   <loc>)
-	    (cnst-alloc-symbol::node    ::symbol    <loc>)
-	    (cnst-alloc-keyword::node   ::keyword   <loc>)
-	    (cnst-alloc-procedure::node ::node      <loc>)
-	    (cnst-alloc-real::node      ::real      <loc>)
-	    (cnst-alloc-list::node      <list>      <loc>)
-	    (cnst-alloc-vector::node    ::vector    <loc>)
-	    (cnst-alloc-tvector::node   <a-tvector> <loc>)))
+	    (cnst-alloc-string::node ::bstring <loc>)
+	    (cnst-alloc-ucs2-string::node ::ucs2string <loc>)
+	    (cnst-alloc-symbol::node ::symbol <loc>)
+	    (cnst-alloc-keyword::node ::keyword <loc>)
+	    (cnst-alloc-procedure::node ::node <loc>)
+	    (cnst-alloc-real::node ::real <loc>)
+	    (cnst-alloc-elong::node ::elong <loc>)
+	    (cnst-alloc-llong::node ::llong <loc>)
+	    (cnst-alloc-list::node <list> <loc>)
+	    (cnst-alloc-vector::node ::vector <loc>)
+	    (cnst-alloc-struct::node ::struct <loc>)
+	    (cnst-alloc-tvector::node <a-tvector> <loc>)))
 
 ;*---------------------------------------------------------------------*/
 ;*    structures                                                       */
@@ -125,14 +105,18 @@
 ;*---------------------------------------------------------------------*/
 ;*    The local hash-tables                                            */
 ;*---------------------------------------------------------------------*/
-(define *string-env*    '())
-(define *real-env*      '())
-(define *symbol-env*    '())
-(define *keyword-env*   '())
-(define *list-env*      '())
-(define *vector-env*    '())
-(define *global-set*    '())
-(define *global-sexp*   '())
+(define *string-env*     '())
+(define *ucs2string-env* '())
+(define *real-env*       '())
+(define *elong-env*      '())
+(define *llong-env*      '())
+(define *symbol-env*     '())
+(define *keyword-env*    '())
+(define *list-env*       '())
+(define *vector-env*     '())
+(define *struct-env*     '())
+(define *global-set*     '())
+(define *global-sexp*    '())
 
 ;*---------------------------------------------------------------------*/
 ;*    *old-debug*                                                      */
@@ -143,13 +127,13 @@
 ;*    cnst-table-id ...                                                */
 ;*---------------------------------------------------------------------*/
 (define (cnst-table-id)
-   '__CNSTS_TABLE)
+   '__cnsts_table)
 
 ;*---------------------------------------------------------------------*/
 ;*    typed-cnst-table-id ...                                          */
 ;*---------------------------------------------------------------------*/
 (define (typed-cnst-table-id)
-   (symbol-append (cnst-table-id) '::cnst*))
+   (make-typed-ident (cnst-table-id) 'cnst*))
 
 ;*---------------------------------------------------------------------*/
 ;*    start-cnst-alloc! ...                                            */
@@ -164,24 +148,18 @@
 			 *module*
 			 'cnst-vector
 			 'now))
+   ;; when compiling for bdb2, default global variable are exported. of
+   ;; course this may lead to incorrect compilation if the global variable
+   ;; (its name is not gensymed nor qualified) is exported.
+   (if (and (>=fx *bdb-debug* 3)
+	    (memq 'bdb (backend-debug-support (the-backend))))
+       (global-import-set! *cnst-table* 'static))
    (set! *cnst-offset* 0)
-   (set! *string-env* (make-hash-table 1024
-				       (lambda (s) (string->0..2^x-1 s 10))
-				       cnst-info-cnst
-				       string=?))
+   (set! *string-env* (make-hashtable))
+   (set! *ucs2string-env* (make-hashtable))
    (set! *real-env*   '())
-   (set! *symbol-env* (make-hash-table 1024
-				       (lambda (s) (string->0..2^x-1
-						    (symbol->string s)
-						    10))
-				       cnst-info-cnst
-				       eq?))
-   (set! *keyword-env* (make-hash-table 1024
-					(lambda (s) (string->0..2^x-1
-						     (keyword->string s)
-						     10))
-					cnst-info-cnst
-					eq?))
+   (set! *symbol-env* (make-hashtable))
+   (set! *keyword-env* (make-hashtable))
    #t)
 
 ;*---------------------------------------------------------------------*/
@@ -189,12 +167,14 @@
 ;*---------------------------------------------------------------------*/
 (define (stop-cnst-alloc!)
    (set! *compiler-debug* *old-debug*)
-   (set! *string-env*  #unspecified)
-   (set! *real-env*    #unspecified) 
-   (set! *symbol-env*  #unspecified)
-   (set! *keyword-env* #unspecified)
-   (set! *list-env*    #unspecified) 
-   (set! *vector-env*  #unspecified) 
+   (set! *string-env*     #unspecified)
+   (set! *ucs2string-env* #unspecified)
+   (set! *real-env*       #unspecified) 
+   (set! *symbol-env*     #unspecified)
+   (set! *keyword-env*    #unspecified)
+   (set! *list-env*       #unspecified) 
+   (set! *vector-env*     #unspecified)
+   (set! *struct-env*     #unspecified)
    #t)
 
 ;*---------------------------------------------------------------------*/
@@ -210,30 +190,32 @@
 	      (variable *cnst-table-ref*)))
       (args (list (instantiate::atom
 		     (loc loc)
-		     (type (get-default-type))
+		     (type *int*)
 		     (value offset))))))
-	 
+
 ;*---------------------------------------------------------------------*/
 ;*    cnst-alloc-string ...                                            */
 ;*    -------------------------------------------------------------    */
 ;*    Strings can't be allocated via the reader because we use         */
 ;*    the string constant compilation to create the                    */
-;*    input-string-port !                                              */
+;*    input-string-port!                                               */
 ;*---------------------------------------------------------------------*/
 (define (cnst-alloc-string string loc)
    (define (alloc-string)
       ;; in lib-mode string are statically allocated
-      (let ((var (def-global-scnst! (symbol-append (gensym 'string) '::bstring)
+      (let ((var (def-global-scnst! (make-typed-ident (gensym 'string)
+						      'bstring)
 		    *module*
 		    string
-		    'sstring)))
+		    'sstring
+		    loc)))
 	 (if *shared-cnst?*
-	     (put-hash! (cnst-info string var) *string-env*))
+	     (hashtable-put! *string-env* string (cnst-info string var)))
 	 (instantiate::var
 	    (loc loc)
 	    (type (variable-type var))
 	    (variable var))))
-   (let ((old (and *shared-cnst?* (get-hash string *string-env*))))
+   (let ((old (and *shared-cnst?* (hashtable-get *string-env* string))))
       (cond
 	 (old
 	  (instantiate::var
@@ -243,16 +225,68 @@
 	 (else
 	  (alloc-string)))))
 
+;*---------------------------------------------------------------------*/
+;*    cnst-alloc-ucs2-string ...                                       */
+;*---------------------------------------------------------------------*/
+(define (cnst-alloc-ucs2-string string loc)
+   (define (lib-alloc-ucs2string bstring)
+      (let ((var (def-global-svar! (make-typed-ident (gensym 'ucs2string)
+						     'ucs2string)
+		    *module*
+		    'an-ucs2-string
+		    'now))
+	    (vs (cnst-alloc-string bstring loc)))
+	 (hashtable-put! *ucs2string-env* string (cnst-info string var))
+	 (add-cnst-sexp! `(set! (@ ,(global-id var) ,(global-module var))
+				,(coerce!
+				  (instantiate::app
+				     (loc loc)
+				     (type *ucs2string*)
+				     (fun (instantiate::var
+					     (loc loc)
+					     (type *ucs2string*)
+					     (variable *string->ucs2string*)))
+				     (args (list vs)))
+				  var
+				  *obj*)))
+	 (instantiate::var
+	    (loc loc)
+	    (type (variable-type var))
+	    (variable var))))
+   (define (read-alloc-ucs2string)
+      (let ((offset *cnst-offset*))
+	 (set! *cnst-offset* (+fx 1 *cnst-offset*))
+	 (set! *global-set* (cons string *global-set*))
+	 (hashtable-put! *ucs2string-env* string (cnst-info string offset))
+	 (make-cnst-table-ref offset loc)))
+   (let ((old (hashtable-get *ucs2string-env* string))
+	 (string-as-bstring (ucs2-string->utf8-string string)))
+      (cond
+	 (old
+	  (if (eq? *init-mode* 'lib)
+	      (instantiate::var
+		 (loc loc)
+		 (type (get-default-type))
+		 (variable (cnst-info-offset old)))
+	      (make-cnst-table-ref (cnst-info-offset old) loc)))
+	 ((or (eq? *init-mode* 'lib)
+	      (=fx (string-length string-as-bstring) 0)
+	      (char=? (string-ref string-as-bstring 0) #\;))
+	  (lib-alloc-ucs2string string-as-bstring))
+	 (else
+	  (read-alloc-ucs2string)))))
+
 ;*---------------------------------------------------------------------*/
 ;*    cnst-alloc-symbol ...                                            */
 ;*---------------------------------------------------------------------*/
 (define (cnst-alloc-symbol symbol loc)
    (define (lib-alloc-symbol)
-      (let ((var (def-global-svar! (symbol-append (gensym 'symbol) '::symbol)
+      (let ((var (def-global-svar! (make-typed-ident (gensym 'symbol) 'symbol)
 		    *module*
 		    'a-symbol
-		    'now)))
-	 (put-hash! (cnst-info symbol var) *symbol-env*)
+		    'now))
+	    (vs (cnst-alloc-string (symbol->string symbol) loc)))
+	 (hashtable-put! *symbol-env* symbol (cnst-info symbol var))
 	 (add-cnst-sexp! `(set! (@ ,(global-id var) ,(global-module var))
 				,(coerce!
 				  (instantiate::app
@@ -261,12 +295,9 @@
 				     (fun (instantiate::var
 					     (loc loc)
 					     (type *symbol*)
-					     (variable *string->symbol*)))
-				     (args (list (instantiate::atom
-						    (loc loc)
-						    (type *string*)
-						    (value (symbol->string
-							    symbol))))))
+					     (variable *bstring->symbol*)))
+				     (args (list vs)))
+				  var
 				  *obj*)))
 	 (instantiate::var
 	    (loc loc)
@@ -276,9 +307,9 @@
       (let ((offset *cnst-offset*))
 	 (set! *cnst-offset* (+fx 1 *cnst-offset*))
 	 (set! *global-set* (cons symbol *global-set*))
-	 (put-hash! (cnst-info symbol offset) *symbol-env*)
+	 (hashtable-put! *symbol-env* symbol (cnst-info symbol offset))
 	 (make-cnst-table-ref offset loc)))
-   (let ((old (get-hash symbol *symbol-env*))
+   (let ((old (hashtable-get *symbol-env* symbol))
 	 (symbol-as-string (symbol->string symbol)))
       (cond
 	 (old
@@ -300,11 +331,13 @@
 ;*---------------------------------------------------------------------*/
 (define (cnst-alloc-keyword keyword loc)
    (define (lib-alloc-keyword)
-      (let ((var (def-global-svar! (symbol-append (gensym 'keyword) '::keyword)
+      (let ((var (def-global-svar! (make-typed-ident (gensym 'keyword)
+						     'keyword)
 		    *module*
 		    'a-keyword
-		    'now)))
-	 (put-hash! (cnst-info keyword var) *keyword-env*)
+		    'now))
+	    (vs (cnst-alloc-string (keyword->string keyword) loc)))
+	 (hashtable-put! *keyword-env* keyword (cnst-info keyword var))
 	 (add-cnst-sexp! `(set! (@ ,(global-id var) ,(global-module var))
 				,(coerce!
 				  (instantiate::app
@@ -313,12 +346,9 @@
 				     (fun (instantiate::var
 					     (loc loc)
 					     (type *keyword*)
-					     (variable *string->keyword*)))
-				     (args (list (instantiate::atom
-						    (loc loc)
-						    (type *string*)
-						    (value (keyword->string
-							    keyword))))))
+					     (variable *bstring->keyword*)))
+				     (args (list vs)))
+				  var
 				  *obj*)))
 	 (instantiate::var
 	    (loc loc)
@@ -328,9 +358,9 @@
       (let ((offset *cnst-offset*))
 	 (set! *cnst-offset* (+fx 1 *cnst-offset*))
 	 (set! *global-set* (cons keyword *global-set*))
-	 (put-hash! (cnst-info keyword offset) *keyword-env*)
+	 (hashtable-put! *keyword-env* keyword (cnst-info keyword offset))
 	 (make-cnst-table-ref offset loc)))
-   (let ((old (get-hash keyword *keyword-env*))
+   (let ((old (hashtable-get *keyword-env* keyword))
 	 (keyword-as-string (keyword->string keyword)))
       (cond
 	 (old
@@ -351,10 +381,11 @@
 ;*    cnst-alloc-procedure ...                                         */
 ;*---------------------------------------------------------------------*/
 (define (cnst-alloc-procedure procedure loc)
-   (let ((var (def-global-scnst! (symbol-append (gensym 'proc) '::procedure)
+   (let ((var (def-global-scnst! (make-typed-ident (gensym 'proc) 'procedure)
 		 *module*
 		 procedure
-		 'sfun)))
+		 'sfun
+		 loc)))
       (instantiate::var
 	 (loc loc)
 	 (type (variable-type var))
@@ -365,10 +396,11 @@
 ;*---------------------------------------------------------------------*/
 (define (cnst-alloc-real real loc)
    (define (lib-alloc-real)
-      (let ((var (def-global-scnst! (symbol-append (gensym 'real) '::real)
+      (let ((var (def-global-scnst! (make-typed-ident (gensym 'real) 'real)
 		    *module*
 		    real
-		    'sreal)))
+		    'sreal
+		    loc)))
 	 (set! *real-env* (cons (cons real var) *real-env*))
 	 (instantiate::var
 	    (loc loc)
@@ -393,6 +425,74 @@
 	 (else
 	  (lib-alloc-real)))))
 
+;*---------------------------------------------------------------------*/
+;*    cnst-alloc-elong ...                                             */
+;*---------------------------------------------------------------------*/
+(define (cnst-alloc-elong elong loc)
+   (define (lib-alloc-elong)
+      (let ((var (def-global-scnst! (make-typed-ident (gensym 'elong) 'belong)
+		    *module*
+		    elong
+		    'selong
+		    loc)))
+	 (set! *elong-env* (cons (cons elong var) *elong-env*))
+	 (instantiate::var
+	    (loc loc)
+	    (type (variable-type var))
+	    (variable var))))
+   (define (find-elong)
+      (let loop ((list *elong-env*))
+	 (cond
+	    ((null? list)
+	     #f)
+	    ((=elong (car (car list)) elong)
+	     (cdr (car list)))
+	    (else
+	     (loop (cdr list))))))
+   (let ((old (find-elong)))
+      (cond
+	 (old
+	  (instantiate::var
+	     (loc loc)
+	     (type (variable-type old))
+	     (variable old)))
+	 (else
+	  (lib-alloc-elong)))))
+
+;*---------------------------------------------------------------------*/
+;*    cnst-alloc-llong ...                                             */
+;*---------------------------------------------------------------------*/
+(define (cnst-alloc-llong llong loc)
+   (define (lib-alloc-llong)
+      (let ((var (def-global-scnst! (make-typed-ident (gensym 'llong) 'bllong)
+		    *module*
+		    llong
+		    'sllong
+		    loc)))
+	 (set! *llong-env* (cons (cons llong var) *llong-env*))
+	 (instantiate::var
+	    (loc loc)
+	    (type (variable-type var))
+	    (variable var))))
+   (define (find-llong)
+      (let loop ((list *llong-env*))
+	 (cond
+	    ((null? list)
+	     #f)
+	    ((=llong (car (car list)) llong)
+	     (cdr (car list)))
+	    (else
+	     (loop (cdr list))))))
+   (let ((old (find-llong)))
+      (cond
+	 (old
+	  (instantiate::var
+	     (loc loc)
+	     (type (variable-type old))
+	     (variable old)))
+	 (else
+	  (lib-alloc-llong)))))
+
 ;*---------------------------------------------------------------------*/
 ;*    cnst-alloc-list ...                                              */
 ;*---------------------------------------------------------------------*/
@@ -421,14 +521,14 @@
 				      (value (car pair))))
 			    (loop (cdr pair)))))))))
    (define (lib-alloc-list)
-      (let ((var (def-global-svar! (symbol-append (gensym 'list) '::pair)
+      (let ((var (def-global-svar! (make-typed-ident (gensym 'list) 'pair)
 		    *module*
 		    'cnst-list
 		    'now)))
 	 (if *shared-cnst?*
 	     (set! *list-env* (cons (cnst-info pair var) *list-env*)))
 	 (add-cnst-sexp! `(set! (@ ,(global-id var) ,(global-module var))
-				,(coerce! (cnst-list pair) *obj*)))
+				,(coerce! (cnst-list pair) var *obj*)))
 	 (instantiate::var
 	    (loc loc)
 	    (type (variable-type var))
@@ -517,6 +617,7 @@
 					     (value (vector-tag vec))))))
 				var-body)))
 			   var-body)))))
+	#unspecified
 	*obj*)))
    (define (read-alloc-vector)
       (let ((offset *cnst-offset*))
@@ -526,7 +627,7 @@
 	     (set! *vector-env* (cons (cnst-info vec offset) *vector-env*)))
 	 (make-cnst-table-ref offset loc)))
    (define (lib-alloc-vector)
-      (let ((var (def-global-svar! (symbol-append (gensym 'vector) '::vector)
+      (let ((var (def-global-svar! (make-typed-ident (gensym 'vector) 'vector)
 		    *module*
 		    'cnst-vector
 		    'now)))
@@ -565,10 +666,11 @@
 ;*---------------------------------------------------------------------*/
 (define (cnst-alloc-tvector tvec loc)
    (define (C-static-alloc-tvector)
-      (let ((var (def-global-scnst! (symbol-append (gensym 'tvec) '::tvector)
+      (let ((var (def-global-scnst! (make-typed-ident (gensym 'tvec) 'tvector)
 		    *module*
 		    tvec
-		    'stvector)))
+		    'stvector
+		    loc)))
 	 (let* ((id  (type-id (a-tvector-type tvec)))
 		(aid (cnst! (instantiate::kwote
 			       (loc loc)
@@ -596,3 +698,76 @@
        (C-static-alloc-tvector)
        (read-alloc-tvector)))
 
+;*---------------------------------------------------------------------*/
+;*    cnst-alloc-struct ...                                            */
+;*---------------------------------------------------------------------*/
+(define (cnst-alloc-struct struct loc)
+   (define (cnst-struct-node)
+      (cnst!
+       (coerce!
+	(let ((var (make-local-svar 'var *struct*)))
+	   (instantiate::let-var
+	      (loc loc)
+	      (type *obj*)
+	      (bindings
+	       (list (cons
+		      var
+		      (instantiate::app
+			 (loc loc)
+			 (type *struct*)
+			 (fun (instantiate::var
+				 (loc loc)
+				 (type (variable-type *list->struct*))
+				 (variable *list->struct*)))
+			 (args (list
+				(instantiate::kwote
+				   (loc loc)
+				   (type *obj*)
+				   (value (struct->list struct)))))))))
+	      (body (instantiate::var
+		       (loc loc)
+		       (type (variable-type var))
+		       (variable var)))))
+	#unspecified
+	*obj*)))
+   (define (read-alloc-struct)
+      (let ((offset *cnst-offset*))
+	 (set! *cnst-offset* (+fx 1 *cnst-offset*))
+	 (set! *global-set* (cons struct *global-set*))
+	 (if *shared-cnst?*
+	     (set! *struct-env* (cons (cnst-info struct offset) *struct-env*)))
+	 (make-cnst-table-ref offset loc)))
+   (define (lib-alloc-struct)
+      (let ((var (def-global-svar! (make-typed-ident (gensym 'struct) 'struct)
+		    *module*
+		    'cnst-struct
+		    'now)))
+	 (if *shared-cnst?*
+	     (set! *struct-env* (cons (cnst-info struct var) *struct-env*)))
+	 (add-cnst-sexp! `(set! (@ ,(global-id var) ,(global-module var))
+				,(cnst-struct-node)))
+	 (instantiate::var
+	    (loc loc)
+	    (type (variable-type var))
+	    (variable var))))
+   (let ((old (and *shared-cnst?*
+		   (let loop ((env *struct-env*))
+		      (cond
+			 ((null? env)
+			  #f)
+			 ((equal? (cnst-info-cnst (car env)) struct)
+			  (car env))
+			 (else
+			  (loop (cdr env))))))))
+      (cond
+	 (old
+	  (if (eq? *init-mode* 'lib)
+	      (instantiate::var
+		 (loc loc)
+		 (type (get-default-type))
+		 (variable (cnst-info-offset old)))
+	      (make-cnst-table-ref (cnst-info-offset old) loc)))
+	 ((eq? *init-mode* 'lib)
+	  (lib-alloc-struct))
+	 (else
+	  (read-alloc-struct)))))
