@@ -1,36 +1,10 @@
-;* --------------------------------------------------------------------*/
-;*    Copyright (c) 1992-1998 by Manuel Serrano. All rights reserved.  */
-;*                                                                     */
-;*                                     ,--^,                           */
-;*                               _ ___/ /|/                            */
-;*                           ,;'( )__, ) '                             */
-;*                          ;;  //   L__.                              */
-;*                          '   \   /  '                               */
-;*                               ^   ^                                 */
-;*                                                                     */
-;*                                                                     */
-;*    This program is distributed in the hope that it will be useful.  */
-;*    Use and copying of this software and preparation of derivative   */
-;*    works based upon this software are permitted, so long as the     */
-;*    following conditions are met:                                    */
-;*           o credit to the authors is acknowledged following         */
-;*             current academic behaviour                              */
-;*           o no fees or compensation are charged for use, copies,    */
-;*             or access to this software                              */
-;*           o this copyright notice is included intact.               */
-;*      This software is made available AS IS, and no warranty is made */
-;*      about the software or its performance.                         */
-;*                                                                     */
-;*      Bug descriptions, use reports, comments or suggestions are     */
-;*      welcome. Send them to                                          */
-;*        Manuel Serrano -- Manuel.Serrano@unice.fr                    */
-;*-------------------------------------------------------------------- */
 ;*=====================================================================*/
-;*    serrano/prgm/project/bigloo/comptime1.9b/Globalize/free.scm      */
+;*    serrano/prgm/project/bigloo/comptime/Globalize/free.scm          */
 ;*    -------------------------------------------------------------    */
 ;*    Author      :  Manuel Serrano                                    */
 ;*    Creation    :  Fri Jan 27 14:20:15 1995                          */
-;*    Last change :  Wed Apr 16 12:19:43 1997 (serrano)                */
+;*    Last change :  Mon Dec 12 19:31:45 2005 (serrano)                */
+;*    Copyright   :  1995-2005 Manuel Serrano, see LICENSE file        */
 ;*    -------------------------------------------------------------    */
 ;*    The search of free variables.                                    */
 ;*=====================================================================*/
@@ -212,11 +186,11 @@
       (node-free fun (node-free* args free))))
 
 ;*---------------------------------------------------------------------*/
-;*    node-free ::pragma ...                                           */
+;*    node-free ::extern ...                                           */
 ;*---------------------------------------------------------------------*/
-(define-method (node-free node::pragma free)
-   (with-access::pragma node (args)
-      (node-free* args free)))
+(define-method (node-free node::extern free)
+   (with-access::extern node (expr*)
+      (node-free* expr* free)))
 
 ;*---------------------------------------------------------------------*/
 ;*    node-free ::cast ...                                             */
@@ -360,9 +334,13 @@
 	  closure
 	  (let* ((gloclo   (make-global-closure global))
 		 (arity    (fun-arity (global-value global)))
-		 (make-clo (if (<fx arity 0)
-			       'make-va-procedure
-			       'make-fx-procedure))
+		 (make-clo (cond
+			      ((<fx arity 0)
+			       'make-va-procedure)
+;* 			      ((sfun-optional? (global-value global))  */
+;* 			       'make-opt-procedure)                    */
+			      (else
+			       'make-fx-procedure)))
 		 (node     (sexp->node `(,make-clo
 					 ,(instantiate::var
 					     (loc loc)
@@ -390,7 +368,9 @@
 				      'sgfun)
 				     (else
 				      'sfun))
-				  'sfun))))
+				  'sfun)
+			      loc)))
+	     (global-library-set! closure (global-library global))
 	     (global-import-set! closure (global-import global))
 	     (fun-the-closure-set! (global-value global) closure)
 	     ;; this next setting is require by the Cfa pass
@@ -411,6 +391,7 @@
       (if (local? (sfun/Ginfo-the-closure info))
 	  (sfun/Ginfo-the-closure info)
 	  (let ((closure (make-local-svar (local-id local) *procedure*)))
+	     (local-user?-set! closure (local-user? local))
 	     (widen!::svar/Ginfo (local-value closure))
 	     (widen!::local/Ginfo closure)
 	     (sfun/Ginfo-the-closure-set! info closure)
