@@ -1,36 +1,10 @@
-;* --------------------------------------------------------------------*/
-;*    Copyright (c) 1992-1998 by Manuel Serrano. All rights reserved.  */
-;*                                                                     */
-;*                                     ,--^,                           */
-;*                               _ ___/ /|/                            */
-;*                           ,;'( )__, ) '                             */
-;*                          ;;  //   L__.                              */
-;*                          '   \   /  '                               */
-;*                               ^   ^                                 */
-;*                                                                     */
-;*                                                                     */
-;*    This program is distributed in the hope that it will be useful.  */
-;*    Use and copying of this software and preparation of derivative   */
-;*    works based upon this software are permitted, so long as the     */
-;*    following conditions are met:                                    */
-;*           o credit to the authors is acknowledged following         */
-;*             current academic behaviour                              */
-;*           o no fees or compensation are charged for use, copies,    */
-;*             or access to this software                              */
-;*           o this copyright notice is included intact.               */
-;*      This software is made available AS IS, and no warranty is made */
-;*      about the software or its performance.                         */
-;*                                                                     */
-;*      Bug descriptions, use reports, comments or suggestions are     */
-;*      welcome. Send them to                                          */
-;*        Manuel Serrano -- Manuel.Serrano@unice.fr                    */
-;*-------------------------------------------------------------------- */
 ;*=====================================================================*/
-;*    serrano/prgm/project/bigloo/comptime1.9b/Globalize/node.scm      */
+;*    serrano/prgm/project/bigloo/comptime/Globalize/node.scm          */
 ;*    -------------------------------------------------------------    */
 ;*    Author      :  Manuel Serrano                                    */
 ;*    Creation    :  Fri Jan 27 14:12:58 1995                          */
-;*    Last change :  Thu Apr  3 14:19:41 1997 (serrano)                */
+;*    Last change :  Fri Dec  9 14:13:49 2005 (serrano)                */
+;*    Copyright   :  1995-2005 Manuel Serrano, see LICENSE file        */
 ;*    -------------------------------------------------------------    */
 ;*    We transforme the ast in order to fix the free variables, to     */
 ;*    remove the useless local functions (globalized or integrated     */
@@ -93,6 +67,7 @@
 	  (let* ((var (make-local-svar (local-id (car formals)) *obj*))
 		 (o.n (cons (car formals) var)))
 	     (local-access-set! var 'cell-globalize)
+	     (local-user?-set! var (local-user? (car formals)))
 	     (widen!::svar/Ginfo (local-value var)
 		(celled? #t)
 		(kaptured? #t))
@@ -278,11 +253,11 @@
       node))
 
 ;*---------------------------------------------------------------------*/
-;*    glo! ::pragma ...                                                */
+;*    glo! ::extern ...                                                */
 ;*---------------------------------------------------------------------*/
-(define-method (glo! node::pragma integrator)
-   (with-access::pragma node (args)
-      (glo*! args integrator)
+(define-method (glo! node::extern integrator)
+   (with-access::extern node (expr*)
+      (glo*! expr* integrator)
       node))
 
 ;*---------------------------------------------------------------------*/
@@ -311,6 +286,7 @@
 		       (let ((a-var (make-local-svar 'aux *obj*))
 			     (loc   (node-loc node)))
 			  (local-access-set! var 'cell-globalize)
+			  (local-user?-set! a-var (local-user? var))
 			  (widen!::svar/Ginfo (local-value a-var)
 			     (kaptured? #f))
 			  (instantiate::let-var
@@ -367,6 +343,12 @@
       (let liip ((obindings locals)
 		 (nbindings '())
 		 (ebindings '()))
+	 (if (not (null? obindings))
+	     (trace (globalize 4)
+		    "glo!(let-fun): " (shape (car obindings)) ": "
+		    "  integrator: " (shape integrator)
+		    "  escape: " (local/Ginfo-escape? (car obindings))
+		    #\Newline))
 	 (cond
 	    ((null? obindings)
 	     (set! locals nbindings)
@@ -397,7 +379,8 @@
 		      (if (celled? var)
 			  (begin
 			     (local-type-set! var *obj*)
-			     (set-cdr! binding (a-make-cell val var))))))
+			     (set-cdr! binding (a-make-cell (cdr binding)
+							    var))))))
 		bindings)
       (set! body (glo! body integrator))
       node))
@@ -524,15 +507,21 @@
    (let* ((fun      (local-value local))
 	  (arity    (sfun-arity fun))
 	  (kaptured (sfun/Ginfo-kaptured (local-value local)))
-	  (loc      (node-loc (sfun-body fun)))
-	  (make-p   (if (>=fx arity 0) 'make-fx-procedure 'make-va-procedure)))
+	  (loc      #unspecified)
+	  (make-p   (cond
+		       ((<fx arity 0)
+			'make-va-procedure)
+;* 		       ((sfun-optional? fun)                           */
+;* 			'make-opt-procedure)                           */
+		       (else
+			'make-fx-procedure))))
       (instantiate::app
 	 (loc loc)
 	 (type *_*)
 	 (fun (instantiate::var
 		 (loc loc)
 		 (type *_*)
-		 (variable (find-global make-p 'foreign))))
+		 (variable (find-global/module make-p 'foreign))))
 	 (args (list (instantiate::var
 			(loc loc)
 			(type *_*)
@@ -584,7 +573,7 @@
       (fun (instantiate::var
 	      (loc loc)
 	      (type *_*)
-	      (variable (find-global 'procedure-set! 'foreign))))
+	      (variable (find-global/module 'procedure-set! 'foreign))))
       (args (list (instantiate::var
 		     (loc loc)
 		     (type *_*)
