@@ -1,36 +1,10 @@
-;* --------------------------------------------------------------------*/
-;*    Copyright (c) 1992-1998 by Manuel Serrano. All rights reserved.  */
-;*                                                                     */
-;*                                     ,--^,                           */
-;*                               _ ___/ /|/                            */
-;*                           ,;'( )__, ) '                             */
-;*                          ;;  //   L__.                              */
-;*                          '   \   /  '                               */
-;*                               ^   ^                                 */
-;*                                                                     */
-;*                                                                     */
-;*    This program is distributed in the hope that it will be useful.  */
-;*    Use and copying of this software and preparation of derivative   */
-;*    works based upon this software are permitted, so long as the     */
-;*    following conditions are met:                                    */
-;*           o credit to the authors is acknowledged following         */
-;*             current academic behaviour                              */
-;*           o no fees or compensation are charged for use, copies,    */
-;*             or access to this software                              */
-;*           o this copyright notice is included intact.               */
-;*      This software is made available AS IS, and no warranty is made */
-;*      about the software or its performance.                         */
-;*                                                                     */
-;*      Bug descriptions, use reports, comments or suggestions are     */
-;*      welcome. Send them to                                          */
-;*        Manuel Serrano -- Manuel.Serrano@unice.fr                    */
-;*-------------------------------------------------------------------- */
 ;*=====================================================================*/
 ;*    serrano/prgm/project/bigloo/comptime/Inline/recursion.scm        */
 ;*    -------------------------------------------------------------    */
 ;*    Author      :  Manuel Serrano                                    */
 ;*    Creation    :  Wed Jun 19 13:40:47 1996                          */
-;*    Last change :  Wed Mar  4 11:26:16 1998 (serrano)                */
+;*    Last change :  Wed Jun 15 06:27:41 2005 (serrano)                */
+;*    Copyright   :  1996-2005 Manuel Serrano, see LICENSE file        */
 ;*    -------------------------------------------------------------    */
 ;*    The inlining of recursive functions.                             */
 ;*=====================================================================*/
@@ -39,7 +13,8 @@
 ;*    The module                                                       */
 ;*---------------------------------------------------------------------*/
 (module inline_recursion
-   (include "Tools/trace.sch")
+   (include "Tools/trace.sch"
+	    "Tools/verbose.sch")
    (import  engine_param
 	    type_type
 	    ast_var
@@ -74,8 +49,7 @@
 	  ;; ok, we do a smart inlining.
 	  (begin
 	     (trace (inline 2) "inline-app-recursive: " (shape node) #\Newline)
-	     (let ((labels-node (inline-app-labels node kfactor stack)))
-		labels-node)))))
+	     (inline-app-labels node kfactor stack)))))
 
 ;*---------------------------------------------------------------------*/
 ;*    inline-app-labels ...                                            */
@@ -91,7 +65,7 @@
 	  (old-sfun       (variable-value variable))
 	  (rec-calls      (isfun-recursive-calls old-sfun))
 	  (old-args       (sfun-args old-sfun))
-	  (inv-args       (invariant-args variable rec-calls))
+	  (inv-args       (invariant-args node variable rec-calls))
 	  (var-args       (variant-args variable))
 	  (new-args       (map (lambda (l)
 				  (clone-local
@@ -117,10 +91,24 @@
 			     (class 'sfun)
 			     (body new-body)))
 	  (new-kfactor    (*inlining-reduce-kfactor* kfactor)))
+      ;; we mark the formal parameter as compiler parameters
+      (for-each (lambda (new old)
+		   (local-user?-set! new (local-user? old)))
+		new-args
+		old-args)
       ;; we set the new sfun
+      (local-user?-set! local (cond
+				 ((global? variable)
+				  (global-user? variable))
+				 ((local? variable)
+				  (local-user? variable))
+				 (else
+				  #f)))
       (local-value-set! local new-sfun)
       ;; some traces
       (trace (inline 3)
+	     "   new-fun  : " (shape local) #\Newline
+	     "   user?    : " (local-user? local) #\Newline
 	     "   rec-calls: " (shape rec-calls) #\Newline
 	     "   inv-args : " (shape inv-args) #\Newline
 	     "   var-args : " (shape var-args) #\Newline)
@@ -265,10 +253,10 @@
 	      (find-recursive-calls* args var))))
 
 ;*---------------------------------------------------------------------*/
-;*    find-recursive-calls ::pragma ...                                */
+;*    find-recursive-calls ::extern ...                                */
 ;*---------------------------------------------------------------------*/
-(define-method (find-recursive-calls node::pragma var)
-   (find-recursive-calls* (pragma-args node) var))
+(define-method (find-recursive-calls node::extern var)
+   (find-recursive-calls* (extern-expr* node) var))
 
 ;*---------------------------------------------------------------------*/
 ;*    find-recursive-calls ::cast ...                                  */
