@@ -1,38 +1,12 @@
-;* --------------------------------------------------------------------*/
-;*    Copyright (c) 1992-1998 by Manuel Serrano. All rights reserved.  */
-;*                                                                     */
-;*                                     ,--^,                           */
-;*                               _ ___/ /|/                            */
-;*                           ,;'( )__, ) '                             */
-;*                          ;;  //   L__.                              */
-;*                          '   \   /  '                               */
-;*                               ^   ^                                 */
-;*                                                                     */
-;*                                                                     */
-;*    This program is distributed in the hope that it will be useful.  */
-;*    Use and copying of this software and preparation of derivative   */
-;*    works based upon this software are permitted, so long as the     */
-;*    following conditions are met:                                    */
-;*           o credit to the authors is acknowledged following         */
-;*             current academic behaviour                              */
-;*           o no fees or compensation are charged for use, copies,    */
-;*             or access to this software                              */
-;*           o this copyright notice is included intact.               */
-;*      This software is made available AS IS, and no warranty is made */
-;*      about the software or its performance.                         */
-;*                                                                     */
-;*      Bug descriptions, use reports, comments or suggestions are     */
-;*      welcome. Send them to                                          */
-;*        Manuel Serrano -- Manuel.Serrano@unice.fr                    */
-;*-------------------------------------------------------------------- */
 ;*=====================================================================*/
 ;*    serrano/prgm/project/bigloo/comptime/Module/impuse.scm           */
 ;*    -------------------------------------------------------------    */
 ;*    Author      :  Manuel Serrano                                    */
 ;*    Creation    :  Tue Jun  4 12:25:53 1996                          */
-;*    Last change :  Wed Feb 11 09:25:06 1998 (serrano)                */
+;*    Last change :  Wed Sep 14 13:40:03 2005 (serrano)                */
+;*    Copyright   :  1996-2005 Manuel Serrano, see LICENSE file        */
 ;*    -------------------------------------------------------------    */
-;*    The import/usr/from clauses compilation                          */
+;*    The import/use/from clauses compilation                          */
 ;*=====================================================================*/
 
 ;*---------------------------------------------------------------------*/
@@ -43,19 +17,24 @@
    (include "Ast/unit.sch"
 	    "Tools/trace.sch")
    
-   (import  module_module
+   (import  read_reader
+	    backend_backend
+	    module_module
 	    module_prototype
 	    module_class
 	    module_include
+	    module_checksum
 	    tools_speek
 	    tools_error
-	    tools_file
+	    tools_location
 	    type_type
+	    object_class
 	    read_access
 	    read_inline
 	    ast_var
 	    ast_find-gdefs
 	    ast_glo-decl
+	    ast_ident
 	    engine_param
 	    init_main)
 	   
@@ -63,8 +42,8 @@
 	    (make-use-compiler)
 	    (make-from-compiler)
 	    (get-imported-modules)
-	    (import-with-module! ::symbol)
-	    (import-parser       ::symbol prototype)))
+	    (import-with-module! ::symbol loc)
+	    (import-parser       ::symbol prototype . import-src)))
 
 ;*---------------------------------------------------------------------*/
 ;*    make-import-compiler ...                                         */
@@ -98,34 +77,44 @@
    (let ((mode (car clause)))
       (match-case clause
 	 ((?- . ?protos)
-	  (for-each (lambda (proto) (impuse-parser proto mode)) protos))
+	  (for-each (lambda (proto) (impuse-parser proto mode clause)) protos))
 	 (else
-	  (user-error "Parse error"
-		      (string-append "Illegal `"
-				     (string-downcase (symbol->string mode))
-				     "' clause")
-		      clause
-		      '())))))
+	  (user-error/location (find-location *module-clause*)
+			       "Parse error"
+			       (string-append
+				"Illegal `"
+				(string-downcase (symbol->string mode))
+				"' clause")
+			       clause
+			       '())))))
    
 ;*---------------------------------------------------------------------*/
 ;*    import-all-module ...                                            */
 ;*---------------------------------------------------------------------*/
-(define (import-all-module module::symbol mode)
-   (let ((b (assq module *import-module-list*)))
+(define (import-all-module module::symbol mode src)
+   (let ((b (assq module *import-module-list*))
+	 (loc (find-location/loc src (find-location *module-clause*))))
       (if (pair? b)
 	  (imode-vars-set! (cdr b) 'all)
 	  (set! *import-module-list*
-		(cons (cons module (imode module mode 'all #unspecified))
-		      *import-module-list*)))))
+		(cons
+		 (cons module (imode module mode 'all #unspecified loc src))
+		 *import-module-list*)))))
 
 ;*---------------------------------------------------------------------*/
 ;*    import-1-module ...                                              */
 ;*---------------------------------------------------------------------*/
-(define (import-1-module module::symbol var mode)
-   (let ((cell (assq module *import-module-list*)))
+(define (import-1-module module::symbol var mode src)
+   (let ((cell (assq module *import-module-list*))
+	 (loc (find-location/loc src (find-location *module-clause*))))
       (if (not (pair? cell))
 	  (set! *import-module-list*
-		(cons (cons module (imode module mode (list var) #unspecified))
+		(cons (cons module (imode module
+					  mode
+					  (list var)
+					  #unspecified
+					  loc
+					  src))
 		      *import-module-list*))
 	  (let ((b (cdr cell)))
 	     (case (imode-vars b)
@@ -140,11 +129,12 @@
 ;*---------------------------------------------------------------------*/
 ;*    import-with-module! ...                                          */
 ;*---------------------------------------------------------------------*/
-(define (import-with-module! module)
-   (let ((b (assq module *import-module-list*)))
+(define (import-with-module! module src)
+   (let ((b (assq module *import-module-list*))
+	 (loc (find-location/loc src (find-location *module-clause*))))
       (if (not (pair? b))
 	  (set! *import-module-list*
-		(cons (cons module (imode module 'with '() 0))
+		(cons (cons module (imode module 'with '() 0 loc src))
 		      *import-module-list*)))))
    
 ;*---------------------------------------------------------------------*/
@@ -156,56 +146,58 @@
 ;*            |  (variable module-name)                                */
 ;*            |  (variable module-name "file-name" *)                  */
 ;*---------------------------------------------------------------------*/
-(define (impuse-parser prototype mode)
+(define (impuse-parser prototype mode import-src)
    (trace (ast 2) "impuse-parser: " prototype " " mode #\Newline)
-   (match-case prototype
-      (((and ?module (? symbol?)) (and ?fname (? string?)) . ?rest)
-       ;; (module-name "file-name" ...)
-       (let loop ((rest   rest)
-		  (fnames (list fname)))
-	  (cond
-	     ((null? rest)
-	      (add-access! module (reverse! fnames))
-	      (import-all-module module mode))
-	     ((string? (car rest))
-	      (loop (cdr rest)
-		    (cons (car rest) fnames)))
-	     (else
-	     (user-error "Parse error"
-			 "Illegal imported clause"
-			 prototype
-			 '())))))
-      (((and ?var (? symbol?)) (and ?module (? symbol?)))
-       ;; (variable module-name)
-       (import-1-module module var mode))
-      (((and (? symbol?) ?v) (and (? symbol?) ?m) (and ?f (? string?)) . ?rest)
-       ;; (variable module-name "file-name" ...) 
-       (let loop ((rest   rest)
-		  (fnames (list f)))
-	  (cond
-	     ((null? rest)
-	      (add-access! m (reverse! fnames))
-	      (import-1-module m v mode))
-	     ((string? (car rest))
-	      (loop (cdr rest) (cons (car rest) fnames)))
-	     (else
-	     (user-error "Parse error"
-			 "Illegal imported clause"
-			 prototype
-			 '())))))
-      ((and ?module (? symbol?))
+   (cond
+      ((symbol? prototype)
        ;; module-name
-       (import-all-module module mode))
+       (import-all-module prototype mode import-src))
+      ((list? prototype)
+       (let ((inv (reverse prototype)))
+	  (let loop ((lst inv)
+		     (files '()))
+	     (cond
+		((not (pair? lst))
+		 (user-error "Parse error"
+			     "Illegal import/use/from clause"
+			     prototype
+			     '()))
+		((string? (car lst))
+		 (loop (cdr lst) (cons (car lst) files)))
+		((symbol? (car lst))
+		 (let ((mod (car lst))
+		       (vars (cdr lst)))
+		    (cond
+		       ((null? vars)
+			;; (module-name "file-name"+)
+			(if (pair? files) (add-access! mod files))
+			(import-all-module mod mode prototype))
+		       ((every? symbol? vars)
+			;; (var1 var2 ... varN module-name "file-name"*)
+			(if (pair? files) (add-access! mod files))
+			(for-each (lambda (v)
+				     (import-1-module mod v mode prototype))
+				  vars))
+		       (else
+			(user-error "Parse error"
+				    "Illegal import/use/from clause"
+				    prototype
+				    '())))))
+		(else
+		 (user-error "Parse error"
+			     "Illegal import/use/from clause"
+			     prototype
+			     '()))))))
       (else
        (user-error "Parse error"
-		   "Illegal `import/use' clause"
-		   prototype
-		   '()))))
-
+			     "Illegal import/use/from clause"
+			     prototype
+			     '()))))
+   
 ;*---------------------------------------------------------------------*/
 ;*    imode ...                                                        */
 ;*---------------------------------------------------------------------*/
-(define-struct imode id mode vars checksum)
+(define-struct imode id mode vars checksum loc src)
 
 ;*---------------------------------------------------------------------*/
 ;*    *import-module-list* ...                                         */
@@ -231,6 +223,8 @@
 ;*---------------------------------------------------------------------*/
 ;*    import-finalizer ...                                             */
 ;*    -------------------------------------------------------------    */
+;*    @label importation unit@                                         */
+;*    -------------------------------------------------------------    */
 ;*    In the `impuse' finalizer we read all the imported modules and   */
 ;*    if we have to, we create a unit in order to initialize imported  */
 ;*    modules.                                                         */
@@ -238,13 +232,35 @@
 (define (import-finalizer)
    (set! *import-module-list* (reverse! *import-module-list*))
    (trace (ast 2) "import-finalizer: " *import-module-list* #\Newline)
+   ;; when not in imported-module-list mode, we do the regular job, that
+   ;; is, we open the imported files.
    (let loop ((init* '()))
       (if (null? *import-module-list*)
 	  (if (pair? init*)
-	      (let loop ((init*      init*)
+	      (let loop ((init* init*)
+			 (init-name* '())
 			 (init-call* '()))
 		 (if (null? init*)
-		     (list (unit 'imported-modules 12 init-call* #t))
+		     ;; the unit number of importation must be smaller
+		     ;; than the one for classes. See module_class module
+		     ;; (@ref class.scm:class unit@).
+		     (let ((body (if (and (>fx *debug-module* 0)
+					  (memq 'module
+						(backend-debug-support
+						 (the-backend))))
+				     `((begin
+					  ,@(map (lambda (n c)
+						    `(begin
+							(pragma::int
+							 ,(string-append
+							   "puts(\"    import ("
+							   (symbol->string *module*)
+							   "): " n "\")"))
+							,c))
+						 init-name*
+						 init-call*)))
+				     init-call*)))
+			(list (unit 'imported-modules 12 body #t)))
 		     (let* ((id          (imode-id (car init*)))
 			    (checksum    (imode-checksum (car init*)))
 			    (init-fun-id (module-initialization-id id)))
@@ -255,9 +271,10 @@
 			   ;; from eval. We mark this.
 			   (global-evaluable?-set! global #f))
 			(loop (cdr init*)
-			      (cons `((@ ,init-fun-id ,id) ,checksum
-							   ,(symbol->string
-							     *module*))
+			      (cons (symbol->string id) init-name*)
+			      (cons `((@ ,init-fun-id ,id)
+				      ,checksum
+				      ,(symbol->string *module*))
 				    init-call*)))))
 	      '())
 	  (let ((module (car (car *import-module-list*)))
@@ -266,7 +283,7 @@
 	     (set! *imported-module-list* (cons module *imported-module-list*))
 	     (if (not (eq? (imode-mode imode) 'with))
 		 (read-imported-module imode))
-	     (loop (if (memq (imode-mode imode) '(import with))
+	     (loop (if (memq (imode-mode imode) '(import with from))
 		       (cons imode init*)
 		       init*))))))
 
@@ -288,10 +305,17 @@
 ;*---------------------------------------------------------------------*/
 ;*    import-parser ...                                                */
 ;*---------------------------------------------------------------------*/
-(define (import-parser module::symbol prototype)
-   (let ((proto (parse-prototype prototype)))
+(define (import-parser module::symbol prototype . import-src)
+   (let ((proto (parse-prototype prototype))
+	 (src   (if (pair? import-src)
+		    (car import-src)
+		    #f)))
       (if (not (pair? proto))
-	  (user-error "Parse error" "Illegal prototype" prototype '())
+	  (user-error/location (find-location *module-clause*)
+			       "Parse error"
+			       "Illegal prototype"
+			       prototype
+			       '())
 	  (case (car proto)
 	     ((sfun sifun sgfun)
 	      (declare-global-sfun! (cadr proto)
@@ -299,15 +323,18 @@
 				    module
 				    'import
 				    (car proto)
-				    prototype))
+				    prototype
+				    src))
 	     ((svar)
-	      (declare-global-svar! (cadr proto) module	'import	prototype))
+	      (declare-global-svar! (cadr proto) module	'import	prototype src))
 	     ((class)
-	      (declare-class! (cdr proto) module 'import #f prototype))
+	      (declare-class! (cdr proto) module 'import #f #f prototype src))
+	     ((abstract-class)
+	      (declare-class! (cdr proto) module 'import #f #t prototype src))
 	     ((final-class)
-	      (declare-class! (cdr proto) module 'import #t prototype))
+	      (declare-class! (cdr proto) module 'import #t #f prototype src))
 	     ((wide-class)
-	      (declare-wide-class! (cdr proto) module 'import prototype))
+	      (declare-wide-class! (cdr proto) module 'import prototype src))
 	     (else
 	      (user-error "Parse error" "Illegal prototype" prototype '()))))))
    
@@ -323,60 +350,59 @@
 	       (if (eq? (imode-mode imode) 'use) "used" "imported")
 	       " module " module "]" #\Newline)
       (if (not b)
-	  (user-error "read-imported-module"
-		      "Can't find such module"
-		      module
-		      '())
+	  (user-error/location (imode-loc imode)
+			       "read-imported-module"
+			       "Can't find such module"
+			       module
+			       '())
 	  (let* ((fnames (cdr b))
 		 (fname  (find-file/path (car fnames) *load-path*)))
 	     (if (not (string? fname))
-		 (user-error "read-imported-module"
-			     "Can't open such file"
-			     (car fnames)
-			     '())
+		 (user-error/location (imode-loc imode)
+				      "read-imported-module"
+				      "Can't open such file"
+				      (car fnames)
+				      '())
 		 (let ((port (open-input-file fname)))
+		    (reader-reset!)
 		    (if (not (input-port? port))
-			(user-error "read-imported-module"
-				    "Can't open such file"
-				    (car fnames)
-				    '())
-			(let ((handler
-			       (lambda (escape proc mes obj)
-				  (notify-error proc mes obj)
-				  (flush-output-port (current-error-port))
-				  (close-input-port port)
-				  (exit-bigloo -7))))
-			   (try
-			    (let* ((mdecl (read port #t))
-				   (prov  (append
-					   (begin
-					      (reset-include-consumed-directive!)
-					      (reset-include-consumed-code!)
-					      (consume-module! module mdecl))
-					   (get-include-consumed-directive)))
-				   (code   (get-include-consumed-code))
-				   (check  (checksum-module mdecl)))
-			       (imode-checksum-set! imode check)
-			       (look-for-inline
-				(if (not (pair? wanted))
-				    (import-everything prov module)
-				    (import-wanted prov wanted module))
-				code
-				port
-				(cdr fnames)
-				module)
-			       (close-input-port port))
-				handler)))))))))
+			(user-error/location (imode-loc imode)
+					     "read-imported-module"
+					     "Can't open such file"
+					     (car fnames)
+					     '())
+			(unwind-protect
+			   (let* ((mdecl (compiler-read port #t))
+				  (prov (append
+					 (begin
+					    (reset-include-consumed-directive!)
+					    (reset-include-consumed-code!)
+					    (consume-module! module mdecl))
+					 (get-include-consumed-directive)))
+				  (code (get-include-consumed-code))
+				  (check (module-checksum mdecl))
+				  (src (imode-src imode)))
+			      (imode-checksum-set! imode check)
+			      (look-for-inline
+			       (if (not (pair? wanted))
+				   (import-everything prov module src)
+				   (import-wanted prov wanted module src))
+			       code
+			       port
+			       (cdr fnames)
+			       module)
+			      (close-input-port port))
+			   (close-input-port port)))))))))
    
 ;*---------------------------------------------------------------------*/
 ;*    import-everything ...                                            */
 ;*---------------------------------------------------------------------*/
-(define (import-everything provided module::symbol)
+(define (import-everything provided module::symbol src)
    (let loop ((provided provided)
 	      (inline   '()))
       (if (null? provided)
 	  inline
-	  (let ((p (import-parser module (car provided))))
+	  (let ((p (import-parser module (car provided) src)))
 	     (if (global? p)
 		 (let ((val (global-value p)))
 		    (loop (cdr provided)
@@ -393,7 +419,7 @@
 ;*---------------------------------------------------------------------*/
 ;*    import-wanted ...                                                */
 ;*---------------------------------------------------------------------*/
-(define (import-wanted provided wanted module::symbol)
+(define (import-wanted provided wanted module::symbol src)
    (let loop ((provided provided)
 	      (inline   '())
 	      (wanted   wanted))
@@ -403,19 +429,21 @@
 	 ((null? wanted) 
 	  inline)
 	 ((null? provided)
-	  (user-error module
-		      "Can't find exportation for these identifiers"
-		      wanted
-		      '()))
+	  (user-error/location (find-location *module-clause*)
+			       module
+			       "Can't find exportation for these identifiers"
+			       wanted
+			       '()))
 	 (else
 	  (let ((proto (parse-prototype (car provided))))
 	     (if (pair? proto)
-		 (let ((id (cadr proto)))
+		 (let ((id (fast-id-of-id (cadr proto)
+					  (find-location (car provided)))))
 		    (if (not (memq id wanted))
 			(loop (cdr provided)
 			      inline
 			      wanted)
-			(let ((p (import-parser module (car provided))))
+			(let ((p (import-parser module (car provided) src)))
 			   (cond
 			      ((global? p)
 			       (loop (cdr provided)
