@@ -1,36 +1,10 @@
-;* --------------------------------------------------------------------*/
-;*    Copyright (c) 1992-1998 by Manuel Serrano. All rights reserved.  */
-;*                                                                     */
-;*                                     ,--^,                           */
-;*                               _ ___/ /|/                            */
-;*                           ,;'( )__, ) '                             */
-;*                          ;;  //   L__.                              */
-;*                          '   \   /  '                               */
-;*                               ^   ^                                 */
-;*                                                                     */
-;*                                                                     */
-;*    This program is distributed in the hope that it will be useful.  */
-;*    Use and copying of this software and preparation of derivative   */
-;*    works based upon this software are permitted, so long as the     */
-;*    following conditions are met:                                    */
-;*           o credit to the authors is acknowledged following         */
-;*             current academic behaviour                              */
-;*           o no fees or compensation are charged for use, copies,    */
-;*             or access to this software                              */
-;*           o this copyright notice is included intact.               */
-;*      This software is made available AS IS, and no warranty is made */
-;*      about the software or its performance.                         */
-;*                                                                     */
-;*      Bug descriptions, use reports, comments or suggestions are     */
-;*      welcome. Send them to                                          */
-;*        Manuel Serrano -- Manuel.Serrano@unice.fr                    */
-;*-------------------------------------------------------------------- */
 ;*=====================================================================*/
 ;*    serrano/prgm/project/bigloo/comptime/Reduce/1occ.scm             */
 ;*    -------------------------------------------------------------    */
 ;*    Author      :  Manuel Serrano                                    */
 ;*    Creation    :  Thu Jul 13 10:29:17 1995                          */
-;*    Last change :  Wed Mar  4 11:26:24 1998 (serrano)                */
+;*    Last change :  Fri Oct  7 06:17:49 2005 (serrano)                */
+;*    Copyright   :  1995-2005 Manuel Serrano, see LICENSE file        */
 ;*    -------------------------------------------------------------    */
 ;*    The removal of the local variables appearing just once.          */
 ;*    The only goal of this pass is to prune the ast.                  */
@@ -46,7 +20,8 @@
 	    tools_error
 	    type_type
 	    type_cache
-	    coerce_typeof
+	    type_typeof
+	    type_misc
 	    coerce_coerce
 	    effect_effect
 	    ast_var
@@ -59,7 +34,7 @@
 ;*    reduce-1occ! ...                                                 */
 ;*---------------------------------------------------------------------*/
 (define (reduce-1occ! globals)
-   (verbose 2 #"      single occurrence      ")
+   (verbose 2 #"      single occurrence")
    ;; we have to recompute the occurrences because both `copy-propagation
    ;; and `cse' have changed the number of occurrence (in two directions).
    (occur-var globals)
@@ -73,10 +48,9 @@
 					  node))
 		   #unspecified))
 	     globals)
-   (verbose 2 "(removed : " *variable-removed* #\) #\newline)
+   (verbose 2 "      (removed : " *variable-removed* #\) #\newline)
    globals)
 
-
 ;*---------------------------------------------------------------------*/
 ;*    Statitics ...                                                    */
 ;*---------------------------------------------------------------------*/
@@ -151,14 +125,13 @@
 	 (values reset' node))))
 
 ;*---------------------------------------------------------------------*/
-;*    node-1occ! ::pragma ...                                          */
+;*    node-1occ! ::extern ...                                          */
 ;*    -------------------------------------------------------------    */
-;*    We don't remove single occurrences for pragma constructions      */
+;*    We don't remove single occurrences for extern constructions      */
 ;*    because it may introduce typing pbm.                             */
 ;*---------------------------------------------------------------------*/
-(define-method (node-1occ! node::pragma 1-exp*)
-   (with-access::pragma node (args)
-      (values (side-effect? node) node)))
+(define-method (node-1occ! node::extern 1-exp*)
+   (values (side-effect? node) node))
 
 ;*---------------------------------------------------------------------*/
 ;*    node-1occ! ::cast ...                                            */
@@ -257,12 +230,36 @@
 ;*    node-1occ! ::let-var ...                                         */
 ;*---------------------------------------------------------------------*/
 (define-method (node-1occ! node::let-var 1-exp*)
+   (with-access::let-var node (body bindings loc type removable?)
+      ;; we first the very special case
+      ;;    (let ((var expr)) (if var ... ...))
+      ;; this special case is important in order to get good
+      ;; compile-type error messages
+      (if (and (pair? bindings)
+	       (null? (cdr bindings))
+	       (let ((var (caar bindings)))
+		  (and (=fx (local-occurrence var) 1)
+		       (eq? (variable-type var) *bool*)
+		       (conditional? body)
+		       (with-access::conditional body (test)
+			  (and (var? test)
+			       (eq? (var-variable test) var))))))
+	  ;; yes
+	  (with-access::conditional body (test)
+	     (set! test (cdar bindings))
+	     (node-1occ! body 1-exp*))
+	  (node-1occ-let-var! node 1-exp*))))
+
+;*---------------------------------------------------------------------*/
+;*    node-1occ-let-var! ...                                           */
+;*---------------------------------------------------------------------*/
+(define (node-1occ-let-var! node::let-var 1-exp*)
    (with-access::let-var node (body bindings loc type removable?)
       (let loop ((obindings bindings)
 		 (reset     #f)
 		 (extend    '()))
 	 (if (null? obindings)
-	     (if reset
+	     (if (or reset (side-effect? body))
 		 (multiple-value-bind (reset' nbody)
 		    (node-1occ! body '())
 		    (set! body nbody)
@@ -316,26 +313,16 @@
 				extend))
 			 ((and (=fx (local-occurrence var) 1)
 			       (eq? (local-access var) 'read)
-			       (not (side-effect? val)))
+			       (not (side-effect? val))
+			       (type-less-specific? (local-type var)
+						    (typeof val)))
 			  (trace (reduce 3)
 				 "***1occ: applying: "
-				 (shape var)
-				 " "
-				 (shape val)
-				 #\Newline)
+				 (shape var) " " (shape val) #\Newline)
 			  (loop (cdr obindings)
 				#f
 				(cons binding extend)))
 			 (else
-			  (if (and (=fx (local-occurrence var) 1)
-				   (eq? (local-access var) 'read)
-				   (side-effect? val))
-			      (trace (reduce 3)
-				     "***1occ: NOT applying: "
-				     (shape var)
-				     " "
-				     (shape val)
-				     #\Newline))
 			  (loop (cdr obindings)
 				#f
 				extend))))))))))
@@ -383,7 +370,7 @@
 	 (node-1occ! value 1-exp*)
 	 (set! value nvalue)
 	 (values reset node))))
-
+ 
 ;*---------------------------------------------------------------------*/
 ;*    node-1occ! ::box-ref ...                                         */
 ;*---------------------------------------------------------------------*/
