@@ -1,68 +1,40 @@
-/* --------------------------------------------------------------------*/
-/*    Copyright (c) 1992-1998 by Manuel Serrano. All rights reserved.  */
-/*                                                                     */
-/*                                     ,--^,                           */
-/*                               _ ___/ /|/                            */
-/*                           ,;'( )__, ) '                             */
-/*                          ;;  //   L__.                              */
-/*                          '   \   /  '                               */
-/*                               ^   ^                                 */
-/*                                                                     */
-/*                                                                     */
-/*    This program is distributed in the hope that it will be useful.  */
-/*    Use and copying of this software and preparation of derivative   */
-/*    works based upon this software are permitted, so long as the     */
-/*    following conditions are met:                                    */
-/*           o credit to the authors is acknowledged following         */
-/*             current academic behaviour                              */
-/*           o no fees or compensation are charged for use, copies,    */
-/*             or access to this software                              */
-/*           o this copyright notice is included intact.               */
-/*      This software is made available AS IS, and no warranty is made */
-/*      about the software or its performance.                         */
-/*                                                                     */
-/*      Bug descriptions, use reports, comments or suggestions are     */
-/*      welcome. Send them to                                          */
-/*        Manuel Serrano -- Manuel.Serrano@unice.fr                    */
-/*-------------------------------------------------------------------- */
 /*=====================================================================*/
-/*    serrano/prgm/project/bigloo/runtime1.9b/Clib/cunicode.c          */
+/*    serrano/prgm/project/bigloo/runtime/Clib/cunicode.c              */
 /*    -------------------------------------------------------------    */
 /*    Author      :  Manuel Serrano                                    */
 /*    Creation    :  Mon May 19 17:47:11 1997                          */
-/*    Last change :  Thu Jun 12 16:12:56 1997 (serrano)                */
+/*    Last change :  Mon Jul  4 21:58:25 2005 (serrano)                */
 /*    -------------------------------------------------------------    */
 /*    Unicode strings handling                                         */
 /*=====================================================================*/
 #include <string.h>
-#include <bigloo1.9c.h>
+#include <bigloo.h>
 
 /*---------------------------------------------------------------------*/
 /*    Importations                                                     */
 /*---------------------------------------------------------------------*/
 extern bool_t ucs2_lower( ucs2_t );
-extern char *integer_to_string();
+extern obj_t integer_to_string();
 extern char *real_to_string();
 extern obj_t make_string();
+extern ucs2_t ucs2_tolower( ucs2_t );
 
 /*---------------------------------------------------------------------*/
 /*    ucs2_string_t                                                    */
 /*    make_ucs2_string ...                                             */
 /*---------------------------------------------------------------------*/
+BGL_RUNTIME_DEF
 obj_t
-make_ucs2_string( long len, ucs2_t c )
-{
-   obj_t   string , aux;
+make_ucs2_string( int len, ucs2_t c ) {
+   obj_t   string;
    ucs2_t *cstring;
 
-   if( len < 0 )
-      FAILURE( string_to_bstring( "make-ucs2-string" ),
-	       string_to_bstring( "Illegal string size" ),
-	       BINT( len ) );
-   {
-      long i;
+   if( len < 0 ) {
+      C_FAILURE( "make-ucs2-string", "Illegal string size", BINT( len ) );
+   } else {
+      int i;
       
-      string  = GC_MALLOC_ATOMIC( UCS2_STRING_SIZE + len );
+      string  = GC_MALLOC_ATOMIC( UCS2_STRING_SIZE + (len * sizeof(ucs2_t)) );
       cstring = (&(string->ucs2_string_t.char0));
 
       string->ucs2_string_t.header = MAKE_HEADER( UCS2_STRING_TYPE, 0 );
@@ -80,9 +52,8 @@ make_ucs2_string( long len, ucs2_t c )
 /*    void                                                             */
 /*    ucs2cpy ...                                                      */
 /*---------------------------------------------------------------------*/
-void
-ucs2cpy( ucs2_t *u1, ucs2_t *u2, long len )
-{
+static void
+ucs2cpy( ucs2_t *u1, ucs2_t *u2, int len ) {
    for( len--; len >= 0; len-- )
       u1[ len ] = u2[ len ];
 }
@@ -91,17 +62,14 @@ ucs2cpy( ucs2_t *u1, ucs2_t *u2, long len )
 /*    ucs2_string_append ...                                           */
 /*---------------------------------------------------------------------*/
 obj_t
-ucs2_string_append( obj_t s1, obj_t s2 )
-{
-   long    l1, l2;
+ucs2_string_append( obj_t s1, obj_t s2 ) {
+   int    l1, l2;
    obj_t   ucs2_string;
-   ucs2_t *cstring;
-   obj_t   a;
    
    l1 = UCS2_STRING( s1 ).length;
    l2 = UCS2_STRING( s2 ).length;
-	
-   ucs2_string = GC_MALLOC_ATOMIC( UCS2_STRING_SIZE + l1 + l2 );
+
+   ucs2_string = GC_MALLOC_ATOMIC( UCS2_STRING_SIZE + ((l1 + l2) * sizeof( ucs2_t )) );
 
    ucs2_string->ucs2_string_t.header = MAKE_HEADER( UCS2_STRING_TYPE, 0 );
    ucs2_string->ucs2_string_t.length = l1 + l2;
@@ -120,17 +88,15 @@ ucs2_string_append( obj_t s1, obj_t s2 )
 /*---------------------------------------------------------------------*/
 /*    c_subucs2_string ...                                             */
 /*---------------------------------------------------------------------*/
+BGL_RUNTIME_DEF
 obj_t
-c_subucs2_string( src_ucs2_string, min, max )
-obj_t src_ucs2_string;
-long  min, max;
-{
-   long  len;
-   obj_t dst_ucs2_string, a;
+c_subucs2_string( obj_t src_ucs2_string, int min, int max ) {
+   int  len;
+   obj_t dst_ucs2_string;
    
    len = max - min;
 
-   dst_ucs2_string = GC_MALLOC_ATOMIC( UCS2_STRING_SIZE + len );
+   dst_ucs2_string = GC_MALLOC_ATOMIC( UCS2_STRING_SIZE + (len * sizeof( ucs2_t ) ) );
 
    dst_ucs2_string->ucs2_string_t.header = MAKE_HEADER( UCS2_STRING_TYPE, 0 );
    dst_ucs2_string->ucs2_string_t.length = len;
@@ -147,15 +113,15 @@ long  min, max;
 /*    obj_t                                                            */
 /*    c_ucs2_string_copy ...                                           */
 /*---------------------------------------------------------------------*/
+BGL_RUNTIME_DEF
 obj_t
-c_ucs2_string_copy( obj_t src )
-{
-   long    len = UCS2_STRING_LENGTH( src );
-   obj_t   string , aux;
+c_ucs2_string_copy( obj_t src ) {
+   int    len = UCS2_STRING_LENGTH( src );
+   obj_t   string;
    ucs2_t *cstring, *cstr;
-   long    i;
+   int    i;
 
-   string  = GC_MALLOC_ATOMIC( UCS2_STRING_SIZE + len );
+   string  = GC_MALLOC_ATOMIC( UCS2_STRING_SIZE + (len * sizeof( ucs2_t ) ) );
    cstring = (&(string->ucs2_string_t.char0));
    cstr    = &UCS2_STRING_REF( src, 0 );
 
@@ -173,15 +139,14 @@ c_ucs2_string_copy( obj_t src )
 /*    ucs2_string_t                                                    */
 /*    string_to_ucs2_string ...                                        */
 /*---------------------------------------------------------------------*/
-obj_t
-string_to_ucs2_string( char *c )
-{
-   long    len = strlen( c );
-   obj_t   string , aux;
+BGL_RUNTIME_DEF obj_t
+string_to_ucs2_string( char *c ) {
+   int     len = (int)strlen( c );
+   obj_t   string;
    ucs2_t *cstring;
-   long    i;
+   int     i;
 
-   string  = GC_MALLOC_ATOMIC( UCS2_STRING_SIZE + len );
+   string  = GC_MALLOC_ATOMIC( UCS2_STRING_SIZE + (len * sizeof( ucs2_t ) ) );
    cstring = (&(string->ucs2_string_t.char0));
 
    string->ucs2_string_t.header = MAKE_HEADER( UCS2_STRING_TYPE, 0 );
@@ -199,15 +164,14 @@ string_to_ucs2_string( char *c )
 /*    bstring_to_ucs2_string ...                                       */
 /*---------------------------------------------------------------------*/
 obj_t
-bstring_to_ucs2_string( obj_t src )
-{
-   long    len = STRING_LENGTH( src );
-   obj_t   string , aux;
+bstring_to_ucs2_string( obj_t src ) {
+   int    len = STRING_LENGTH( src );
+   obj_t   string;
    ucs2_t *cstring;
-   long    i;
-   char   *c = (&(src->string_t.char0));
+   int    i;
+   char   *c = (char *)(&(src->string_t.char0));
 
-   string  = GC_MALLOC_ATOMIC( UCS2_STRING_SIZE + len );
+   string  = GC_MALLOC_ATOMIC( UCS2_STRING_SIZE + (len * sizeof( ucs2_t ) ) );
    cstring = (&(string->ucs2_string_t.char0));
 
    string->ucs2_string_t.header = MAKE_HEADER( UCS2_STRING_TYPE, 0 );
@@ -224,37 +188,31 @@ bstring_to_ucs2_string( obj_t src )
 /*    integer_to_ucs2_string ...                                       */
 /*---------------------------------------------------------------------*/
 obj_t
-integer_to_ucs2_string( x, radix )
-long x, radix;
-{
-   return string_to_ucs2_string( integer_to_string( x, radix ) );
+integer_to_ucs2_string( long x, long radix ) {
+   return string_to_ucs2_string( BSTRING_TO_STRING( integer_to_string( x, radix ) ) );
 }
 
 /*---------------------------------------------------------------------*/
 /*    real_to_ucs2_string ...                                          */
 /*---------------------------------------------------------------------*/
 obj_t
-real_to_ucs2_string( x )
-double x;
-{
-   return string_to_ucs2_string( real_to_string( x ) );
+real_to_ucs2_string( double x ) {
+   return string_to_ucs2_string( BSTRING_TO_STRING( real_to_string( x ) ) );
 }
 
 /*---------------------------------------------------------------------*/
 /*    bool_t                                                           */
 /*    ucs2_strcmp ...                                                  */
 /*---------------------------------------------------------------------*/
+BGL_RUNTIME_DEF
 bool_t
-ucs2_strcmp( o1, o2 )
-obj_t o1, o2;
-{
-   long l1, l2;
+ucs2_strcmp( obj_t o1, obj_t o2 ) {
+   int l1, l2;
    
    l1 = UCS2_STRING_LENGTH( o1 );
    l2 = UCS2_STRING_LENGTH( o2 );
 
-   if( l1 == l2 )
-   {
+   if( l1 == l2 ) {
       ucs2_t *c1, *c2;
    
       c1 = BUCS2_STRING_TO_UCS2_STRING( o1 );
@@ -265,28 +223,26 @@ obj_t o1, o2;
 	    return 0;
 
       return 1;
-   }
-   else
+   } else {
       return 0;
+   }
 }
 
 /*---------------------------------------------------------------------*/
 /*    ucs2_strcicmp ...                                                */
 /*---------------------------------------------------------------------*/
+BGL_RUNTIME_DEF
 bool_t        
-ucs2_strcicmp( bst1, bst2 )
-obj_t bst1, bst2;
-{
-   long l1, l2;
+ucs2_strcicmp( obj_t bst1, obj_t bst2 ) {
+   int l1, l2;
 
    l1 = UCS2_STRING_LENGTH( bst1 );
    l2 = UCS2_STRING_LENGTH( bst2 );
 
-   if( l1 == l2 )
-   {
+   if( l1 == l2 ) {
       ucs2_t *st1 = BUCS2_STRING_TO_UCS2_STRING( bst1 );
       ucs2_t *st2 = BUCS2_STRING_TO_UCS2_STRING( bst2 );
-      long i;
+      int i;
    
       for( i = 0;
 	   ucs2_tolower( *st1 ) == ucs2_tolower( *st2 );
@@ -301,14 +257,14 @@ obj_t bst1, bst2;
 /*---------------------------------------------------------------------*/
 /*    ucs2_string_lt ...                                               */
 /*---------------------------------------------------------------------*/
+BGL_RUNTIME_DEF
 bool_t       
-ucs2_string_lt( bst1, bst2 )
-obj_t bst1, bst2;
+ucs2_string_lt( obj_t bst1, obj_t bst2 )
 {
    ucs2_t *st1 = BUCS2_STRING_TO_UCS2_STRING( bst1 );
    ucs2_t *st2 = BUCS2_STRING_TO_UCS2_STRING( bst2 );
-   long l1, l2;
-   long i, min;
+   int l1, l2;
+   int i, min;
 
    l1 = UCS2_STRING_LENGTH( bst1 );
    l2 = UCS2_STRING_LENGTH( bst2 );
@@ -326,14 +282,13 @@ obj_t bst1, bst2;
 /*---------------------------------------------------------------------*/
 /*    ucs2_string_le ...                                               */
 /*---------------------------------------------------------------------*/
+BGL_RUNTIME_DEF
 bool_t        
-ucs2_string_le( bst1, bst2 )
-obj_t bst1, bst2;
-{
+ucs2_string_le( obj_t bst1, obj_t bst2 ) {
    ucs2_t *st1 = BUCS2_STRING_TO_UCS2_STRING( bst1 );
    ucs2_t *st2 = BUCS2_STRING_TO_UCS2_STRING( bst2 );
-   long l1, l2;
-   long i, min;
+   int l1, l2;
+   int i, min;
 
    l1 = UCS2_STRING_LENGTH( bst1 );
    l2 = UCS2_STRING_LENGTH( bst2 );
@@ -351,14 +306,13 @@ obj_t bst1, bst2;
 /*---------------------------------------------------------------------*/
 /*    ucs2_string_gt ...                                               */
 /*---------------------------------------------------------------------*/
+BGL_RUNTIME_DEF
 bool_t       
-ucs2_string_gt( bst1, bst2 )
-obj_t bst1, bst2;
-{
+ucs2_string_gt( obj_t bst1, obj_t bst2 ) {
    ucs2_t *st1 = BUCS2_STRING_TO_UCS2_STRING( bst1 );
    ucs2_t *st2 = BUCS2_STRING_TO_UCS2_STRING( bst2 );
-   long l1, l2;
-   long i, min;
+   int l1, l2;
+   int i, min;
 
    l1 = UCS2_STRING_LENGTH( bst1 );
    l2 = UCS2_STRING_LENGTH( bst2 );
@@ -376,14 +330,13 @@ obj_t bst1, bst2;
 /*---------------------------------------------------------------------*/
 /*    ucs2_string_ge ...                                               */
 /*---------------------------------------------------------------------*/
+BGL_RUNTIME_DEF
 bool_t        
-ucs2_string_ge( bst1, bst2 )
-obj_t bst1, bst2;
-{
+ucs2_string_ge( obj_t bst1, obj_t bst2 ) {
    ucs2_t *st1 = BUCS2_STRING_TO_UCS2_STRING( bst1 );
    ucs2_t *st2 = BUCS2_STRING_TO_UCS2_STRING( bst2 );
-   long l1, l2;
-   long i, min;
+   int l1, l2;
+   int i, min;
 
    l1 = UCS2_STRING_LENGTH( bst1 );
    l2 = UCS2_STRING_LENGTH( bst2 );
@@ -401,14 +354,13 @@ obj_t bst1, bst2;
 /*---------------------------------------------------------------------*/
 /*    ucs2_string_cilt ...                                             */
 /*---------------------------------------------------------------------*/
+BGL_RUNTIME_DEF
 bool_t       
-ucs2_string_cilt( bst1, bst2 )
-obj_t bst1, bst2;
-{
+ucs2_string_cilt( obj_t bst1, obj_t bst2 ) {
    ucs2_t *st1 = BUCS2_STRING_TO_UCS2_STRING( bst1 );
    ucs2_t *st2 = BUCS2_STRING_TO_UCS2_STRING( bst2 );
-   long l1, l2;
-   long i, min;
+   int l1, l2;
+   int i, min;
 
    l1 = UCS2_STRING_LENGTH( bst1 );
    l2 = UCS2_STRING_LENGTH( bst2 );
@@ -428,14 +380,13 @@ obj_t bst1, bst2;
 /*---------------------------------------------------------------------*/
 /*    ucs2_string_cile ...                                             */
 /*---------------------------------------------------------------------*/
+BGL_RUNTIME_DEF
 bool_t        
-ucs2_string_cile( bst1, bst2 )
-obj_t bst1, bst2;
-{
+ucs2_string_cile( obj_t bst1, obj_t bst2 ) {
    ucs2_t *st1 = BUCS2_STRING_TO_UCS2_STRING( bst1 );
    ucs2_t *st2 = BUCS2_STRING_TO_UCS2_STRING( bst2 );
-   long l1, l2;
-   long i, min;
+   int l1, l2;
+   int i, min;
 
    l1 = UCS2_STRING_LENGTH( bst1 );
    l2 = UCS2_STRING_LENGTH( bst2 );
@@ -455,14 +406,13 @@ obj_t bst1, bst2;
 /*---------------------------------------------------------------------*/
 /*    ucs2_string_cigt ...                                             */
 /*---------------------------------------------------------------------*/
+BGL_RUNTIME_DEF
 bool_t       
-ucs2_string_cigt( bst1, bst2 )
-obj_t bst1, bst2;
-{
+ucs2_string_cigt( obj_t bst1, obj_t bst2 ) {
    ucs2_t *st1 = BUCS2_STRING_TO_UCS2_STRING( bst1 );
    ucs2_t *st2 = BUCS2_STRING_TO_UCS2_STRING( bst2 );
-   long l1, l2;
-   long i, min;
+   int l1, l2;
+   int i, min;
 
    l1 = UCS2_STRING_LENGTH( bst1 );
    l2 = UCS2_STRING_LENGTH( bst2 );
@@ -482,14 +432,13 @@ obj_t bst1, bst2;
 /*---------------------------------------------------------------------*/
 /*    ucs2_string_cige ...                                             */
 /*---------------------------------------------------------------------*/
+BGL_RUNTIME_DEF
 bool_t        
-ucs2_string_cige( bst1, bst2 )
-obj_t bst1, bst2;
-{
+ucs2_string_cige( obj_t bst1, obj_t bst2 ) {
    ucs2_t *st1 = BUCS2_STRING_TO_UCS2_STRING( bst1 );
    ucs2_t *st2 = BUCS2_STRING_TO_UCS2_STRING( bst2 );
-   long l1, l2;
-   long i, min;
+   int l1, l2;
+   int i, min;
 
    l1 = UCS2_STRING_LENGTH( bst1 );
    l2 = UCS2_STRING_LENGTH( bst2 );
@@ -506,85 +455,12 @@ obj_t bst1, bst2;
       return l1 >= l2;
 }
 
-/*---------------------------------------------------------------------*/
-/*    obj_t                                                            */
-/*    display_ucs2string ...                                           */
-/*---------------------------------------------------------------------*/
-obj_t
-display_ucs2string( obj_t obj, obj_t port )
-{
-   long    len  = UCS2_STRING_LENGTH( obj );
-   ucs2_t *ucs2 = BUCS2_STRING_TO_UCS2_STRING( obj );
-   long    i;
-   char    aux[ 10 ];
-   
-   if( OUTPUT_STRING_PORTP( port ) )
-   {
-      for( i = 0; i < len; i++ )
-      {
-	 ucs2_t ch = ucs2[ i ];
-	 
-#if( UCS2_DISPLAYABLE )
-#else
-	 if( UCS2_ISOLATIN1P( ch ) )
-	    strputc( ch, port );
-#endif
-      }
-   }
-   else
-   {
-      FILE *fout = OUTPUT_PORT( port ).file;
-      
-      for( i = 0; i < len; i++ )
-      {
-	 ucs2_t ch = ucs2[ i ];
-	 
-#if( UCS2_DISPLAYABLE )
-#else
-	 if( UCS2_ISOLATIN1P( ch ) )
-	    fputc( ch, fout );
-#endif
-      }
-   }
-   return obj;
-}
-
-/*---------------------------------------------------------------------*/
-/*    obj_t                                                            */
-/*    write_utf8string ...                                             */
-/*---------------------------------------------------------------------*/
-obj_t
-write_utf8string( obj_t string, obj_t port )
-{
-   char *aux = BSTRING_TO_STRING( string );
-   long  len = STRING_LENGTH( string );
-   
-   if( OUTPUT_STRING_PORTP( port ) )
-   {
-      strputs( "#u\"", port );
-      lstrputs( aux, port, len );
-      strputc( '"', port );
-   }
-   else
-   {
-      FILE *fout = OUTPUT_PORT( port ).file;
-
-      fprintf( fout, "#u\"" );
-      fwrite( aux, 1, len, fout );
-      fputc( '"', fout );
-      return string;
-   }
-
-   return string;
-}
-
 /*---------------------------------------------------------------------*/
 /*    static int                                                       */
 /*    utf8_size ...                                                    */
 /*---------------------------------------------------------------------*/
 static int
-utf8_size( ucs2_t ucs2 )
-{
+utf8_size( ucs2_t ucs2 ) {
    if( ucs2 <= 0x7f )
       return 1;
    if( ucs2 <= 0x7ff )
@@ -598,21 +474,23 @@ utf8_size( ucs2_t ucs2 )
       return 3;
    
    C_FAILURE( "utf8_size", "Illegal ucs2 character", BUCS2( ucs2 ) );
+
+   return -1;
 }
 
 /*---------------------------------------------------------------------*/
 /*    obj_t                                                            */
 /*    ucs2_string_to_utf8_string ...                                   */
 /*---------------------------------------------------------------------*/
+BGL_RUNTIME_DEF
 obj_t
-ucs2_string_to_utf8_string( obj_t bucs2 )
-{
-   long           len      = UCS2_STRING_LENGTH( bucs2 );
-   long           utf8_len = 0;
-   ucs2_t        *cucs2    = BUCS2_STRING_TO_UCS2_STRING( bucs2 );
-   long           read, write;
-   obj_t          result;
-   unsigned char *cresult;
+ucs2_string_to_utf8_string( obj_t bucs2 ) {
+   int len = UCS2_STRING_LENGTH( bucs2 );
+   int utf8_len = 0;
+   ucs2_t *cucs2 = BUCS2_STRING_TO_UCS2_STRING( bucs2 );
+   int read, write;
+   obj_t result;
+   char *cresult;
    
    /* First, we compute the size of the futur utf8 string */
    for( read = 0, utf8_len = 0; read < len; read++ )
@@ -620,20 +498,17 @@ ucs2_string_to_utf8_string( obj_t bucs2 )
 
    /* now, we allocate the new string */
    result  = make_string( utf8_len, '0' );
-   cresult = (unsigned char *)BSTRING_TO_STRING( result );
+   cresult = BSTRING_TO_STRING( result );
    
    /* and we fill it */
-   for( read = 0, write = 0; read < len; read++ )
-   {
+   for( read = 0, write = 0; read < len; read++ ) {
       ucs2_t ucs2 = cucs2[ read ];
       int    len  = utf8_size( ucs2 );
 
       if( len == 1 )
 	 cresult[ write++ ] = (unsigned char)ucs2;
-      else
-      {
-	 if( len == 3 )
-	 {
+      else {
+	 if( len == 3 ) {
 	    cresult[ write + 2 ] = (unsigned char)(0x80 + (ucs2 & 0x3f));
 	    ucs2 >>= 6;
 	 }
@@ -653,39 +528,34 @@ ucs2_string_to_utf8_string( obj_t bucs2 )
 /*    obj_t                                                            */
 /*    utf8_string_to_ucs2_string ...                                   */
 /*---------------------------------------------------------------------*/
+BGL_RUNTIME_DEF
 obj_t
-utf8_string_to_ucs2_string( obj_t butf8 )
-{
-   long           len = STRING_LENGTH( butf8 );
-   long           nbchar = 0;
-   ucs2_t        *aux = (ucs2_t *)alloca( len * sizeof( ucs2_t ) );
-   unsigned char *cutf8 = (unsigned char *)BSTRING_TO_STRING( butf8 );
-   long           read, write;
-   obj_t          string;
-   ucs2_t        *cstring;
+utf8_string_to_ucs2_string( obj_t butf8 ) {
+   int len = STRING_LENGTH( butf8 );
+   ucs2_t *aux = (ucs2_t *)alloca( len * sizeof( ucs2_t ) );
+   char *cutf8 = BSTRING_TO_STRING( butf8 );
+   int read, write;
+   obj_t string;
+   ucs2_t *cstring;
    
-   for( read = 0, write = 0; read < len; write++ )
-   {
+   for( read = 0, write = 0; read < len; write++ ) {
       unsigned char byte = cutf8[ read++ ];
 
       if( byte <= 0x7f )
 	 aux[ write ] = (ucs2_t)byte;
-      else
-      {
+      else {
 	 if( (byte <= 0xbf) || (byte >= 0xfd) )
 	    C_FAILURE( "utf8-string->ucs2-string",
 		       "Illegal first byte",
 		       BCHAR( byte ) );
-	 else
-	 {
+	 else {
 	    ucs2_t ucs2;
-	    int    bits;
+	    int bits;
 	    
 	    ucs2 = (ucs2_t)byte;
 	    bits = 6;
 
-	    while( byte & 0x40 )
-	    {
+	    while( byte & 0x40 ) {
 	       unsigned char next = cutf8[ read++ ];
 	       
 	       if( (next <= 0x7f) || (next > 0xbf) )
@@ -700,7 +570,7 @@ utf8_string_to_ucs2_string( obj_t butf8 )
 	    ucs2 &= ((ucs2_t)1<<bits) - 1;
 
 	    if( (ucs2 > 0xd7ff && ucs2 <= 0xdfff) ||
-		(ucs2 > 0xfffd && ucs2 <= 0xffff) ||
+		(ucs2 > 0xfffd) ||
 		!(ucs2 & (~(unsigned long)0<<(bits - 5))) )
 	       C_FAILURE( "utf8-string->ucs2-string",
 			  "Illegal utf8 character encoding",
@@ -711,7 +581,7 @@ utf8_string_to_ucs2_string( obj_t butf8 )
       }
    }
 	 
-   string  = GC_MALLOC_ATOMIC( UCS2_STRING_SIZE + len );
+   string  = GC_MALLOC_ATOMIC( UCS2_STRING_SIZE + (len * sizeof( ucs2_t ) ) );
    cstring = (&(string->ucs2_string_t.char0));
 
    string->ucs2_string_t.header = MAKE_HEADER( UCS2_STRING_TYPE, 0 );
@@ -720,4 +590,3 @@ utf8_string_to_ucs2_string( obj_t butf8 )
 		
    return BUCS2STRING( string );
 }
-   
