@@ -1,36 +1,9 @@
-;* --------------------------------------------------------------------*/
-;*    Copyright (c) 1992-1998 by Manuel Serrano. All rights reserved.  */
-;*                                                                     */
-;*                                     ,--^,                           */
-;*                               _ ___/ /|/                            */
-;*                           ,;'( )__, ) '                             */
-;*                          ;;  //   L__.                              */
-;*                          '   \   /  '                               */
-;*                               ^   ^                                 */
-;*                                                                     */
-;*                                                                     */
-;*    This program is distributed in the hope that it will be useful.  */
-;*    Use and copying of this software and preparation of derivative   */
-;*    works based upon this software are permitted, so long as the     */
-;*    following conditions are met:                                    */
-;*           o credit to the authors is acknowledged following         */
-;*             current academic behaviour                              */
-;*           o no fees or compensation are charged for use, copies,    */
-;*             or access to this software                              */
-;*           o this copyright notice is included intact.               */
-;*      This software is made available AS IS, and no warranty is made */
-;*      about the software or its performance.                         */
-;*                                                                     */
-;*      Bug descriptions, use reports, comments or suggestions are     */
-;*      welcome. Send them to                                          */
-;*        Manuel Serrano -- Manuel.Serrano@unice.fr                    */
-;*-------------------------------------------------------------------- */
 ;*=====================================================================*/
-;*    serrano/prgm/project/bigloo/runtime1.9b/Ieee/flonum.scm          */
+;*    serrano/prgm/project/bigloo/runtime/Ieee/flonum.scm              */
 ;*    -------------------------------------------------------------    */
 ;*    Author      :  Manuel Serrano                                    */
 ;*    Creation    :  Thu Nov 26 14:04:03 1992                          */
-;*    Last change :  Fri Aug  1 10:53:39 1997 (serrano)                */
+;*    Last change :  Tue Oct 18 15:25:04 2005 (serrano)                */
 ;*    -------------------------------------------------------------    */
 ;*    6.5. Numbers (page 18, r4) The `flonum' functions                */
 ;*=====================================================================*/
@@ -39,132 +12,204 @@
 ;*    The module                                                       */
 ;*---------------------------------------------------------------------*/
 (module __r4_numbers_6_5_flonum
-
-   (import  (__error                   "Llib/error.scm"))
    
-   (use     (__type                    "Llib/type.scm")
-	    (__bigloo                  "Llib/bigloo.scm")
-	    (__tvector                 "Llib/tvector.scm")
-	    (__r4_booleans_6_1         "Ieee/boolean.scm")
-	    (__r4_vectors_6_8          "Ieee/vector.scm")
-	    (__r4_strings_6_7          "Ieee/string.scm")
-	    (__r4_characters_6_6       "Ieee/char.scm")
-	    (__r4_pairs_and_lists_6_3  "Ieee/pair-list.scm")
-	    (__r4_symbols_6_4          "Ieee/symbol.scm")
-	    (__r4_numbers_6_5_fixnum   "Ieee/fixnum.scm")
+   (import  __error)
+   
+   (use     __type
+	    __bigloo
+	    __tvector
+	    __r4_booleans_6_1
+	    __r4_vectors_6_8
+	    __r4_strings_6_7
+	    __r4_characters_6_6
+	    __r4_pairs_and_lists_6_3
+	    __r4_symbols_6_4
+	    __r4_numbers_6_5_fixnum
 	    
-	    (__evenv                   "Eval/evenv.scm"))
-
-   (foreign (macro bool       c-flonum?   (obj)           "REALP")
-	    (infix macro bool c-=fl       (double double) "==")
-	    (infix macro bool c-<fl       (double double) "<")
-	    (infix macro bool c-<=fl      (double double) "<=")
-	    (infix macro bool c->fl       (double double) ">")
-	    (infix macro bool c->=fl      (double double) ">=")
-	    (infix macro double c-+fl     (double double) "+")
-	    (infix macro double c--fl     (double double) "-")
-	    (infix macro double c-*fl     (double double) "*")
-	    (infix macro double c-/fl     (double double) "/")
-	    (macro double       c-negfl   (double)        "NEG")
-	    (macro double c-floor         (double)        "floor")
-	    (macro double c-ceiling       (double)        "ceil")
-	    (macro double c-exp           (double)        "exp")
-	    (macro double c-log           (double)        "log")
-	    (macro double c-sin           (double)        "sin")
-	    (macro double c-cos           (double)        "cos")
-	    (macro double c-tan           (double)        "tan")
-	    (macro double c-asin          (double)        "asin")
-	    (macro double c-acos          (double)        "acos")
-	    (macro double c-atan          (double)        "atan")
-	    (macro double c-atan2         (double double) "atan2")
-	    (macro double c-sqrt          (double)        "sqrt")
-	    (macro double c-pow           (double double) "pow")
-	    (string       c-real->string  (double)        "real_to_string")
-	    (macro double strtod          (string long)   "strtod"))
+	    __evenv)
    
-   (export  (inline real?::bool          ::obj)
-	    (inline flonum?::bool        ::obj)
-	    (inline =fl::bool            ::double ::double)
-	    (inline >fl::bool            ::double ::double)
-	    (inline >=fl::bool           ::double ::double)
-	    (inline <fl::bool            ::double ::double)
-	    (inline <=fl::bool           ::double ::double)
-	    (inline zerofl?::bool        ::double)
-	    (inline positivefl?::bool    ::double)
-	    (inline negativefl?::bool    ::double)
-	    (maxfl::double               ::double . rn)
-	    (minfl::double               ::double . rn)
-	    (inline max-2fl::double      ::double ::double)
-	    (inline min-2fl::double      ::double ::double)
-	    (inline +fl::double          ::double ::double)
-	    (inline -fl::double          ::double ::double)
-	    (inline *fl::double          ::double ::double)
-	    (inline /fl::double          ::double ::double)
-	    (inline negfl::double        ::double)
-	    (inline absfl::double        ::double)
-	    (inline floorfl::double      ::double)
-	    (inline ceilingfl::double    ::double)
-	    (inline truncatefl::double   ::double)
-	    (inline roundfl::double      ::double)
-	    (inline expfl::double        ::double)
-	    (inline logfl::double        ::double)
-	    (inline sinfl::double        ::double)
-	    (inline cosfl::double        ::double)
-	    (inline tanfl::double        ::double)
-	    (inline asinfl::double       ::double)
-	    (inline acosfl::double       ::double)
-	    (atanfl::double              ::double . y)
-	    (inline atan-1fl::double     ::double)
-	    (inline atan-2fl::double     ::double ::double)
-	    (inline atan-2fl-ur::double  ::double ::double)
-	    (inline sqrtfl::double       ::double)
-	    (inline sqrtfl-ur::double    ::double)
-	    (inline exptfl::double       ::double ::double)
+   (extern  (macro c-flonum?::bool (::obj) "REALP")
+	    (infix macro c-=fl::bool (::double ::double) "==")
+	    (infix macro c-<fl::bool (::double ::double) "<")
+	    (infix macro c-<=fl::bool (::double ::double) "<=")
+	    (infix macro c->fl::bool (::double ::double) ">")
+	    (infix macro c->=fl::bool (::double ::double) ">=")
+	    (infix macro c-+fl::double (::double ::double) "+")
+	    (infix macro c--fl::double (::double ::double) "-")
+	    (infix macro c-*fl::double (::double ::double) "*")
+	    (infix macro c-/fl::double (::double ::double) "/")
+	    (macro c-negfl::double (::double) "NEG")
+	    (macro c-floor::double (::double) "floor")
+	    (macro c-ceiling::double (::double) "ceil")
+	    (macro c-fmod::double (::double ::double) "fmod")
+	    (macro c-exp::double (::double) "exp")
+	    (macro c-log::double (::double) "log")
+	    (macro c-sin::double (::double) "sin")
+	    (macro c-cos::double (::double) "cos")
+	    (macro c-tan::double (::double) "tan")
+	    (macro c-asin::double (::double) "asin")
+	    (macro c-acos::double (::double) "acos")
+	    (macro c-atan::double (::double) "atan")
+	    (macro c-atan2::double (::double ::double) "atan2")
+	    (macro c-sqrt::double (::double) "sqrt")
+	    (macro c-pow::double (::double ::double) "pow")
+	    (macro strtod::double (::string ::long) "strtod")
+	    (c-real->string::bstring (::double) "real_to_string")
+	    (%ieee-string->double::double (::bstring)
+					  "bgl_ieee_string_to_double")
+	    (%double->ieee-string::bstring (::double)
+					   "bgl_double_to_ieee_string"))
+   
+   (java    (class foreign
+	       (method static c-flonum?::bool (::obj)
+		       "REALP")
+	       (method static c-=fl::bool (::double ::double)
+		       "EQ_FL")
+	       (method static c-<fl::bool (::double ::double)
+		       "LT_FL")
+	       (method static c-<=fl::bool (::double ::double)
+		       "LE_FL")
+	       (method static c->fl::bool (::double ::double)
+		       "GT_FL")
+	       (method static c->=fl::bool (::double ::double)
+		       "GE_FL")
+	       (method static c-+fl::double (::double ::double)
+		       "PLUS_FL")
+	       (method static c--fl::double (::double ::double)
+		       "MINUS_FL")
+	       (method static c-*fl::double (::double ::double)
+		       "MUL_FL")
+	       (method static c-/fl::double (::double ::double)
+		       "DIV_FL")
+	       (method static c-negfl::double (::double)
+		       "NEG_FL")
+	       (method static c-fmod::double (::double ::double)
+		       "fmod")
+	       (method static c-floor::double (::double)
+		       "floor")
+	       (method static c-ceiling::double (::double)
+		       "ceil")
+	       (method static c-exp::double (::double)
+		       "exp")
+	       (method static c-log::double (::double)
+		       "log")
+	       (method static c-sin::double (::double)
+		       "sin")
+	       (method static c-cos::double (::double)
+		       "cos")
+	       (method static c-tan::double (::double)
+		       "tan")
+	       (method static c-asin::double (::double)
+		       "asin")
+	       (method static c-acos::double (::double)
+		       "acos")
+	       (method static c-atan::double (::double)
+		       "atan")
+	       (method static c-atan2::double (::double ::double)
+		       "atan2")
+	       (method static c-sqrt::double (::double)
+		       "sqrt")
+	       (method static c-pow::double (::double ::double)
+		       "pow")
+	       (method static strtod::double (::string ::long)
+		       "strtod")
+	       (method static c-real->string::bstring (::double)
+		       "real_to_string")
+	       (method static %ieee-string->double::double (::bstring)
+		       "bgl_ieee_string_to_double")
+	       (method static %double->ieee-string::bstring (::double)
+		       "bgl_double_to_ieee_string")))
+   
+   (export  (inline real?::bool ::obj)
+	    (inline flonum?::bool ::obj)
+	    (inline =fl::bool ::double ::double)
+	    (inline >fl::bool ::double ::double)
+	    (inline >=fl::bool ::double ::double)
+	    (inline <fl::bool ::double ::double)
+	    (inline <=fl::bool ::double ::double)
+	    (inline zerofl?::bool ::double)
+	    (inline positivefl?::bool ::double)
+	    (inline negativefl?::bool ::double)
+	    (maxfl::double ::double . rn)
+	    (minfl::double ::double . rn)
+	    (inline max-2fl::double ::double ::double)
+	    (inline min-2fl::double ::double ::double)
+	    (inline +fl::double ::double ::double)
+	    (inline -fl::double ::double ::double)
+	    (inline *fl::double ::double ::double)
+	    (inline /fl::double ::double ::double)
+	    (inline negfl::double ::double)
+	    (inline absfl::double ::double)
+	    (inline floorfl::double ::double)
+	    (inline ceilingfl::double ::double)
+	    (inline truncatefl::double ::double)
+	    (roundfl::double ::double)
+	    (inline remainderfl::double ::double ::double)
+	    (inline expfl::double ::double)
+	    (inline logfl::double ::double)
+	    (inline sinfl::double ::double)
+	    (inline cosfl::double ::double)
+	    (inline tanfl::double ::double)
+	    (inline asinfl::double ::double)
+	    (inline acosfl::double ::double)
+	    (atanfl::double ::double . y)
+	    (inline atan-1fl::double ::double)
+	    (inline atan-2fl::double ::double ::double)
+	    (inline atan-2fl-ur::double ::double ::double)
+	    (inline sqrtfl::double ::double)
+	    (inline sqrtfl-ur::double ::double)
+	    (inline exptfl::double ::double ::double)
 	    (inline string->real::double ::string)
-	    (inline real->string::string ::double))
-
-   (pragma  (c-flonum? (predicate-of real) no-cfa-top)
-	    (real? side-effect-free no-cfa-top)
-	    (c-=fl side-effect-free no-cfa-top)
-	    (c->fl side-effect-free no-cfa-top)
-	    (c->=fl side-effect-free no-cfa-top)
-	    (c-<fl side-effect-free no-cfa-top)
-	    (c-<=fl side-effect-free no-cfa-top)
-	    (c-+fl side-effect-free no-cfa-top)
-	    (c--fl side-effect-free no-cfa-top)
-	    (c-*fl side-effect-free no-cfa-top)
-	    (c-/fl side-effect-free no-cfa-top)
-	    (c-negfl side-effect-free no-cfa-top)
-	    (c-exp side-effect-free no-cfa-top)
-	    (c-log side-effect-free no-cfa-top)
-	    (c-sin side-effect-free no-cfa-top)
-	    (c-cos side-effect-free no-cfa-top)
-	    (c-tan side-effect-free no-cfa-top)
-	    (c-asin side-effect-free no-cfa-top)
-	    (c-acos side-effect-free no-cfa-top)
-	    (c-atan side-effect-free no-cfa-top)
-	    (c-sqrt side-effect-free no-cfa-top)
-	    (flonum? (predicate-of real) no-cfa-top)
-	    (real? side-effect-free no-cfa-top)
-	    (=fl side-effect-free no-cfa-top)
-	    (>fl side-effect-free no-cfa-top)
-	    (>=fl side-effect-free no-cfa-top)
-	    (<fl side-effect-free no-cfa-top)
-	    (<=fl side-effect-free no-cfa-top)
-	    (+fl side-effect-free no-cfa-top)
-	    (-fl side-effect-free no-cfa-top)
-	    (*fl side-effect-free no-cfa-top)
-	    (/fl side-effect-free no-cfa-top)
-	    (negfl side-effect-free no-cfa-top)
-	    (expfl side-effect-free no-cfa-top)
-	    (logfl side-effect-free no-cfa-top)
-	    (sinfl side-effect-free no-cfa-top)
-	    (cosfl side-effect-free no-cfa-top)
-	    (tanfl side-effect-free no-cfa-top)
-	    (asinfl side-effect-free no-cfa-top)
-	    (acosfl side-effect-free no-cfa-top)
-	    (atanfl side-effect-free no-cfa-top)
-	    (sqrtfl side-effect-free no-cfa-top)))
+	    (inline real->string::bstring ::real)
+	    (inline ieee-string->real::real ::bstring)
+	    (inline real->ieee-string::bstring ::real)
+	    (inline ieee-string->double::double ::bstring)
+	    (inline double->ieee-string::bstring ::double)
+	    (inline ieee-string->float::float ::bstring)
+	    (inline float->ieee-string::bstring ::float))
+   
+   (pragma  (c-flonum? side-effect-free (predicate-of real) no-cfa-top nesting)
+	    (real? side-effect-free no-cfa-top nesting)
+	    (c-=fl side-effect-free no-cfa-top nesting args-safe)
+	    (c->fl side-effect-free no-cfa-top nesting args-safe)
+	    (c->=fl side-effect-free no-cfa-top nesting args-safe)
+	    (c-<fl side-effect-free no-cfa-top nesting args-safe)
+	    (c-<=fl side-effect-free no-cfa-top nesting args-safe)
+	    (c-+fl side-effect-free no-cfa-top nesting args-safe)
+	    (c--fl side-effect-free no-cfa-top nesting args-safe)
+	    (c-*fl side-effect-free no-cfa-top nesting args-safe)
+	    (c-/fl side-effect-free no-cfa-top nesting args-safe)
+	    (c-negfl side-effect-free no-cfa-top nesting args-safe)
+	    (c-exp side-effect-free no-cfa-top nesting args-safe)
+	    (c-log side-effect-free no-cfa-top nesting args-safe)
+	    (c-sin side-effect-free no-cfa-top nesting args-safe)
+	    (c-cos side-effect-free no-cfa-top nesting args-safe)
+	    (c-tan side-effect-free no-cfa-top nesting args-safe)
+	    (c-asin side-effect-free no-cfa-top nesting args-safe)
+	    (c-acos side-effect-free no-cfa-top nesting args-safe)
+	    (c-atan side-effect-free no-cfa-top nesting args-safe)
+	    (c-sqrt side-effect-free no-cfa-top nesting args-safe)
+	    (flonum? (predicate-of real) no-cfa-top nesting)
+	    (real? side-effect-free no-cfa-top nesting)
+	    (=fl side-effect-free no-cfa-top nesting)
+	    (>fl side-effect-free no-cfa-top nesting)
+	    (>=fl side-effect-free no-cfa-top nesting)
+	    (<fl side-effect-free no-cfa-top nesting)
+	    (<=fl side-effect-free no-cfa-top nesting)
+	    (+fl side-effect-free no-cfa-top nesting)
+	    (-fl side-effect-free no-cfa-top nesting)
+	    (*fl side-effect-free no-cfa-top nesting)
+	    (/fl side-effect-free no-cfa-top nesting)
+	    (negfl side-effect-free no-cfa-top nesting)
+	    (expfl side-effect-free no-cfa-top nesting)
+	    (logfl side-effect-free no-cfa-top nesting)
+	    (sinfl side-effect-free no-cfa-top nesting)
+	    (cosfl side-effect-free no-cfa-top nesting)
+	    (tanfl side-effect-free no-cfa-top nesting)
+	    (asinfl side-effect-free no-cfa-top nesting)
+	    (acosfl side-effect-free no-cfa-top nesting)
+	    (atanfl side-effect-free no-cfa-top nesting)
+	    (sqrtfl side-effect-free no-cfa-top nesting)))
 
 ;*---------------------------------------------------------------------*/
 ;*    real? ...                                                        */
@@ -220,7 +265,7 @@
 ;*    positivefl? ...                                                  */
 ;*---------------------------------------------------------------------*/
 (define-inline (positivefl? r)
-   (>=fl r 0.0))
+   (>fl r 0.0))
 
 ;*---------------------------------------------------------------------*/
 ;*    negativefl? ...                                                  */
@@ -329,8 +374,18 @@
 ;*---------------------------------------------------------------------*/
 ;*    roundfl ...                                                      */
 ;*---------------------------------------------------------------------*/
-(define-inline (roundfl r)
-   (floorfl (+fl r 0.5)))
+(define (roundfl r)
+   (let* ((int (floorfl r))
+          (frac (-fl r int)))
+      (cond ((<fl frac 0.5) int)
+            ((>fl frac 0.5) (+fl int 1.0))
+            (else (*fl 2.0 (ceilingfl (/fl int 2.0)))))))
+
+;*---------------------------------------------------------------------*/
+;*    remainderfl ...                                                  */
+;*---------------------------------------------------------------------*/
+(define-inline (remainderfl n1 n2)
+   (c-fmod n1 n2))
 
 ;*---------------------------------------------------------------------*/
 ;*    expfl ...                                                        */
@@ -400,7 +455,11 @@
 	  (let ((proc::obj (string->bstring "atanfl"))
 		(msg::obj (string->bstring "Domain error"))
 		(obj::obj (double->real 0.0)))
-	     (error proc msg obj))
+	     ;; !!! Warning
+	     ;; the_failure is prefered to Error in order to fix a
+	     ;; registers allocation bug in gcc 2.96
+	     (the_failure proc msg obj)
+	     0.0)
 	  (c-atan2 x y))))
 
 ;*---------------------------------------------------------------------*/
@@ -417,7 +476,9 @@
        (let ((proc::obj (string->bstring "sqrtfl"))
 	     (msg::obj (string->bstring "Domain error"))
 	     (obj::obj (double->real r)))
-	  (error proc msg obj))
+	  (begin
+	     (error proc msg obj)
+	     0.0))
        (c-sqrt r)))
 
 ;*---------------------------------------------------------------------*/
@@ -443,3 +504,39 @@
 ;*---------------------------------------------------------------------*/
 (define-inline (real->string real)
    (c-real->string real))
+
+;*---------------------------------------------------------------------*/
+;*    ieee-string->real ...                                            */
+;*---------------------------------------------------------------------*/
+(define-inline (ieee-string->real string)
+   (%ieee-string->double string))
+
+;*---------------------------------------------------------------------*/
+;*    real->ieee-string ...                                            */
+;*---------------------------------------------------------------------*/
+(define-inline (real->ieee-string real)
+   (%double->ieee-string real))
+
+;*---------------------------------------------------------------------*/
+;*    ieee-string->double ...                                          */
+;*---------------------------------------------------------------------*/
+(define-inline (ieee-string->double string)
+   (%ieee-string->double string))
+
+;*---------------------------------------------------------------------*/
+;*    double->ieee-string ...                                          */
+;*---------------------------------------------------------------------*/
+(define-inline (double->ieee-string double)
+   (%double->ieee-string double))
+
+;*---------------------------------------------------------------------*/
+;*    ieee-string->float ...                                           */
+;*---------------------------------------------------------------------*/
+(define-inline (ieee-string->float string)
+   (%ieee-string->double string))
+
+;*---------------------------------------------------------------------*/
+;*    float->ieee-string ...                                           */
+;*---------------------------------------------------------------------*/
+(define-inline (float->ieee-string float)
+   (%double->ieee-string float))
