@@ -1,38 +1,15 @@
-;* --------------------------------------------------------------------*/
-;*    Copyright (c) 1992-1998 by Manuel Serrano. All rights reserved.  */
-;*                                                                     */
-;*                                     ,--^,                           */
-;*                               _ ___/ /|/                            */
-;*                           ,;'( )__, ) '                             */
-;*                          ;;  //   L__.                              */
-;*                          '   \   /  '                               */
-;*                               ^   ^                                 */
-;*                                                                     */
-;*                                                                     */
-;*    This program is distributed in the hope that it will be useful.  */
-;*    Use and copying of this software and preparation of derivative   */
-;*    works based upon this software are permitted, so long as the     */
-;*    following conditions are met:                                    */
-;*           o credit to the authors is acknowledged following         */
-;*             current academic behaviour                              */
-;*           o no fees or compensation are charged for use, copies,    */
-;*             or access to this software                              */
-;*           o this copyright notice is included intact.               */
-;*      This software is made available AS IS, and no warranty is made */
-;*      about the software or its performance.                         */
-;*                                                                     */
-;*      Bug descriptions, use reports, comments or suggestions are     */
-;*      welcome. Send them to                                          */
-;*        Manuel Serrano -- Manuel.Serrano@unice.fr                    */
-;*-------------------------------------------------------------------- */
 ;*=====================================================================*/
-;*    serrano/prgm/project/bigloo/runtime1.9b/Ieee/number.scm          */
+;*    serrano/prgm/project/bigloo/runtime/Ieee/number.scm              */
 ;*    -------------------------------------------------------------    */
 ;*    Author      :  Manuel Serrano                                    */
 ;*    Creation    :  Fri Mar 24 09:59:43 1995                          */
-;*    Last change :  Mon Apr 14 11:33:44 1997 (serrano)                */
+;*    Last change :  Sun Aug  6 21:31:18 2006 (serrano)                */
 ;*    -------------------------------------------------------------    */
 ;*    6.5. Numbers (page 18, r4)                                       */
+;*    -------------------------------------------------------------    */
+;*    Source documentation:                                            */
+;*       @path ../../manuals/body.texi@                                */
+;*       @node Numbers@                                                */
 ;*=====================================================================*/
  
 ;*---------------------------------------------------------------------*/
@@ -40,33 +17,78 @@
 ;*---------------------------------------------------------------------*/
 (module __r4_numbers_6_5
    
-   (import  (__error                   "Llib/error.scm"))
+   (import  __error)
 
-   (use     (__type                    "Llib/type.scm")
-	    (__bigloo                  "Llib/bigloo.scm")
-	    (__tvector                 "Llib/tvector.scm")
-	    (__r4_equivalence_6_2      "Ieee/equiv.scm")
-	    (__r4_numbers_6_5_fixnum   "Ieee/fixnum.scm")
-	    (__r4_booleans_6_1         "Ieee/boolean.scm")
-	    (__r4_characters_6_6       "Ieee/char.scm")
-	    (__r4_pairs_and_lists_6_3  "Ieee/pair-list.scm")
-	    (__r4_vectors_6_8          "Ieee/vector.scm")
-	    (__r4_numbers_6_5_flonum   "Ieee/flonum.scm")
-	    (__r4_symbols_6_4          "Ieee/symbol.scm")
-	    (__r4_strings_6_7          "Ieee/string.scm")
+   (use     __type
+	    __bigloo
+	    __tvector
+	    __r4_equivalence_6_2
+	    __r4_numbers_6_5_fixnum
+	    __r4_booleans_6_1
+	    __r4_characters_6_6
+	    __r4_pairs_and_lists_6_3
+	    __r4_vectors_6_8
+	    __r4_numbers_6_5_flonum
+	    __r4_symbols_6_4
+	    __r4_strings_6_7
 
-	    (__evenv                   "Eval/evenv.scm"))
+	    __evenv)
    
-   (foreign (macro double c-fixnum->flonum (long)   "(double)")
-	    (macro long   c-flonum->fixnum (double) "(long)"))
+   (extern  (macro c-fixnum->flonum::double (::long)   "(double)")
+	    (macro c-flonum->fixnum::long   (::double) "(long)")
+	    
+	    (macro c-elong->fixnum::long (::elong)     "(long)")
+	    (macro c-fixnum->elong::elong (::long)     "(long)")
+	    (macro c-llong->fixnum::long (::llong)     "(long)")
+	    (macro c-fixnum->llong::llong (::long)     "(BGL_LONGLONG_T)")
+	    
+	    (macro c-elong->flonum::double  (::elong)  "(double)")
+	    (macro c-flonum->elong::long    (::double) "(long)")
+	    (macro c-llong->flonum::double  (::llong)  "(double)")
+	    (macro c-flonum->llong::llong   (::double) "DOUBLE_TO_LLONG")
+	    
+	    (export exact->inexact "bgl_exact_to_inexact")
+	    (export inexact->exact "bgl_inexact_to_exact"))
+   
+   (java    (class foreign
+	       (method static c-fixnum->flonum::double (::long)
+		       "FIXNUM_TO_FLONUM")
+	       (method static c-flonum->fixnum::long   (::double)
+		       "FLONUM_TO_FIXNUM")
+	       
+	       (method static c-elong->fixnum::long (::elong)
+		       "ELONG_TO_LONG")
+	       (method static c-fixnum->elong::elong (::long)
+		       "LONG_TO_ELONG")
+	       (method static c-llong->fixnum::long (::llong)
+		       "LLONG_TO_LONG")
+	       (method static c-fixnum->llong::llong (::long)
+		       "LONG_TO_LLONG")
+	       
+	       (method static c-elong->flonum::double (::elong)
+		       "ELONG_TO_FLONUM")
+	       (method static c-flonum->elong::elong   (::double)
+		       "FLONUM_TO_ELONG")
+	       (method static c-llong->flonum::double (::llong)
+		       "LLONG_TO_FLONUM")
+	       (method static c-flonum->llong::llong   (::double)
+		       "FLONUM_TO_LLONG")))
    
    (export  (inline number?::bool           obj)
 	    (inline exact?::bool            z)
 	    (inline inexact?::bool          z)
 	    (complex?::bool                 x)
 	    (rational?::bool                x)
-	    (inline flonum->fixnum::long    x)
-	    (inline fixnum->flonum::double  x)
+	    (inline flonum->fixnum::long    ::double)
+	    (inline fixnum->flonum::double  ::long)
+	    (inline fixnum->elong::elong    ::long)
+	    (inline elong->fixnum::long     ::elong)
+	    (inline fixnum->llong::llong    ::long)
+	    (inline llong->fixnum::long     ::llong)
+	    (inline flonum->elong::elong    ::double)
+	    (inline elong->flonum::double   ::elong)
+	    (inline flonum->llong::llong    ::double)
+	    (inline llong->flonum::double   ::llong)
 	    (2=::bool                       x y)
 	    (=::bool                        x y . z)
 	    (2<::bool                       x y) 
@@ -81,7 +103,9 @@
 	    (positive?::bool                x)
 	    (negative?::bool                x)
 	    (max                            x . y)
+	    (2max                           x y)
 	    (min                            x . y)
+	    (2min                           x y)
 	    (2+                             x y)
 	    (+                              . x)
 	    (2*                             x y)
@@ -110,49 +134,60 @@
 	    (number->string::string         x . radix)
 	    (string->number                 x . radix))
 
-   (pragma  (2= side-effect-free)
-	    (= side-effect-free)
-	    (2< side-effect-free)
-	    (< side-effect-free)
-	    (2> side-effect-free)
-	    (> side-effect-free)
-	    (2<= side-effect-free)
-	    (<= side-effect-free)
-	    (2>= side-effect-free)
-	    (>= side-effect-free)
-	    (zero? side-effect-free)
-	    (positive? side-effect-free)
-	    (negative? side-effect-free)
-	    (max side-effect-free)
-	    (min side-effect-free)
-	    (2+ side-effect-free)
-	    (+ side-effect-free)
-	    (2* side-effect-free)
-	    (* side-effect-free)
-	    (2/ side-effect-free)
-	    (/ side-effect-free)
-	    (2- side-effect-free)
-	    (- side-effect-free)
-	    (abs side-effect-free)
-	    (floor side-effect-free)
-	    (ceiling side-effect-free)
-	    (truncate side-effect-free)
-	    (round side-effect-free)
-	    (exp side-effect-free)
-	    (log side-effect-free)
-	    (sin side-effect-free)
-	    (cos side-effect-free)
-	    (tan side-effect-free)
-	    (asin side-effect-free)
-	    (acos side-effect-free)
-	    (atan side-effect-free)
-	    (sqrt side-effect-free)
-	    (expt side-effect-free)
-	    (exact->inexact side-effect-free)
-	    (inexact->exact side-effect-free)
-	    (number->string side-effect-free)
-	    (string->number side-effect-free)))
-		  
+   (pragma  (c-fixnum->flonum side-effect-free args-safe (effect))
+	    (c-flonum->fixnum side-effect-free args-safe (effect))
+	    (c-elong->fixnum side-effect-free args-safe (effect))
+	    (c-fixnum->elong side-effect-free args-safe (effect))
+	    (c-llong->fixnum side-effect-free args-safe (effect))
+	    (c-fixnum->llong side-effect-free args-safe (effect))
+	    (c-elong->flonum side-effect-free args-safe (effect))
+	    (c-flonum->elong side-effect-free args-safe (effect))
+	    (c-llong->flonum side-effect-free args-safe (effect))
+	    (c-flonum->llong side-effect-free args-safe (effect))
+	    (2= side-effect-free (effect))
+	    (= side-effect-free (effect))
+	    (2< side-effect-free (effect))
+	    (< side-effect-free (effect))
+	    (2> side-effect-free (effect))
+	    (> side-effect-free (effect))
+	    (2<= side-effect-free (effect))
+	    (<= side-effect-free (effect))
+	    (2>= side-effect-free (effect))
+	    (>= side-effect-free (effect))
+	    (zero? side-effect-free (effect))
+	    (positive? side-effect-free (effect))
+	    (negative? side-effect-free (effect))
+	    (max side-effect-free (effect))
+	    (2max side-effect-free (effect))
+	    (min side-effect-free (effect))
+	    (2min side-effect-free (effect))
+	    (2+ side-effect-free (effect))
+	    (+ side-effect-free (effect))
+	    (2* side-effect-free (effect))
+	    (* side-effect-free (effect))
+	    (2/ side-effect-free (effect))
+	    (/ side-effect-free (effect))
+	    (2- side-effect-free (effect))
+	    (- side-effect-free (effect))
+	    (abs side-effect-free (effect))
+	    (floor side-effect-free (effect))
+	    (ceiling side-effect-free (effect))
+	    (truncate side-effect-free (effect))
+	    (round side-effect-free (effect))
+	    (exp side-effect-free (effect))
+	    (log side-effect-free (effect))
+	    (sin side-effect-free (effect))
+	    (cos side-effect-free (effect))
+	    (tan side-effect-free (effect))
+	    (asin side-effect-free (effect))
+	    (acos side-effect-free (effect))
+	    (atan side-effect-free (effect))
+	    (sqrt side-effect-free (effect))
+	    (expt side-effect-free (effect))
+	    (exact->inexact side-effect-free args-safe (effect))
+	    (inexact->exact side-effect-free args-safe (effect))
+	    (number->string side-effect-free (effect))
+	    (string->number side-effect-free (effect))))
 
 ;*---------------------------------------------------------------------*/
 ;*    number? ...                                                      */
@@ -160,13 +195,21 @@
 (define-inline (number? obj)
    (if (fixnum? obj)
        #t
-       (flonum? obj)))
+       (if (flonum? obj)
+	   #t
+	   (if (elong? obj)
+	       #t
+	       (llong? obj)))))
 
 ;*---------------------------------------------------------------------*/
 ;*    exact? ...                                                       */
 ;*---------------------------------------------------------------------*/
 (define-inline (exact? z)
-   (integer? z))
+   (if (fixnum? z)
+       #t
+       (if (elong? z)
+	   #t
+	   (llong? z))))
 
 ;*---------------------------------------------------------------------*/
 ;*    inexact? ...                                                     */
@@ -198,29 +241,119 @@
 (define-inline (fixnum->flonum x)
    (c-fixnum->flonum x))
 		       
+;*---------------------------------------------------------------------*/
+;*    fixnum->elong ...                                                */
+;*---------------------------------------------------------------------*/
+(define-inline (fixnum->elong x)
+   (c-fixnum->elong x))
+
+;*---------------------------------------------------------------------*/
+;*    elong->fixnum ...                                                */
+;*---------------------------------------------------------------------*/
+(define-inline (elong->fixnum x)
+   (c-elong->fixnum x))
+		       
+;*---------------------------------------------------------------------*/
+;*    fixnum->llong ...                                                */
+;*---------------------------------------------------------------------*/
+(define-inline (fixnum->llong x)
+   (c-fixnum->llong x))
+
+;*---------------------------------------------------------------------*/
+;*    llong->fixnum ...                                                */
+;*---------------------------------------------------------------------*/
+(define-inline (llong->fixnum x)
+   (c-llong->fixnum x))
+		       
+;*---------------------------------------------------------------------*/
+;*    flonum->elong ...                                                */
+;*---------------------------------------------------------------------*/
+(define-inline (flonum->elong x)
+   (c-flonum->elong x))
+
+;*---------------------------------------------------------------------*/
+;*    elong->flonum ...                                                */
+;*---------------------------------------------------------------------*/
+(define-inline (elong->flonum x)
+   (c-elong->flonum x))
+		       
+;*---------------------------------------------------------------------*/
+;*    flonum->llong ...                                                */
+;*---------------------------------------------------------------------*/
+(define-inline (flonum->llong x)
+   (c-flonum->llong x))
+
+;*---------------------------------------------------------------------*/
+;*    llong->flonum ...                                                */
+;*---------------------------------------------------------------------*/
+(define-inline (llong->flonum x)
+   (c-llong->flonum x))
+
+;*---------------------------------------------------------------------*/
+;*    2op :: ...                                                       */
+;*---------------------------------------------------------------------*/
+(define-macro (2op op x y)
+   (let ((opfx (symbol-append op 'fx))
+	 (opfl (symbol-append op 'fl))
+	 (opelong (symbol-append op 'elong))
+	 (opllong (symbol-append op 'llong)))
+      `(cond
+	  ((fixnum? ,x)
+	   (cond
+	      ((fixnum? ,y)
+	       (,opfx ,x ,y))
+	      ((flonum? ,y)
+	       (,opfl (fixnum->flonum ,x) ,y))
+	      ((elong? ,y)
+	       (,opelong (fixnum->elong ,x) ,y))
+	      ((llong? y)
+	       (,opllong (fixnum->llong ,x) ,y))
+	      (else
+	       (error ',op "not a number" ,y))))
+	  ((flonum? ,x)
+	   (cond
+	      ((flonum? ,y)
+	       (,opfl ,x ,y))
+	      ((fixnum? ,y)
+	       (,opfl ,x (fixnum->flonum ,y)))
+	      ((elong? y)
+	       (,opfl ,x (elong->flonum ,y)))
+	      ((llong? y)
+	       (,opfl ,x (llong->flonum ,y)))
+	      (else
+	       (error ',op "not a number" ,y))))
+	  ((elong? ,x)
+	   (cond
+	      ((fixnum? ,y)
+	       (,opelong ,x (fixnum->elong ,y)))
+	      ((flonum? ,y)
+	       (,opfl (elong->flonum ,x) ,y))
+	      ((elong? ,y)
+	       (,opelong ,x ,y))
+	      ((llong? ,y)
+	       (,opllong (flonum->llong (elong->flonum ,x)) ,y))
+	      (else
+	       (error ',op "not a number" ,y))))
+	  ((llong? ,x)
+	   (cond
+	      ((fixnum? ,y)
+	       (,opllong ,x (fixnum->llong ,y)))
+	      ((flonum? y)
+	       (,opfl (llong->flonum ,x) ,y))
+	      ((elong? ,y)
+	       (,opllong ,x (flonum->llong (elong->flonum ,y))))
+	      ((llong? ,y)
+	       (,opllong ,x ,y))
+	      (else
+	       (error ',op "not a number" ,y))))
+	  (else
+	   (error ',op "not a number" ,x)))))
+
 ;*---------------------------------------------------------------------*/
 ;*    2= ...                                                           */
 ;*---------------------------------------------------------------------*/
 (define (2= x y)
-   (cond
-      ((fixnum? x)
-       (cond
-	  ((fixnum? y)
-	   (=fx x y))
-	  ((flonum? y)
-	   (=fl (fixnum->flonum x) y))
-	  (else
-	   (error "=" "not a number" y))))
-      ((flonum? x)
-       (cond
-	  ((flonum? y)
-	   (=fl x y))
-	  ((fixnum? y)
-	   (=fl x (fixnum->flonum y)))
-	  (else
-	   (error "=" "not a number" y))))
-      (else
-       (error "=" "not a number" x))))
+   (2op = x y))
 
 ;*---------------------------------------------------------------------*/
 ;*    = ...                                                            */
@@ -239,25 +372,7 @@
 ;*    2< ...                                                           */
 ;*---------------------------------------------------------------------*/
 (define (2< x y)
-   (cond
-      ((fixnum? x)
-       (cond
-	  ((fixnum? y)
-	   (<fx x y))
-	  ((flonum? y)
-	   (<fl (fixnum->flonum x) y))
-	  (else
-	   (error "<" "not a number" y))))
-      ((flonum? x)
-       (cond
-	  ((flonum? y)
-	   (<fl x y))
-	  ((fixnum? y)
-	   (<fl x (fixnum->flonum y)))
-	  (else
-	   (error "<" "not a number" y))))
-      (else
-       (error "<" "not a number" x))))
+   (2op < x y))
 
 ;*---------------------------------------------------------------------*/
 ;*    < ...                                                            */
@@ -277,25 +392,7 @@
 ;*    2> ...                                                           */
 ;*---------------------------------------------------------------------*/
 (define (2> x y)
-   (cond
-      ((fixnum? x)
-       (cond
-	  ((fixnum? y)
-	   (>fx x y))
-	  ((flonum? y)
-	   (>fl (fixnum->flonum x) y))
-	  (else
-	   (error ">" "not a number" y))))
-      ((flonum? x)
-       (cond
-	  ((flonum? y)
-	   (>fl x y))
-	  ((fixnum? y)
-	   (>fl x (fixnum->flonum y)))
-	  (else
-	   (error ">" "not a number" y))))
-      (else
-       (error ">" "not a number" x))))
+   (2op > x y))
 
 ;*---------------------------------------------------------------------*/
 ;*    > ...                                                            */
@@ -314,25 +411,7 @@
 ;*    2<= ...                                                          */
 ;*---------------------------------------------------------------------*/
 (define (2<= x y)
-   (cond
-      ((fixnum? x)
-       (cond
-	  ((fixnum? y)
-	   (<=fx x y))
-	  ((flonum? y)
-	   (<=fl (fixnum->flonum x) y))
-	  (else
-	   (error "<=" "not a number" y))))
-      ((flonum? x)
-       (cond
-	  ((flonum? y)
-	   (<=fl x y))
-	  ((fixnum? y)
-	   (<=fl x (fixnum->flonum y)))
-	  (else
-	   (error "<=" "not a number" y))))
-      (else
-       (error "<=" "not a number" x))))
+   (2op <= x y))
 
 ;*---------------------------------------------------------------------*/
 ;*    <= ...                                                           */
@@ -351,25 +430,7 @@
 ;*    2>= ...                                                          */
 ;*---------------------------------------------------------------------*/
 (define (2>= x y)
-   (cond
-      ((fixnum? x)
-       (cond
-	  ((fixnum? y)
-	   (>=fx x y))
-	  ((flonum? y)
-	   (>=fl (fixnum->flonum x) y))
-	  (else
-	   (error ">=" "not a number" y))))
-      ((flonum? x)
-       (cond
-	  ((flonum? y)
-	   (>=fl x y))
-	  ((fixnum? y)
-	   (>=fl x (fixnum->flonum y)))
-	  (else
-	   (error ">=" "not a number" y))))
-      (else
-       (error ">=" "not a number" x))))
+   (2op >= x y))
 
 ;*---------------------------------------------------------------------*/
 ;*    >= ...                                                           */
@@ -393,6 +454,10 @@
        (zerofx? x))
       ((flonum? x)
        (zerofl? x))
+      ((elong? x)
+       (=elong x #e0))
+      ((llong? x)
+       (=llong x #l0))
       (else
        (error "zero" "not a number" x))))
 
@@ -405,6 +470,10 @@
        (positivefx? x))
       ((flonum? x)
        (positivefl? x))
+      ((elong? x)
+       (>elong x #e0))
+      ((llong? x)
+       (>llong x #l0))
       (else
        (error "positive" "not a number" x))))
 
@@ -417,9 +486,31 @@
        (negativefx? x))
       ((flonum? x)
        (negativefl? x))
+      ((elong? x)
+       (<elong x #e0))
+      ((llong? x)
+       (<llong x #l0))
       (else
        (error "negative" "not a number" x))))
 
+;*---------------------------------------------------------------------*/
+;*    2max ...                                                         */
+;*---------------------------------------------------------------------*/
+(define-inline (2maxfx x y)
+   (if (>fx x y) x y))
+(define-inline (2maxfl x y)
+   (if (>fl x y) x y))
+(define-inline (2maxelong x y)
+   (if (>elong x y) x y))
+(define-inline (2maxllong x y)
+   (if (>llong x y) x y))
+
+;*---------------------------------------------------------------------*/
+;*    2max ...                                                         */
+;*---------------------------------------------------------------------*/
+(define (2max x y)
+   (2op 2max x y))
+
 ;*---------------------------------------------------------------------*/
 ;*    max ...                                                          */
 ;*---------------------------------------------------------------------*/
@@ -427,12 +518,27 @@
    (let loop ((x x)
 	      (y y))
       (if (pair? y)
-	  (loop (if (> x (car y))
-		    x
-		    (car y))
-		(cdr y))
+	  (loop (2max x (car y)) (cdr y))
 	  x)))
 
+;*---------------------------------------------------------------------*/
+;*    2min ...                                                         */
+;*---------------------------------------------------------------------*/
+(define-inline (2minfx x y)
+   (if (>fx x y) y x))
+(define-inline (2minfl x y)
+   (if (>fl x y) y x))
+(define-inline (2minelong x y)
+   (if (>elong x y) y x))
+(define-inline (2minllong x y)
+   (if (>llong x y) y x))
+
+;*---------------------------------------------------------------------*/
+;*    2min ...                                                         */
+;*---------------------------------------------------------------------*/
+(define (2min x y)
+   (2op 2min x y))
+
 ;*---------------------------------------------------------------------*/
 ;*    min ...                                                          */
 ;*---------------------------------------------------------------------*/
@@ -440,35 +546,14 @@
    (let loop ((x x)
 	      (y y))
       (if (pair? y)
-	  (loop (if (< x (car y))
-		    x
-		    (car y))
-		(cdr y))
+	  (loop (2min x (car y)) (cdr y))
 	  x)))
 
 ;*---------------------------------------------------------------------*/
 ;*    2+ ...                                                           */
 ;*---------------------------------------------------------------------*/
 (define (2+ x y)
-   (cond
-      ((fixnum? x)
-       (cond
-	  ((fixnum? y)
-	   (+fx x y))
-	  ((flonum? y)
-	   (+fl (fixnum->flonum x) y))
-	  (else
-	   (error "+" "not a number" y))))
-      ((flonum? x)
-       (cond
-	  ((flonum? y)
-	   (+fl x y))
-	  ((fixnum? y)
-	   (+fl x (fixnum->flonum y)))
-	  (else
-	   (error "+" "not a number" y))))
-      (else
-       (error "+" "not a number" x))))
+   (2op + x y))
 
 ;*---------------------------------------------------------------------*/
 ;*    + ...                                                            */
@@ -485,25 +570,7 @@
 ;*    2* ...                                                           */
 ;*---------------------------------------------------------------------*/
 (define (2* x y)
-   (cond
-      ((fixnum? x)
-       (cond
-	  ((fixnum? y)
-	   (*fx x y))
-	  ((flonum? y)
-	   (*fl (fixnum->flonum x) y))
-	  (else
-	   (error "*" "not a number" y))))
-      ((flonum? x)
-       (cond
-	  ((flonum? y)
-	   (*fl x y))
-	  ((fixnum? y)
-	   (*fl x (fixnum->flonum y)))
-	  (else
-	   (error "*" "not a number" y))))
-      (else
-       (error "*" "not a number" x))))
+   (2op * x y))
 
 ;*---------------------------------------------------------------------*/
 ;*    * ...                                                            */
@@ -519,25 +586,7 @@
 ;*    2- ...                                                           */
 ;*---------------------------------------------------------------------*/
 (define (2- x y)
-   (cond
-      ((fixnum? x)
-       (cond
-	  ((fixnum? y)
-	   (-fx x y))
-	  ((flonum? y)
-	   (-fl (fixnum->flonum x) y))
-	  (else
-	   (error "-" "not a number" y))))
-      ((flonum? x)
-       (cond
-	  ((flonum? y)
-	   (-fl x y))
-	  ((fixnum? y)
-	   (-fl x (fixnum->flonum y)))
-	  (else
-	   (error "-" "not a number" y))))
-      (else
-       (error "-" "not a number" x))))
+   (2op - x y))
 
 ;*---------------------------------------------------------------------*/
 ;*    - ...                                                            */
@@ -564,18 +613,74 @@
 	       (/fl (fixnum->flonum x) (fixnum->flonum y))))
 	  ((flonum? y)
 	   (/fl (fixnum->flonum x) y))
+	  ((elong? y)
+	   (let ((ex (fixnum->elong x)))
+	      (if (=elong (remainderelong ex y) #e0)
+		  (/elong ex y)
+		  (/fl (fixnum->flonum x) (elong->flonum y)))))
+	  ((llong? y)
+	   (let ((lx (fixnum->llong x)))
+	      (if (=llong (remainderllong lx y) #l0)
+		  (/llong lx y)
+		  (/fl (fixnum->flonum x) (llong->flonum y)))))
 	  (else
-	   (error "/" "not a number" y))))
+	   (error '/ "not a number" y))))
       ((flonum? x)
        (cond
 	  ((flonum? y)
 	   (/fl x y))
 	  ((fixnum? y)
 	   (/fl x (fixnum->flonum y)))
+	  ((elong? y)
+	   (/fl x (elong->flonum y)))
+	  ((llong? y)
+	   (/fl x (llong->flonum y)))
 	  (else
-	   (error "/" "not a number" y))))
+	   (error '/ "not a number" y))))
+      ((elong? x)
+       (cond
+	  ((fixnum? y)
+	   (let ((ey (fixnum->elong y)))
+	      (if (=elong (remainderelong x ey) #e0)
+		  (/elong x ey)
+		  (/fl (elong->flonum x) (fixnum->flonum y)))))
+	  ((flonum? y)
+	   (/fl (elong->flonum x) y))
+	  ((elong? y)
+	   (if (=elong (remainderelong x y) #e0)
+	       (/elong x y)
+	       (/fl (elong->flonum x) (elong->flonum y))))
+	  ((llong? y)
+	   (let* ((fx (elong->flonum x))
+		  (lx (flonum->llong fx)))
+	      (if (=llong (remainderllong lx y) #l0)
+		  (/llong lx y)
+		  (/fl fx (llong->flonum y)))))
+	  (else
+	   (error '/ "not a number" y))))
+      ((llong? x)
+       (cond
+	  ((fixnum? y)
+	   (let ((ly (fixnum->llong y)))
+	      (if (=llong (remainderllong x ly) #l0)
+		  (/llong x ly)
+		  (/fl (llong->flonum x) (fixnum->flonum y)))))
+	  ((flonum? y)
+	   (/fl (llong->flonum x) y))
+	  ((elong? y)
+	   (let* ((fy (elong->flonum y))
+		  (ly (flonum->llong fy)))
+	      (if (=llong (remainderllong x ly) #l0)
+		  (/llong x y)
+		  (/fl (llong->flonum x) fy))))
+	  ((llong? y)
+	   (if (=llong (remainderllong x y) #l0)
+	       (/llong x y)
+	       (/fl (llong->flonum x) (llong->flonum y))))
+	  (else
+	   (error '/ "not a number" y))))
       (else
-       (error "/" "not a number" x))))
+       (error / "not a number" x))))
 
 ;*---------------------------------------------------------------------*/
 ;*    / ...                                                            */
@@ -599,6 +704,14 @@
        (absfx x))
       ((flonum? x)
        (absfl x))
+      ((elong? x)
+       (if (<elong x #e0)
+	   (negelong x)
+	   x))
+      ((llong? x)
+       (if (<llong x #l0)
+	   (negllong x)
+	   x))
       (else
        (error "abs" "not a number" x))))
 
@@ -611,6 +724,10 @@
        x)
       ((flonum? x)
        (floorfl x))
+      ((elong? x)
+       x)
+      ((llong? x)
+       x)
       (else
        (error "floor" "not a number" x))))
 
@@ -623,6 +740,10 @@
        x)
       ((flonum? x)
        (ceilingfl x))
+      ((elong? x)
+       x)
+      ((llong? x)
+       x)
       (else
        (error "ceiling" "not a number" x))))
 
@@ -635,6 +756,10 @@
        x)
       ((flonum? x)
        (truncatefl x))
+      ((elong? x)
+       x)
+      ((llong? x)
+       x)
       (else
        (error "truncate" "not a number" x))))
 
@@ -647,6 +772,10 @@
        x)
       ((flonum? x)
        (roundfl x))
+      ((elong? x)
+       x)
+      ((llong? x)
+       x)
       (else
        (error "round" "not a number" x))))
 
@@ -655,10 +784,14 @@
 ;*---------------------------------------------------------------------*/
 (define (exp x)
    (cond
-      ((fixnum? x)
-       (expfl (fixnum->flonum x)))
       ((flonum? x)
        (expfl x))
+      ((fixnum? x)
+       (expfl (fixnum->flonum x)))
+      ((elong? x)
+       (expfl (elong->flonum x)))
+      ((llong? x)
+       (expfl (llong->flonum x)))
       (else
        (error "exp" "not a number" x))))
 
@@ -667,10 +800,14 @@
 ;*---------------------------------------------------------------------*/
 (define (log x)
    (cond
-      ((fixnum? x)
-       (logfl (fixnum->flonum x)))
       ((flonum? x)
        (logfl x))
+      ((fixnum? x)
+       (logfl (fixnum->flonum x)))
+      ((elong? x)
+       (logfl (elong->flonum x)))
+      ((llong? x)
+       (logfl (llong->flonum x)))
       (else
        (error "log" "not a number" x))))
 
@@ -679,10 +816,14 @@
 ;*---------------------------------------------------------------------*/
 (define (sin x)
    (cond
-      ((fixnum? x)
-       (sinfl (fixnum->flonum x)))
       ((flonum? x)
        (sinfl x))
+      ((fixnum? x)
+       (sinfl (fixnum->flonum x)))
+      ((elong? x)
+       (sin (elong->flonum x)))
+      ((llong? x)
+       (sin (llong->flonum x)))
       (else
        (error "sin" "not a number" x))))
 
@@ -691,10 +832,14 @@
 ;*---------------------------------------------------------------------*/
 (define (cos x)
    (cond
-      ((fixnum? x)
-       (cosfl (fixnum->flonum x)))
       ((flonum? x)
        (cosfl x))
+      ((fixnum? x)
+       (cosfl (fixnum->flonum x)))
+      ((elong? x)
+       (cos (elong->flonum x)))
+      ((llong? x)
+       (cos (llong->flonum x)))
       (else
        (error "cos" "not a number" x))))
 
@@ -703,10 +848,14 @@
 ;*---------------------------------------------------------------------*/
 (define (tan x)
    (cond
-      ((fixnum? x)
-       (tanfl (fixnum->flonum x)))
       ((flonum? x)
        (tanfl x))
+      ((fixnum? x)
+       (tanfl (fixnum->flonum x)))
+      ((elong? x)
+       (tan (elong->flonum x)))
+      ((llong? x)
+       (tan (llong->flonum x)))
       (else
        (error "tan" "not a number" x))))
 
@@ -715,10 +864,14 @@
 ;*---------------------------------------------------------------------*/
 (define (asin x)
    (cond
-      ((fixnum? x)
-       (asinfl (fixnum->flonum x)))
       ((flonum? x)
        (asinfl x))
+      ((fixnum? x)
+       (asinfl (fixnum->flonum x)))
+      ((elong? x)
+       (asin (elong->flonum x)))
+      ((llong? x)
+       (asin (llong->flonum x)))
       (else
        (error "asin" "not a number" x))))
 
@@ -727,10 +880,14 @@
 ;*---------------------------------------------------------------------*/
 (define (acos x)
    (cond
-      ((fixnum? x)
-       (acosfl (fixnum->flonum x)))
       ((flonum? x)
        (acosfl x))
+      ((fixnum? x)
+       (acosfl (fixnum->flonum x)))
+      ((elong? x)
+       (acos (elong->flonum x)))
+      ((llong? x)
+       (acos (llong->flonum x)))
       (else
        (error "acos" "not a number" x))))
 
@@ -753,10 +910,14 @@
 	     (atanfl x y)
 	     (atanfl x)))
       (cond
-	 ((fixnum? x)
-	  (do-atanfl (fixnum->flonum x)))
 	 ((flonum? x)
 	  (do-atanfl x))
+	 ((fixnum? x)
+	  (do-atanfl (fixnum->flonum x)))
+	 ((elong? x)
+	  (do-atanfl (elong->flonum x)))
+	 ((llong? x)
+	  (do-atanfl (llong->flonum x)))
 	 (else
 	  (error "atan" "not a number" x)))))
 
@@ -769,6 +930,10 @@
        (sqrtfl (fixnum->flonum x)))
       ((flonum? x)
        (sqrtfl x))
+      ((elong? x)
+       (sqrtfl (elong->flonum x)))
+      ((llong? x)
+       (sqrtfl (llong->flonum x)))
       (else
        (error "sqrt" "not a number" x))))
 
@@ -776,36 +941,51 @@
 ;*    expt ...                                                         */
 ;*---------------------------------------------------------------------*/
 (define (expt x y)
-   (if (and (flonum? x) (flonum? y) (=fl x 0.0) (=fl y 0.0))
-       1.0
-       (cond
-	  ((and (fixnum? x)
-		(fixnum? y))
-	   (flonum->fixnum (exptfl (fixnum->flonum x)  (fixnum->flonum y))))
-	  ((fixnum? x)
-	   (cond
-	      ((flonum? y)
-	       (exptfl (fixnum->flonum x) y))
-	      (else
-	       (error "expt" "not a number" y))))
-	  ((flonum? x)
-	   (cond
-	      ((flonum? y)
-	       (exptfl x y))
-	      ((fixnum? y)
-	       (exptfl x (fixnum->flonum y)))
-	      (else
-	       (error "expt" "not a number" y))))
-	  (else
-	   (error "expt" "not a number" x)))))
+   (cond
+      ((and (flonum? x) (flonum? y) (=fl x 0.0) (=fl y 0.0))
+       1.0)
+      ((and (fixnum? x) (fixnum? y) (>=fx y 0))
+       (flonum->fixnum (exptfl (fixnum->flonum x) (fixnum->flonum y))))
+      (else
+       (let ((x1 (cond
+		    ((flonum? x)
+		     x)
+		    ((fixnum? x)
+		     (fixnum->flonum x))
+		    ((elong? x)
+		     (elong->flonum x))
+		    ((llong? x)
+		     (llong->flonum x))
+		    (else
+		     (error 'expr "not a number" x))))
+	     (y1 (cond
+		    ((flonum? y)
+		     y)
+		    ((fixnum? y)
+		     (fixnum->flonum y))
+		    ((elong? y)
+		     (elong->flonum y))
+		    ((llong? y)
+		     (llong->flonum y))
+		    (else
+		     (error 'expr "not a number" y)))))
+	  (exptfl x1 y1)))))
 
 ;*---------------------------------------------------------------------*/
 ;*    exact->inexact ...                                               */
 ;*---------------------------------------------------------------------*/
 (define-inline (exact->inexact z)
-   (if (exact? z)
-       (fixnum->flonum z)
-       z))
+   (cond
+      ((fixnum? z)
+       (fixnum->flonum z))
+      ((flonum? z)
+       z)
+      ((elong? z)
+       (elong->flonum z))
+      ((llong? z)
+       (llong->flonum z))
+      (else
+       z)))
 
 ;*---------------------------------------------------------------------*/
 ;*    inexact->exact ...                                               */
@@ -827,32 +1007,102 @@
        (integer->string x radix))
       ((flonum? x)
        (real->string x))
+      ((elong? x)
+       (elong->string x radix))
+      ((llong? x)
+       (llong->string x radix))
       (else
        (error "number->string" "Argument not a number" x))))
 
 ;*---------------------------------------------------------------------*/
-;*    string->number ...                                               */
+;*    @deffn string->number@ ...                                       */
 ;*---------------------------------------------------------------------*/
 (define (string->number x . radix)
-   (define (integer-string? x)
-      (let loop ((i (-fx (string-length x) 1)))
-	 (cond
-	    ((=fx i -1)
-	     #t)
-	    ((char=? (string-ref x i) #\-)
-	     (=fx i 0))
-	    ((and (char>=? (string-ref x i) #\0)
-		  (char<=? (string-ref x i) #\9))
-	     (loop (-fx i 1)))
-	    ((and (char>=? (string-ref x i) #\a)
-		  (char<=? (string-ref x i) #\f))
-	     (loop (-fx i 1)))
-	    ((and (char>=? (string-ref x i) #\A)
-		  (char<=? (string-ref x i) #\F))
-	     (loop (-fx i 1)))
-	    (else
-	     #f))))
-   (if (integer-string? x)
-       (apply string->integer (cons x radix))
-       (apply string->real (cons x radix))))
-
+   (define (integer-string? x r)
+      (let ((len (string-length x)))
+	 (let loop ((i (-fx len 1)))
+	    (cond ((=fx -1 i)
+		   #t)
+		  ((and (char>=? (string-ref x i) #\0)
+			(char<=? (string-ref x i) #\1)
+			(>=fx r 2))
+		   (loop (-fx i 1)))
+		  ((and (char>=? (string-ref x i) #\2)
+			(char<=? (string-ref x i) #\7)
+			(>=fx r 8))
+		   (loop (-fx i 1)))
+		  ((and (char>=? (string-ref x i) #\8)
+			(char<=? (string-ref x i) #\9)
+			(>=fx r 10))
+		   (loop (-fx i 1)))
+		  ((and (char>=? (string-ref x i) #\a)
+			(char<=? (string-ref x i) #\f)
+			(=fx r 16))
+		   (loop (-fx i 1)))
+		  ((and (char>=? (string-ref x i) #\A)
+			(char<=? (string-ref x i) #\F)
+			(=fx r 16))
+		   (loop (-fx i 1)))
+		  ((or (char=? (string-ref x i) #\-)
+		       (char=? (string-ref x i) #\+))
+		   (and (=fx i 0) (>fx len 1)))
+		  (else #f)))))
+   (define (real-string? x)
+      (let ((len (string-length x)))
+	 (let loop ((i 0)
+		    (e #f)
+		    (p 0)
+		    (d #f))
+	    (cond ((=fx i len)
+		   d)
+		  ((and (char>=? (string-ref x i) #\0)
+			(char<=? (string-ref x i) #\9))
+		   (loop (+fx i 1)
+			 e
+			 0
+			 #t))
+		  ((char=? (string-ref x i) #\.)
+		   (loop (+fx i 1)
+			 e
+			 0
+			 d))
+		  ((or (char=? (string-ref x i) #\e)
+		       (char=? (string-ref x i) #\E))
+		   (if e
+		       #f
+		       (loop (+fx i 1)
+			     #t
+			     (+fx i 1)
+			     d)))
+		  ((or (char=? (string-ref x i) #\-)
+		       (char=? (string-ref x i) #\+))
+		   (and (or (=fx i 0) (=fx i p))
+			(loop (+fx i 1)
+			      e
+			      0
+			      d)))
+		  (else #f)))))
+   (let ((rx (match-case radix
+		(() 10)
+		((?val)
+		 (case val
+		    ((2 8 10 16) val)
+		    (else (error "string->number"
+				 "Illegal radix"
+				 val))))
+		(else (error "string->number"
+			     "Illegal number of optional arguments"
+			     radix)))))
+      (cond
+	 ((=fx (string-length x) 0)
+	  #f)
+	 ((integer-string? x rx)
+	  (string->integer x rx))
+	 ((real-string? x)
+	  (if (=fx rx 10)
+	      (string->real x)
+	      (error "string->number"
+		     "Only radix `10' is legal for floating point number"
+		     rx)))
+	 (else
+	  #f))))
