@@ -1,30 +1,3 @@
-;* --------------------------------------------------------------------*/
-;*    Copyright (c) 1992-1998 by Manuel Serrano. All rights reserved.  */
-;*                                                                     */
-;*                                     ,--^,                           */
-;*                               _ ___/ /|/                            */
-;*                           ,;'( )__, ) '                             */
-;*                          ;;  //   L__.                              */
-;*                          '   \   /  '                               */
-;*                               ^   ^                                 */
-;*                                                                     */
-;*                                                                     */
-;*    This program is distributed in the hope that it will be useful.  */
-;*    Use and copying of this software and preparation of derivative   */
-;*    works based upon this software are permitted, so long as the     */
-;*    following conditions are met:                                    */
-;*           o credit to the authors is acknowledged following         */
-;*             current academic behaviour                              */
-;*           o no fees or compensation are charged for use, copies,    */
-;*             or access to this software                              */
-;*           o this copyright notice is included intact.               */
-;*      This software is made available AS IS, and no warranty is made */
-;*      about the software or its performance.                         */
-;*                                                                     */
-;*      Bug descriptions, use reports, comments or suggestions are     */
-;*      welcome. Send them to                                          */
-;*        Manuel Serrano -- Manuel.Serrano@unice.fr                    */
-;*-------------------------------------------------------------------- */
 ;; ---------------------------------------------------------------------- ;;
 ;; FICHIER               : lalr.scm                                       ;;
 ;; DATE DE CREATION      : Thu Jun 29 15:07:55 1995                       ;;
@@ -39,35 +12,36 @@
    
    (include "Lalr/lalr.sch")
 
-   (import (__error                   "Llib/error.scm")
-	   (__lalr_util               "Lalr/util.scm")
-	   (__lalr_gen                "Lalr/gen.scm")
-	   (__lalr_global             "Lalr/global.scm")
-	   (__lalr_rewrite            "Lalr/rewrite.scm"))
+   (import __error
+	   __lalr_util
+	   __lalr_gen
+	   __lalr_global
+	   __lalr_rewrite
 		
-   (import (__type                    "Llib/type.scm")
-		
-	   (__bigloo                  "Llib/bigloo.scm")
-	   (__tvector                 "Llib/tvector.scm")
-	   (__structure               "Llib/struct.scm")
-	   (__tvector                 "Llib/tvector.scm")
-	   (__rgc                     "Rgc/runtime.scm")
-	   (__r4_numbers_6_5          "Ieee/number.scm")
-	   (__r4_numbers_6_5_fixnum   "Ieee/fixnum.scm")
-	   (__r4_numbers_6_5_flonum   "Ieee/flonum.scm")
-	   (__r4_characters_6_6       "Ieee/char.scm")
-	   (__r4_equivalence_6_2      "Ieee/equiv.scm")
-	   (__r4_booleans_6_1         "Ieee/boolean.scm")
-	   (__r4_symbols_6_4          "Ieee/symbol.scm")
-	   (__r4_strings_6_7          "Ieee/string.scm")
-	   (__r4_pairs_and_lists_6_3  "Ieee/pair-list.scm")
-	   (__r4_input_6_10_2         "Ieee/input.scm")
-	   (__r4_control_features_6_9 "Ieee/control.scm")
-	   (__r4_vectors_6_8          "Ieee/vector.scm")
-	   (__r4_ports_6_10_1         "Ieee/port.scm")
-	   (__r4_output_6_10_3        "Ieee/output.scm")
-	   (__bit                     "Llib/bit.scm")
-	   (__evenv                   "Eval/evenv.scm"))
+	   __type
+	   __bigloo
+	   __tvector
+	   __structure
+	   __tvector
+	   __bexit
+	   __param
+
+	   __r4_numbers_6_5
+	   __r4_numbers_6_5_fixnum
+	   __r4_numbers_6_5_flonum
+	   __r4_characters_6_6
+	   __r4_equivalence_6_2
+	   __r4_booleans_6_1
+	   __r4_symbols_6_4
+	   __r4_strings_6_7
+	   __r4_pairs_and_lists_6_3
+	   __r4_input_6_10_2
+	   __r4_control_features_6_9
+	   __r4_vectors_6_8
+	   __r4_ports_6_10_1
+	   __r4_output_6_10_3
+	   __bit
+	   __evenv)
 
    (export (expand-lalr-grammar x e)))
 
@@ -76,32 +50,41 @@
 ;*---------------------------------------------------------------------*/
 (define *default* (gensym 'default))
 
+
 ;; ---------------------------------------------------------------------- ;;
 ;; C'est le point d'entre ...                                            ;;
 ;; ---------------------------------------------------------------------- ;;
 (define (expand-lalr-grammar x e)
    (match-case x
-      ((?- . (and (? pair?) ?rules))
-       (initialize-all)
-       (rewrite-grammar! rules)
-       (pack-grammar)
-       (set-derives)
-       (set-nullable)
-       (generate-states)
-       (lalr)
-       (build-tables)
-       (compact-action-table)
-       (let ((code (gen-lalr-code)))
-	  (clean-plist)
+      ((?- . ?rules)
+       (match-case rules
+	  (((and (? list?) ?tokens) . (and (? list?) ?rules))
+	   (check-lalr-rules x tokens rules)))
+       (let ((code (unwind-protect
+		      (begin
+			 (initialize-all)
+			 (rewrite-grammar! rules)
+			 (pack-grammar)
+			 (set-derives)
+			 (set-nullable)
+			 (generate-states)
+			 (lalr)
+			 (build-tables)
+			 (compact-action-table)
+			 (gen-lalr-code))
+		      (clean-plist))))
 	  (e code e)))
       (else
        (error "lalr-grammar" "Illegal form" x))))
 
 ;; --- Encodage compact de la grammaire ... ----------------------------- ;;
+;; Also populates `rprec', which is the precedence for every rule, from 
+;; the precedence of the last terminal in the production.
 (define (pack-grammar)
   (set! rlhs   (make-vector nrules #f))
   (set! rrhs   (make-vector nrules #f))
   (set! ritem  (make-vector (+fx 1 nitems) #f))
+  (set! rprec  (make-vector nrules #f))
   
   (let loop ((p grammar) (item-no 0) (rule-no 1))
     (if (not (null? p))
@@ -118,14 +101,54 @@
 			  (vector-set! ritem it-no3 (negfx rl-no2))
 			  (loop2 (cdr prods) (+fx it-no3 1) (+fx rl-no2 1)))
 			(let* ((sym (car rhs))
-			       (sym-no (getprop (if (pair? sym)
-						    (car sym)
-						    sym)
-						'sym-no)))
-						       
+			       (asym (if (pair? sym)
+				       (car sym)
+				       sym))
+			       (sym-no (getprop asym 'sym-no)))
+
+			  (if (getprop asym 'prec)
+			    (vector-set! rprec rl-no2 (getprop asym 'prec)))
 			  (vector-set! ritem it-no3 sym-no)
 			  (loop3 (cdr rhs) (+fx it-no3 1))))))))))))
 
+;; ---------------------------------------------------------------------- ;;
+;; On verifie la syntaxe des regles
+;; ---------------------------------------------------------------------- ;;
+(define (check-lalr-rules x tokens rules)
+   (define (check-lalr-rule rule)
+      (match-case rule
+	 (((? symbol?) . (and (? list?) ?subrules))
+	  (for-each (lambda (sr)
+		       (match-case sr
+			  (((and ?sr (? list?)) . ?-)
+			   (for-each (lambda (sym)
+					(if (not (symbol? sym))
+					    (error "lalr-grammar"
+						   "Illegal form "
+						   rule)))
+				     sr))
+			  (else
+			   (error "lalr-grammar"
+				  "Illegal form (Illegal left hand side)"
+				  rule))))
+		    subrules))
+	 (else
+	  (error "lalr-grammar" "Illegal form" x))))
+   (for-each (lambda (t)
+	       (cond
+		 ((symbol? t)
+		  #t)
+		 ((and (pair? t) (memq (car t) '(left: right: none:)))
+		  (for-each (lambda (t)
+			       (if (not (symbol? t))
+				   (bigloo-type-error 'lalr-grammar
+						      'symbol
+						      t)))
+			    (cdr t)))
+		 (else
+		  (error "lalr-grammar" "Illegal token" t))))
+	     tokens)
+   (for-each check-lalr-rule rules))
 
 ;; ---------------------------------------------------------------------- ;;
 ;; Les autres fonctions sont tires du code de Bison ...                  ;;
@@ -844,35 +867,84 @@
 	    (cons (vector-ref *symv* x)
 		  (loop (+fx pos 1)))))))))
 
+(define (get-precedence Sym Act)
+  (if (<=fx Act 0) 
+    (vector-ref rprec (negfx Act))			 ; Reduce (from rule)
+    (getprop (vector-ref *symv* (+fx Sym nvars)) 'prec))); Shift (from terminal)
+
 (define (build-tables)
+
   (define (add-action St Sym Act)
-     (let* ((x (vector-ref ACTION-TABLE St))
+     (let* ((x (vector-ref action-table St))
 	    (y (assv Sym x)))
 	(if y
-	    (if (not (= Act (cdr y)))
-		;; -- there is a conflict 
-		(begin
-		   (if (and (<=fx (cdr y) 0)
-			    (<=fx Act 0))
-		       (begin
-			  (warning "lalr-grammar"
-				   "** Reduce/Reduce conflict: "
-				   #"\n - reduce by rule " (build-rule (negfx Act))
-				   #"\n - reduce by rule " (build-rule (negfx (cdr y)))
-				   #"\non token `" (vector-ref *symv* (+fx nvars Sym))
-				   "'")
-			  (set-cdr! y (max (cdr y) Act)))
-		       (begin
-			  (warning "lalr-grammar"
-				   "** Shift/Reduce conflict: "
-				   #"\n - shift to state "  Act
-				   #"\n - reduce rule " (build-rule (negfx (cdr y)))
-				   #"\non token `"  (vector-ref *symv* (+fx nvars sym))
-				   "'")
-			  (set-cdr! y Act)))))
-	    (vector-set! ACTION-TABLE St
-			 (cons (cons Sym Act) x)))))
+	  (let ((cur-prec (get-precedence Sym (cdr y)))
+		(new-prec (get-precedence Sym Act)))
+	    (cond
+	      ((=fx Act (cdr y))	  ; ** Same action as in the table.
+	       #t)
+	      ((and (<=fx (cdr y) 0)	  ; ** Reduce/Reduce conflict.
+		    (<=fx Act 0))
+	       (warning "lalr-grammar"
+			"** Reduce/Reduce conflict: "
+			#"\n - reduce by rule " (build-rule (negfx Act))
+			#"\n - reduce by rule " (build-rule (negfx (cdr y)))
+			#"\non token `" (vector-ref *symv* (+fx nvars Sym))
+			"'")
+	       (set-cdr! y (max (cdr y) Act)))
+	      ((or cur-prec new-prec)	  ; ** Resolve with precedence.
+	       (cond
+		 ((not new-prec)	  ; ** Old symbol has precedence.
+		  #t)
+		 ((not cur-prec)	  ; ** New symbol has precedence.
+		  (set-cdr! y Act))
+		 ((=fx (cdr cur-prec) (cdr new-prec))
+		  ; ** Equal precedence, use associativity.
+		  (let ((shift (max (cdr y) Act))
+			(reduce (min (cdr y) Act)))
+		    (case (car cur-prec)
+		      ((left:)
+		       (set-cdr! y reduce))
+		      ((right:)
+		       (set-cdr! y shift))
+		      ((none:)
+		       ; Remove from action table so we cause an error during
+		       ; parse.  NOTE: If we have another reduce later, we
+		       ; have no record so we can't point out the
+		       ; reduce/reduce conflict.
+		       (vector-set! action-table St
+				    (let loop ((iter (vector-ref
+						       action-table
+						       St)))
+				      (cond
+					((null? iter)
+					 iter)
+					((eqv? (car iter) St)
+					 (loop (cdr iter)))
+					(else
+					 (cons (car iter)
+					       (loop (cdr iter)))))))))))
+		 ((>fx (cdr cur-prec) (cdr new-prec))
+		  ; ** New action has higher precedence (lower precedence num).
+		  (set-cdr! y Act))
+		 (else
+		  ; ** Current action has higher precedence.  Do nothing.
+		  #t)))
+	      (else			  ; ** Shift/Reduce conflict.
+	       (warning "lalr-grammar"
+			"** Shift/Reduce conflict: "
+			#"\n - shift to state "  Act
+			#"\n - reduce rule " (build-rule (negfx (cdr y)))
+			#"\non token `"  (vector-ref *symv* (+fx nvars Sym))
+			"'")
+	       (set-cdr! y Act))))
+	  (vector-set! action-table St (cons (cons Sym Act) x)))))
 
+  (define (add-action-for-all-terminals state action)
+     (do ((i 1 (+ i 1)))
+	 ((= i nterms))
+	 (add-action state i action)))
+   
   (set! action-table (make-vector nstates '()))
 
   (do ((i 0 (+fx i 1)))  ; i = state
@@ -880,7 +952,11 @@
     (let ((red (vector-ref reduction-table i)))
       (if (and red (>=fx (red-nreds red) 1))
 	  (if (and (= (red-nreds red) 1) (vector-ref consistent i))
-	      (add-action i *default* (negfx (car (red-rules red))))
+;*---------------------------------------------------------------------*/
+;*    Patch Dominique Boucher, 5 sep 2005.                             */
+;*---------------------------------------------------------------------*/
+;;	      (add-action i *default* (negfx (car (red-rules red))))
+              (add-action-for-all-terminals i (- (car (red-rules red))))
 	      (let ((k (vector-ref lookaheads (+fx i 1))))
 		(let loop ((j (vector-ref lookaheads i)))
 		  (if (<fx j k)
