@@ -1,36 +1,9 @@
-;* --------------------------------------------------------------------*/
-;*    Copyright (c) 1992-1998 by Manuel Serrano. All rights reserved.  */
-;*                                                                     */
-;*                                     ,--^,                           */
-;*                               _ ___/ /|/                            */
-;*                           ,;'( )__, ) '                             */
-;*                          ;;  //   L__.                              */
-;*                          '   \   /  '                               */
-;*                               ^   ^                                 */
-;*                                                                     */
-;*                                                                     */
-;*    This program is distributed in the hope that it will be useful.  */
-;*    Use and copying of this software and preparation of derivative   */
-;*    works based upon this software are permitted, so long as the     */
-;*    following conditions are met:                                    */
-;*           o credit to the authors is acknowledged following         */
-;*             current academic behaviour                              */
-;*           o no fees or compensation are charged for use, copies,    */
-;*             or access to this software                              */
-;*           o this copyright notice is included intact.               */
-;*      This software is made available AS IS, and no warranty is made */
-;*      about the software or its performance.                         */
-;*                                                                     */
-;*      Bug descriptions, use reports, comments or suggestions are     */
-;*      welcome. Send them to                                          */
-;*        Manuel Serrano -- Manuel.Serrano@unice.fr                    */
-;*-------------------------------------------------------------------- */
 ;*=====================================================================*/
 ;*    serrano/prgm/project/bigloo/runtime/Llib/unicode.scm             */
 ;*    -------------------------------------------------------------    */
 ;*    Author      :  Manuel Serrano                                    */
 ;*    Creation    :  Mon Mar 20 19:17:18 1995                          */
-;*    Last change :  Fri Feb  6 16:58:46 1998 (serrano)                */
+;*    Last change :  Sat Dec 17 14:48:38 2005 (serrano)                */
 ;*    -------------------------------------------------------------    */
 ;*    Unicode (UCS-2) strings handling.                                */
 ;*=====================================================================*/
@@ -39,48 +12,53 @@
 ;*    The module                                                       */
 ;*---------------------------------------------------------------------*/
 (module __unicode
-
-   (import  (__error                   "Llib/error.scm"))
    
-   (use     (__type                    "Llib/type.scm")
-	    (__bigloo                  "Llib/bigloo.scm")
-	    (__tvector                 "Llib/tvector.scm")
-	    (__ucs2                    "Llib/ucs2.scm")
+   (import  __error
+	    __param)
+   
+   (use     __type
+	    __bigloo
+	    __tvector
+	    __ucs2
 	    
-	    (__r4_numbers_6_5_fixnum   "Ieee/fixnum.scm")
-	    (__r4_equivalence_6_2      "Ieee/equiv.scm")
-	    (__r4_vectors_6_8          "Ieee/vector.scm")
-	    (__r4_booleans_6_1         "Ieee/boolean.scm")
-	    (__r4_characters_6_6       "Ieee/char.scm")
-	    (__r4_symbols_6_4          "Ieee/symbol.scm")
-	    (__r4_pairs_and_lists_6_3  "Ieee/pair-list.scm")
-	    (__r4_strings_6_7          "Ieee/string.scm")
-
-	    (__evenv                   "Eval/evenv.scm"))
+	    __r4_ports_6_10_1
+	    __r4_output_6_10_3
+	    
+	    __r4_numbers_6_5_fixnum
+	    __r4_numbers_6_5_flonum
+	    __r4_equivalence_6_2
+	    __r4_vectors_6_8
+	    __r4_booleans_6_1
+	    __r4_characters_6_6
+	    __r4_symbols_6_4
+	    __r4_pairs_and_lists_6_3
+	    __r4_strings_6_7
+	    
+	    __evenv)
    
-   (extern (macro c-ucs2-string?::bool       (::obj)
+   (extern (macro c-ucs2-string?::bool (::obj)
 		  "UCS2_STRINGP")
-	   (c-make-ucs2-string::ucs2string   (::long ::ucs2)
-		  "make_ucs2_string")
-	   (macro c-ucs2-string-length::long (::ucs2string)
+	   (c-make-ucs2-string::ucs2string (::int ::ucs2)
+					   "make_ucs2_string")
+	   (macro c-ucs2-string-length::int (::ucs2string)
 		  "UCS2_STRING_LENGTH")
-	   (macro c-ucs2-string-ref::ucs2    (::ucs2string ::long)
+	   (macro c-ucs2-string-ref::ucs2 (::ucs2string ::int)
 		  "UCS2_STRING_REF")
-	   (macro c-ucs2-string-set!::obj    (::ucs2string ::long ::ucs2)
+	   (macro c-ucs2-string-set!::obj (::ucs2string ::int ::ucs2)
 		  "UCS2_STRING_SET")
-	    
-	   (c-ucs2-string=?::bool  (::ucs2string ::ucs2string)
-				   "ucs2_strcmp")
-	   (ucs2-strcicmp::bool    (::ucs2string ::ucs2string)
-				   "ucs2_strcicmp")
-	   (ucs2-string_lt::bool   (::ucs2string ::ucs2string)
-				   "ucs2_string_lt")
-	   (ucs2-string_le::bool   (::ucs2string ::ucs2string)
-				   "ucs2_string_le")
-	   (ucs2-string_gt::bool   (::ucs2string ::ucs2string)
-				   "ucs2_string_gt")
-	   (ucs2-string_ge::bool   (::ucs2string ::ucs2string)
-				   "ucs2_string_ge")
+	   
+	   (c-ucs2-string=?::bool (::ucs2string ::ucs2string)
+				  "ucs2_strcmp")
+	   (ucs2-strcicmp::bool (::ucs2string ::ucs2string)
+				"ucs2_strcicmp")
+	   (ucs2-string_lt::bool (::ucs2string ::ucs2string)
+				 "ucs2_string_lt")
+	   (ucs2-string_le::bool (::ucs2string ::ucs2string)
+				 "ucs2_string_le")
+	   (ucs2-string_gt::bool (::ucs2string ::ucs2string)
+				 "ucs2_string_gt")
+	   (ucs2-string_ge::bool (::ucs2string ::ucs2string)
+				 "ucs2_string_ge")
 	   (ucs2-string_cilt::bool (::ucs2string ::ucs2string)
 				   "ucs2_string_cilt")
 	   (ucs2-string_cile::bool (::ucs2string ::ucs2string)
@@ -89,72 +67,118 @@
 				   "ucs2_string_cigt")
 	   (ucs2-string_cige::bool (::ucs2string ::ucs2string)
 				   "ucs2_string_cige")
-
-	   (escape-C-string::ucs2string      (::string)
-					     "escape_ucs2_string")
-	   (c-ucs2-string-copy::ucs2string   (::ucs2string)
-					     "c_ucs2_string_copy")
-	   (c-subucs2-string::ucs2string     (::ucs2string ::long ::long)
-					     "c_subucs2_string")
+	   
+	   (c-ucs2-string-copy::ucs2string (::ucs2string)
+					   "c_ucs2_string_copy")
+	   (c-subucs2-string::ucs2string (::ucs2string ::int ::int)
+					 "c_subucs2_string")
 	   (c-ucs2-string-append::ucs2string (::ucs2string ::ucs2string)
 					     "ucs2_string_append")
-	    
-	   (ucs2-string-bound-check?::bool      (::long ::long)
-						"BOUND_CHECK")
+	   
 	   (c-ucs2-string->utf8-string::bstring (::ucs2string)
 						"ucs2_string_to_utf8_string")
 	   (c-utf8-string->ucs2-string::ucs2string (::bstring)
-						"utf8_string_to_ucs2_string"))
+						   "utf8_string_to_ucs2_string"))
    
-   (export  (inline ucs2-string?::bool             ::obj)
-	    (inline make-ucs2-string::ucs2string   ::int . ucs2)
-	    (inline ucs2-string::ucs2string        . ucs2s)
-	    (inline ucs2-string-length::long       ::ucs2string)
-	    (inline ucs2-string-ref::ucs2          ::ucs2string ::long)
-	    (inline ucs2-string-set!::obj          ::ucs2string ::long ::ucs2)
-	    (inline ucs2-string-ref-ur::ucs2       ::ucs2string ::long)
-	    (inline ucs2-string-set-ur!::obj       ::ucs2string ::long ::ucs2)
-	    (inline ucs2-string=?::bool            ::ucs2string ::ucs2string)
-	    (inline ucs2-string-ci=?::bool         ::ucs2string ::ucs2string)
-	    (inline ucs2-string<?::bool            ::ucs2string ::ucs2string)
-	    (inline ucs2-string>?::bool            ::ucs2string ::ucs2string)
-	    (inline ucs2-string<=?::bool           ::ucs2string ::ucs2string)
-	    (inline ucs2-string>=?::bool           ::ucs2string ::ucs2string)
-	    (inline ucs2-string-ci<?::bool         ::ucs2string ::ucs2string)
-	    (inline ucs2-string-ci>?::bool         ::ucs2string ::ucs2string)
-	    (inline ucs2-string-ci<=?::bool        ::ucs2string ::ucs2string)
-	    (inline ucs2-string-ci>=?::bool        ::ucs2string ::ucs2string)
-	    (inline subucs2-string::ucs2string     ::ucs2string ::long ::long)
-	    (inline subucs2-string-ur::ucs2string  ::ucs2string ::long ::long)
-	    (ucs2-string-append::ucs2string        . ucs2-strings)
-	    (ucs2-string->list                     ::ucs2string)
-	    (list->ucs2-string::ucs2string         ::obj)
-	    (inline ucs2-string-copy::ucs2string   ::ucs2string)
-	    (ucs2-string-fill!::ucs2string         ::ucs2string ::ucs2)
-	    (ucs2-string-upcase::ucs2string        ::ucs2string)
-	    (ucs2-string-downcase::ucs2string      ::ucs2string)
-	    (ucs2-string-upcase!::ucs2string       ::ucs2string)
-	    (ucs2-string-downcase!::ucs2string     ::ucs2string)
-	    (inline ucs2-string->utf8-string::bstring    ::ucs2string)
-	    (inline utf8-string->ucs2-string::ucs2string ::bstring))
-    
-   (pragma  (c-ucs2-string? (predicate-of ucs2string) no-cfa-top)
-	    (ucs2-string? side-effect-free no-cfa-top)
-	    (c-ucs2-string-ref side-effect-free no-cfa-top)
-	    (ucs2-string-ref-ur side-effect-free no-cfa-top)
-	    (ucs2-string-ref side-effect-free no-cfa-top)
-	    (c-ucs2-string-length side-effect-free no-cfa-top)
-	    (ucs2-string-length side-effect-free no-cfa-top)
-	    (ucs2-string=? side-effect-free)
-	    (ucs2-string-ci=? side-effect-free)
-	    (ucs2-string<? side-effect-free)
-	    (ucs2-string>? side-effect-free)
-	    (ucs2-string<=? side-effect-free)
-	    (ucs2-string>=? side-effect-free)
-	    (ucs2-string-ci<? side-effect-free)
-	    (ucs2-string-ci>? side-effect-free)
-	    (ucs2-string-ci<=? side-effect-free)
-	    (ucs2-string-ci>=? side-effect-free)))
+   (java   (class foreign
+	      (method static c-ucs2-string?::bool (::obj)
+		      "UCS2_STRINGP")
+	      (method static c-make-ucs2-string::ucs2string (::int ::ucs2)
+		      "make_ucs2_string")
+	      (method static c-ucs2-string-length::int (::ucs2string)
+		      "UCS2_STRING_LENGTH")
+	      (method static c-ucs2-string-ref::ucs2 (::ucs2string ::int)
+		      "UCS2_STRING_REF")
+	      (method static c-ucs2-string-set!::obj (::ucs2string ::int ::ucs2)
+		      "UCS2_STRING_SET")
+	      
+	      (method static c-ucs2-string=?::bool (::ucs2string ::ucs2string)
+		      "ucs2_strcmp")
+	      (method static ucs2-strcicmp::bool (::ucs2string ::ucs2string)
+		      "ucs2_strcicmp")
+	      (method static ucs2-string_lt::bool (::ucs2string ::ucs2string)
+		      "ucs2_string_lt")
+	      (method static ucs2-string_le::bool (::ucs2string ::ucs2string)
+		      "ucs2_string_le")
+	      (method static ucs2-string_gt::bool (::ucs2string ::ucs2string)
+		      "ucs2_string_gt")
+	      (method static ucs2-string_ge::bool (::ucs2string ::ucs2string)
+		      "ucs2_string_ge")
+	      (method static ucs2-string_cilt::bool (::ucs2string ::ucs2string)
+		      "ucs2_string_cilt")
+	      (method static ucs2-string_cile::bool (::ucs2string ::ucs2string)
+		      "ucs2_string_cile")
+	      (method static ucs2-string_cigt::bool (::ucs2string ::ucs2string)
+		      "ucs2_string_cigt")
+	      (method static ucs2-string_cige::bool (::ucs2string ::ucs2string)
+		      "ucs2_string_cige")
+	      
+	      (method static c-ucs2-string-copy::ucs2string (::ucs2string)
+		      "c_ucs2_string_copy")
+	      (method static c-subucs2-string::ucs2string (::ucs2string ::int ::int)
+		      "c_subucs2_string")
+	      (method static c-ucs2-string-append::ucs2string (::ucs2string ::ucs2string)
+		      "c_ucs2_string_append")
+	      
+	      (method static c-ucs2-string->utf8-string::bstring (::ucs2string)
+		      "ucs2_string_to_utf8_string")
+	      (method static c-utf8-string->ucs2-string::ucs2string (::bstring)
+		      "utf8_string_to_ucs2_string")))
+   
+   (export  (inline ucs2-string?::bool ::obj)
+	    (inline make-ucs2-string::ucs2string ::int
+		    #!optional (filler::ucs2 (char->ucs2 #\space)))
+	    (inline ucs2-string::ucs2string . ucs2s)
+	    (inline ucs2-string-length::int ::ucs2string)
+	    (inline ucs2-string-ref::ucs2 ::ucs2string ::int)
+	    (inline ucs2-string-set!::obj ::ucs2string ::int ::ucs2)
+	    (inline ucs2-string-ref-ur::ucs2 ::ucs2string ::int)
+	    (inline ucs2-string-set-ur!::obj ::ucs2string ::int ::ucs2)
+	    (inline ucs2-string=?::bool ::ucs2string ::ucs2string)
+	    (inline ucs2-string-ci=?::bool ::ucs2string ::ucs2string)
+	    (inline ucs2-string<?::bool ::ucs2string ::ucs2string)
+	    (inline ucs2-string>?::bool ::ucs2string ::ucs2string)
+	    (inline ucs2-string<=?::bool ::ucs2string ::ucs2string)
+	    (inline ucs2-string>=?::bool ::ucs2string ::ucs2string)
+	    (inline ucs2-string-ci<?::bool ::ucs2string ::ucs2string)
+	    (inline ucs2-string-ci>?::bool ::ucs2string ::ucs2string)
+	    (inline ucs2-string-ci<=?::bool ::ucs2string ::ucs2string)
+	    (inline ucs2-string-ci>=?::bool ::ucs2string ::ucs2string)
+	    (inline subucs2-string::ucs2string ::ucs2string ::int ::int)
+	    (inline subucs2-string-ur::ucs2string ::ucs2string ::int ::int)
+	    (ucs2-string-append::ucs2string . ucs2-strings)
+	    (ucs2-string->list ::ucs2string)
+	    (list->ucs2-string::ucs2string ::obj)
+	    (inline ucs2-string-copy::ucs2string ::ucs2string)
+	    (ucs2-string-fill!::ucs2string ::ucs2string ::ucs2)
+	    (ucs2-string-upcase::ucs2string ::ucs2string)
+	    (ucs2-string-downcase::ucs2string ::ucs2string)
+	    (ucs2-string-upcase!::ucs2string ::ucs2string)
+	    (ucs2-string-downcase!::ucs2string ::ucs2string)
+	    (inline ucs2-string->utf8-string::bstring ::ucs2string)
+	    (inline utf8-string->ucs2-string::ucs2string ::bstring)
+	    (utf8->iso-latin::bstring ::bstring)
+	    (utf8->iso-latin!::bstring ::bstring)
+	    (iso-latin->utf8::bstring ::bstring)
+	    (iso-latin->utf8!::bstring ::bstring))
+   
+   (pragma  (c-ucs2-string? (predicate-of ucs2string) no-cfa-top nesting)
+	    (ucs2-string? side-effect-free no-cfa-top nesting)
+	    (c-ucs2-string-ref side-effect-free no-cfa-top nesting)
+	    (ucs2-string-ref-ur side-effect-free no-cfa-top nesting)
+	    (ucs2-string-ref side-effect-free no-cfa-top nesting)
+	    (c-ucs2-string-length side-effect-free no-cfa-top nesting)
+	    (ucs2-string-length side-effect-free no-cfa-top nesting)
+	    (ucs2-string=? side-effect-free nesting)
+	    (ucs2-string-ci=? side-effect-free nesting)
+	    (ucs2-string<? side-effect-free nesting)
+	    (ucs2-string>? side-effect-free nesting)
+	    (ucs2-string<=? side-effect-free nesting)
+	    (ucs2-string>=? side-effect-free nesting)
+	    (ucs2-string-ci<? side-effect-free nesting)
+	    (ucs2-string-ci>? side-effect-free nesting)
+	    (ucs2-string-ci<=? side-effect-free nesting)
+	    (ucs2-string-ci>=? side-effect-free nesting)))
   
 ;*---------------------------------------------------------------------*/
 ;*    ucs2-string? ...                                                 */
@@ -165,10 +189,9 @@
 ;*---------------------------------------------------------------------*/
 ;*    make-ucs2-string ...                                             */
 ;*---------------------------------------------------------------------*/
-(define-inline (make-ucs2-string k . ucs2)
-   (if (null? ucs2)
-       (c-make-ucs2-string k (char->ucs2 #\space))
-       (c-make-ucs2-string k (car ucs2))))
+(define-inline (make-ucs2-string k #!optional
+				 (filler::ucs2 (char->ucs2 #\space)))
+   (c-make-ucs2-string k filler))
  
 ;*---------------------------------------------------------------------*/
 ;*    ucs2-string ...                                                  */
@@ -188,7 +211,12 @@
 (define-inline (ucs2-string-ref ucs2-string k)
    (if (string-bound-check? k (ucs2-string-length ucs2-string))
        (c-ucs2-string-ref ucs2-string k)
-       (error "ucs2-string-ref" "index out of range" k)))
+       (error 'ucs2-string-ref
+	      (string-append "index out of range [0.."
+			     (integer->string
+			      (-fx (ucs2-string-length ucs2-string) 1))
+			     "]")
+	      k)))
  
 ;*---------------------------------------------------------------------*/
 ;*    ucs2-string-set! ...                                             */
@@ -196,7 +224,12 @@
 (define-inline (ucs2-string-set! ucs2-string k ucs2)
    (if (string-bound-check? k (ucs2-string-length ucs2-string))
        (c-ucs2-string-set! ucs2-string k ucs2)
-       (error "ucs2-string-set!" "index out of range" k)))
+       (error 'ucs2-string-set!
+	      (string-append "index out of range [0.."
+			     (integer->string
+			      (-fx (ucs2-string-length ucs2-string) 1))
+			     "]")
+	      k)))
 
 ;*---------------------------------------------------------------------*/
 ;*    ucs2-string-ref-ur ...                                           */
@@ -422,3 +455,142 @@
 ;*---------------------------------------------------------------------*/
 (define-inline (utf8-string->ucs2-string utf8)
    (c-utf8-string->ucs2-string utf8))
+
+;*---------------------------------------------------------------------*/
+;*    utf8->iso-latin-length ...                                       */
+;*---------------------------------------------------------------------*/
+(define (utf8->iso-latin-length str len)
+   (let loop ((size 0)
+	      (i 0))
+      (cond
+	 ((>=fx i len)
+	  size)
+	 ((char=? (string-ref str i) (integer->char #xc2))
+	  (loop (+fx size 1) (+fx i 2)))
+	 ((char=? (string-ref str i) (integer->char #xc3))
+	  (loop (+fx size 1) (+fx i 2)))
+	 (else
+	  (loop (+fx size 1) (+fx i 1))))))
+
+;*---------------------------------------------------------------------*/
+;*    utf8->iso-latin-fill! ...                                        */
+;*---------------------------------------------------------------------*/
+(define (utf8->iso-latin-fill! nstr str len)
+   (let loop ((r 0)
+	      (w 0))
+      (if (=fx r len)
+	  nstr
+	  (let ((c (char->integer (string-ref str r))))
+	     (case c
+		((#xc2)
+		 (if (=fx r (-fx len 1))
+		     (error 'utf8->iso-latin
+			    "String too short"
+			    (string-for-read
+			     (substring
+			      str (maxfx 0 (-fx r 10)) len)))
+		     (let ((nc (string-ref str (+fx r 1))))
+			(if (and (>=fx (char->integer nc) #x80)
+				 (<fx (char->integer nc) #xc0))
+			    (begin
+			       (string-set! nstr w nc)
+			       (loop (+fx r 2) (+fx w 1)))
+			    (error 'utf8->iso-latin
+			 	   "Illegal string"
+				   (string-for-read
+				    (substring
+				     str r (minfx len (+fx r 10)))))))))
+		((#xc3)
+		 (if (=fx r (-fx len 1))
+		     (error 'utf8->iso-latin
+			    "String too short"
+			    (string-for-read
+			     (substring
+			      str (maxfx 0 (-fx r 10)) len)))
+		     (let ((inc (char->integer (string-ref str (+fx r 1)))))
+			(if (>=fx inc #x80)
+			    (let ((v (+fx inc #x40)))
+			       (if (>fx v 255)
+				   (error 'utf8->iso-latin
+					  "Can't convert to iso-latin"
+					  inc)
+				   (string-set! nstr w (integer->char v)))
+			       (loop (+fx r 2) (+fx w 1)))
+			    (error 'utf8->iso-latin
+				   "Illegal string"
+				   inc)))))
+		(else
+		 (string-set! nstr w (integer->char c))
+		 (loop (+fx r 1) (+fx w 1))))))))
+   
+;*---------------------------------------------------------------------*/
+;*    utf8->iso-latin ...                                              */
+;*---------------------------------------------------------------------*/
+(define (utf8->iso-latin str)
+   (let* ((len (string-length str))
+	  (nlen (utf8->iso-latin-length str len)))
+      (utf8->iso-latin-fill! (make-string nlen) str len)))
+
+;*---------------------------------------------------------------------*/
+;*    utf8->iso-latin! ...                                             */
+;*---------------------------------------------------------------------*/
+(define (utf8->iso-latin! str)
+   (let* ((len (string-length str))
+	  (nlen (utf8->iso-latin-length str len)))
+      (if (=fx len nlen)
+	  str
+	  (utf8->iso-latin-fill! (make-string nlen) str len))))
+
+;*---------------------------------------------------------------------*/
+;*    iso-latin->utf8-length ...                                       */
+;*---------------------------------------------------------------------*/
+(define (iso-latin->utf8-length str len)
+   (let loop ((size 0)
+	      (i 0))
+      (cond
+	 ((=fx i len)
+	  size)
+	 ((char>=? (string-ref-ur str i) #a128)
+	  (loop (+fx size 2) (+fx i 1)))
+	 (else
+	  (loop (+fx size 1) (+fx i 1))))))
+
+;*---------------------------------------------------------------------*/
+;*    iso-latin->utf8-fill! ...                                        */
+;*---------------------------------------------------------------------*/
+(define (iso-latin->utf8-fill! nstr str len)
+   (let loop ((r 0)
+	      (w 0))
+      (if (=fx r len)
+	  nstr
+	  (let ((c (char->integer (string-ref-ur str r))))
+	     (cond
+		((>=fx c #xc0)
+		 (string-set-ur! nstr w (integer->char #xc3))
+		 (string-set-ur! nstr (+fx w 1) (integer->char (-fx c #x40)))
+		 (loop (+fx r 1) (+fx w 2)))
+		((>=fx c #x80)
+		 (string-set-ur! nstr w (integer->char #xc2))
+		 (string-set-ur! nstr (+fx w 1) (integer->char c))
+		 (loop (+fx r 1) (+fx w 2)))
+		(else
+		 (string-set-ur! nstr w (integer->char c))
+		 (loop (+fx r 1) (+fx w 1))))))))
+
+;*---------------------------------------------------------------------*/
+;*    iso-latin->utf8 ...                                              */
+;*---------------------------------------------------------------------*/
+(define (iso-latin->utf8 str)
+   (let* ((len (string-length str))
+	  (nlen (iso-latin->utf8-length str len)))
+      (iso-latin->utf8-fill! (c-make-string/wo-fill nlen) str len)))
+   
+;*---------------------------------------------------------------------*/
+;*    iso-latin->utf8! ...                                             */
+;*---------------------------------------------------------------------*/
+(define (iso-latin->utf8! str)
+   (let* ((len (string-length str))
+	  (nlen (iso-latin->utf8-length str len)))
+      (if (=fx len nlen)
+	  str
+	  (iso-latin->utf8-fill! (c-make-string/wo-fill nlen) str len))))
