@@ -1,36 +1,9 @@
-;* --------------------------------------------------------------------*/
-;*    Copyright (c) 1992-1998 by Manuel Serrano. All rights reserved.  */
-;*                                                                     */
-;*                                     ,--^,                           */
-;*                               _ ___/ /|/                            */
-;*                           ,;'( )__, ) '                             */
-;*                          ;;  //   L__.                              */
-;*                          '   \   /  '                               */
-;*                               ^   ^                                 */
-;*                                                                     */
-;*                                                                     */
-;*    This program is distributed in the hope that it will be useful.  */
-;*    Use and copying of this software and preparation of derivative   */
-;*    works based upon this software are permitted, so long as the     */
-;*    following conditions are met:                                    */
-;*           o credit to the authors is acknowledged following         */
-;*             current academic behaviour                              */
-;*           o no fees or compensation are charged for use, copies,    */
-;*             or access to this software                              */
-;*           o this copyright notice is included intact.               */
-;*      This software is made available AS IS, and no warranty is made */
-;*      about the software or its performance.                         */
-;*                                                                     */
-;*      Bug descriptions, use reports, comments or suggestions are     */
-;*      welcome. Send them to                                          */
-;*        Manuel Serrano -- Manuel.Serrano@unice.fr                    */
-;*-------------------------------------------------------------------- */
 ;*---------------------------------------------------------------------*/
-;*    serrano/prgm/project/bigloo/runtime1.8/Match/compiler.scm        */
+;*    serrano/prgm/project/bigloo/runtime/Match/compiler.scm           */
 ;*                                                                     */
 ;*    Author      :  Jean-Marie Geffroy                                */
 ;*    Creation    :  Thu Jan 14 10:29:50 1993                          */
-;*    Last change :  Mon Jul 10 14:28:12 1995 (serrano)                */
+;*    Last change :  Mon Jul 31 16:49:23 2006 (serrano)                */
 ;*                                                                     */
 ;*---------------------------------------------------------------------*/
 ;*    A hand-written pattern "compiler"                                */
@@ -54,31 +27,32 @@
    
    (export  (pcompile f))
 
-   (import  (__error                   "Llib/error.scm")
-	    (__match_s2cfun            "Match/s2cfun.scm")
-	    (__match_descriptions      "Match/descr.scm"))
+   (import  __error
+	    __match_s2cfun
+	    __match_descriptions
+	    __param)
 
-   (use     (__type                    "Llib/type.scm")
-	    (__bigloo                  "Llib/bigloo.scm")
-	    (__tvector                 "Llib/tvector.scm")
-	    (__structure               "Llib/struct.scm")
-	    (__tvector                 "Llib/tvector.scm")
-	    (__rgc                     "Rgc/runtime.scm")
-	    (__r4_numbers_6_5          "Ieee/number.scm")
-	    (__r4_numbers_6_5_fixnum   "Ieee/fixnum.scm")
-	    (__r4_numbers_6_5_flonum   "Ieee/flonum.scm")
-	    (__r4_characters_6_6       "Ieee/char.scm")
-	    (__r4_equivalence_6_2      "Ieee/equiv.scm")
-	    (__r4_booleans_6_1         "Ieee/boolean.scm")
-	    (__r4_symbols_6_4          "Ieee/symbol.scm")
-	    (__r4_strings_6_7          "Ieee/string.scm")
-	    (__r4_pairs_and_lists_6_3  "Ieee/pair-list.scm")
-	    (__r4_input_6_10_2         "Ieee/input.scm")
-	    (__r4_control_features_6_9 "Ieee/control.scm")
-	    (__r4_vectors_6_8          "Ieee/vector.scm")
-	    (__r4_ports_6_10_1         "Ieee/port.scm")
-	    (__r4_output_6_10_3        "Ieee/output.scm")
-	    (__evenv                   "Eval/evenv.scm")))
+   (use     __type
+	    __bigloo
+	    __tvector
+	    __structure
+	    __tvector
+
+	    __r4_numbers_6_5
+	    __r4_numbers_6_5_fixnum
+	    __r4_numbers_6_5_flonum
+	    __r4_characters_6_6
+	    __r4_equivalence_6_2
+	    __r4_booleans_6_1
+	    __r4_symbols_6_4
+	    __r4_strings_6_7
+	    __r4_pairs_and_lists_6_3
+	    __r4_input_6_10_2
+	    __r4_control_features_6_9
+	    __r4_vectors_6_8
+	    __r4_ports_6_10_1
+	    __r4_output_6_10_3
+	    __evenv))
 	    
 (define (pcompile f)
    (let ((s (jim-gensym "E-")))
@@ -132,10 +106,31 @@
        ,(k r z d)  
        ,(z d)))
 
+(define (build-atom-equality-test e const)
+   (cond
+      ;; John Malecki improvement (15 aug 2003)
+      ((null? const)
+       `(null? ,e))
+      ((or (fixnum? const)
+	   (char? const)
+	   (boolean? const)
+	   (symbol? const))
+       `(eq? ,e (quote ,const)))
+      ((flonum? const)
+       `(and (flonum? ,e) (=fl ,e ,const)))
+      ((number? const)
+       `(and (number? ,e) (= ,e ,const)))
+      ((string? const)
+       `(and (string? ,e) (string=? ,e ,const)))
+      ((null? const)
+       `(null? ,e))
+      (else
+       `(equal? ,e (quote ,const)))))
+
 (define (compile-quote f e r m k z d)
-  (build-if `(eq? ,e (quote ,(cadr f)))
-	    (k r z `(quote ,(cadr f)))
-	    (z (pattern-minus d f))))
+   (build-if (build-atom-equality-test e (cadr f))
+	     (k r z `(quote ,(cadr f)))
+	     (z (pattern-minus d f))))
 
 (define (compile-not f e r m k z d)
    (compile f e r m (lambda (r2 z2 d2) (z d2)) 
@@ -266,20 +261,6 @@
 		 (pattern-car c) ) ) )
       (build-let *car* *cdr* e body) ) )
 
-;*  	  (body (compile   */
-;*  		 f1 `(car ,e) r m  */
-;*  		 (lambda (r2 z2 c2)  */
-;*  		    (compile f2 `(cdr ,e) r2 m  */
-;*  			     (lambda (r3 z3 c3)   */
-;*  				(k r3 z3 (list 'cons c2 c3) ) )  */
-;*  			     (lambda (c3) (z (list 'cons c2 c3)))  */
-;*  			     (pattern-cdr c)) )  */
-;*  		 (lambda (c2) (z (list 'cons c2 (pattern-cdr c))))  */
-;*  		 (pattern-car c) ) )  */
-;*  	  )  */
-;*        body))  */
-
-
 ;*---------------------------------------------------------------------*/
 ;*    Instanciate-try may be re-written in order to avoid generating   */
 ;*    more than X specialized versions of a recursive pattern: we just */
@@ -295,7 +276,7 @@
 
 (define (compile-times n f1 f2 e r m k0 z0 d0)
 
-   (let ((F-Env '()) (D-env '()))
+   (let ((F-env '()) (D-env '()))
       (letrec
 	    ((instanciate-try
 	      (lambda (r m k z d)
@@ -477,7 +458,8 @@
 
 
 (define (extend fn pt im)
-   (lambda (x) (if (eq? x pt) im (fn x))))
+   (lambda (x)
+      (if (eq? x pt) im (fn x))))
 
 (define (build-if tst then else)
    (cond
