@@ -1,30 +1,3 @@
-;* --------------------------------------------------------------------*/
-;*    Copyright (c) 1992-1998 by Manuel Serrano. All rights reserved.  */
-;*                                                                     */
-;*                                     ,--^,                           */
-;*                               _ ___/ /|/                            */
-;*                           ,;'( )__, ) '                             */
-;*                          ;;  //   L__.                              */
-;*                          '   \   /  '                               */
-;*                               ^   ^                                 */
-;*                                                                     */
-;*                                                                     */
-;*    This program is distributed in the hope that it will be useful.  */
-;*    Use and copying of this software and preparation of derivative   */
-;*    works based upon this software are permitted, so long as the     */
-;*    following conditions are met:                                    */
-;*           o credit to the authors is acknowledged following         */
-;*             current academic behaviour                              */
-;*           o no fees or compensation are charged for use, copies,    */
-;*             or access to this software                              */
-;*           o this copyright notice is included intact.               */
-;*      This software is made available AS IS, and no warranty is made */
-;*      about the software or its performance.                         */
-;*                                                                     */
-;*      Bug descriptions, use reports, comments or suggestions are     */
-;*      welcome. Send them to                                          */
-;*        Manuel Serrano -- Manuel.Serrano@unice.fr                    */
-;*-------------------------------------------------------------------- */
 ;;;--------------------------------------------------------------------*/
 ;;;   geffroy/Match3.0/descriptions.scm ...                            */
 ;;;   Author      :  Jean-Marie Geffroy                                */
@@ -71,30 +44,31 @@
 	    (inline isVector-end? c)
 	    (inline isTree? c) )
 
-   (import  (__error                   "Llib/error.scm")
-	    (__match_s2cfun            "Match/s2cfun.scm"))
-   
-   (use     (__type                    "Llib/type.scm")
-	    (__bigloo                  "Llib/bigloo.scm")
-	    (__tvector                 "Llib/tvector.scm")
-	    (__structure               "Llib/struct.scm")
-	    (__tvector                 "Llib/tvector.scm")
-	    (__rgc                     "Rgc/runtime.scm")
-	    (__r4_numbers_6_5          "Ieee/number.scm")
-	    (__r4_numbers_6_5_fixnum   "Ieee/fixnum.scm")
-	    (__r4_numbers_6_5_flonum   "Ieee/flonum.scm")
-	    (__r4_characters_6_6       "Ieee/char.scm")
-	    (__r4_equivalence_6_2      "Ieee/equiv.scm")
-	    (__r4_booleans_6_1         "Ieee/boolean.scm")
-	    (__r4_symbols_6_4          "Ieee/symbol.scm")
-	    (__r4_strings_6_7          "Ieee/string.scm")
-	    (__r4_pairs_and_lists_6_3  "Ieee/pair-list.scm")
-	    (__r4_input_6_10_2         "Ieee/input.scm")
-	    (__r4_control_features_6_9 "Ieee/control.scm")
-	    (__r4_vectors_6_8          "Ieee/vector.scm")
-	    (__r4_ports_6_10_1         "Ieee/port.scm")
-	    (__r4_output_6_10_3        "Ieee/output.scm")
-	    (__evenv                   "Eval/evenv.scm")))
+   (import  __error
+	    __match_s2cfun
+	    __param)
+  
+   (use     __type
+	    __bigloo
+	    __tvector
+	    __structure
+	    __tvector
+
+	    __r4_numbers_6_5
+	    __r4_numbers_6_5_fixnum
+	    __r4_numbers_6_5_flonum
+	    __r4_characters_6_6
+	    __r4_equivalence_6_2
+	    __r4_booleans_6_1
+	    __r4_symbols_6_4
+	    __r4_strings_6_7
+	    __r4_pairs_and_lists_6_3
+	    __r4_input_6_10_2
+	    __r4_control_features_6_9
+	    __r4_vectors_6_8
+	    __r4_ports_6_10_1
+	    __r4_output_6_10_3
+	    __evenv))
 	    
 ;;;--------------------------------------------------------------------*/
 ;;;   Renvoie la liste des variables apparaissant dans un filtre       */
@@ -355,7 +329,7 @@
    (cond
       ((or (isAny? descr)
 	   (isAny? pat)
-	   (isOr? pat) (isT-Or? pat) (isTagged-or? pat)
+	   (isOr? pat) (isT-Or? pat) (isTagged-Or? pat)
 	   (isSuccess? pat)
 	   (isCheck? pat)
 	   (isTimes? pat)
@@ -384,7 +358,7 @@
       ((and (isVar? descr) (isVar? pat))
        (if (eq? (env (cadr descr)) 'unbound)
 	   (if (eq? (env (cadr pat)) 'unbound)
-	       (let ((s (list 'quote (jim-gensym))))
+	       (let ((s (list 'var (jim-gensym "VAR-")))) ;;; johnm
 		  (k (extend (extend env (cadr descr) s)
 			     (cadr pat)
 			     s)))
@@ -431,11 +405,11 @@
 ;;;   Filtrage d'une expression par une DESCRIPION                     */
 ;;;--------------------------------------------------------------------*/
 (define (match d e env k z)
-
    (case (car d)
       ((any)   (k env))
       
-      ((quote) (if (eq? e (cadr d)) (k env) (z env)))
+      ((quote) (if (or (eq? e (cadr d)) (and (string? e) (string=? e (cadr d))))
+		   (k env) (z env)))
 
       ((and)   (match (cadr d) e env
 		      (lambda (env)
@@ -454,7 +428,7 @@
       ((not)
        (if (isVar? (cadr d))
 	   (let ((s (jim-gensym "VAR-")))
-	      (k (extend env (cadadr d) `(not (quote s)))))
+	      (k (extend env (cadadr d) `(not (quote ,s))))) ;;; johnm
 	   (match (cadr d) e env z k)))
 
 ;;;-- Le cas ou la descr. est un vector et le filtre un vector-begin --*/
@@ -491,7 +465,7 @@
 	       (env (lambda (x) 'unbound))
 	       (k (lambda (f e) f)) )
       (cond
-	 ((or (boolean? f) (symbol? f) (string? f) (integer? f) )
+	 ((or (boolean? f) (symbol? f) (string? f) (fixnum? f) )
 	  (k f env))
 	 ((null? f)
 	  (k f env))
@@ -499,7 +473,7 @@
 	  (k f env))
 	 ((eq? (car f) 'var)
 	  (if (eq? (env (cadr f)) 'unbound)
-	      (let ((s (jim-gensym)))
+	      (let ((s (jim-gensym "ALPHA-")))
 		 (k (list 'var s) (extend env (cadr f) s)))
 	      (k (list 'var (env (cadr f))) env)))
 	 (else (loop (car f) env
@@ -564,7 +538,7 @@
 ;;;   to extend it.                                                    */
 ;;;--------------------------------------------------------------------*/
 
-(define (Vector-Plus v i d)
+(define (vector-plus v i d)
    (if (>=fx i (vector-length (caddr v)))
        (set-car! (cddr v)
 		 (extend-vector (caddr v) i '(any)))
@@ -575,7 +549,7 @@
 		   (pattern-plus (vector-ref (caddr v) i) d))
       res))
 
-(define (Vector-Minus v i d)
+(define (vector-minus v i d)
    (if (>=fx i (vector-length (caddr v)))
        (set-car! (cddr v)
 		 (extend-vector (caddr v) i '(any)))
