@@ -1,62 +1,36 @@
-;* --------------------------------------------------------------------*/
-;*    Copyright (c) 1992-1998 by Manuel Serrano. All rights reserved.  */
-;*                                                                     */
-;*                                     ,--^,                           */
-;*                               _ ___/ /|/                            */
-;*                           ,;'( )__, ) '                             */
-;*                          ;;  //   L__.                              */
-;*                          '   \   /  '                               */
-;*                               ^   ^                                 */
-;*                                                                     */
-;*                                                                     */
-;*    This program is distributed in the hope that it will be useful.  */
-;*    Use and copying of this software and preparation of derivative   */
-;*    works based upon this software are permitted, so long as the     */
-;*    following conditions are met:                                    */
-;*           o credit to the authors is acknowledged following         */
-;*             current academic behaviour                              */
-;*           o no fees or compensation are charged for use, copies,    */
-;*             or access to this software                              */
-;*           o this copyright notice is included intact.               */
-;*      This software is made available AS IS, and no warranty is made */
-;*      about the software or its performance.                         */
-;*                                                                     */
-;*      Bug descriptions, use reports, comments or suggestions are     */
-;*      welcome. Send them to                                          */
-;*        Manuel Serrano -- Manuel.Serrano@unice.fr                    */
-;*-------------------------------------------------------------------- */
 
 (module __match_normalize
 
-   (import  (__error                   "Llib/error.scm")
-	    (__match_s2cfun            "Match/s2cfun.scm"))
+   (import  __error
+	    __match_s2cfun
+	    __param)
    
-   (use     (__type                    "Llib/type.scm")
-	    (__bigloo                  "Llib/bigloo.scm")
-	    (__tvector                 "Llib/tvector.scm")
-	    (__structure               "Llib/struct.scm")
-	    (__tvector                 "Llib/tvector.scm")
-	    (__rgc                     "Rgc/runtime.scm")
-	    (__r4_numbers_6_5          "Ieee/number.scm")
-	    (__r4_numbers_6_5_fixnum   "Ieee/fixnum.scm")
-	    (__r4_numbers_6_5_flonum   "Ieee/flonum.scm")
-	    (__r4_characters_6_6       "Ieee/char.scm")
-	    (__r4_equivalence_6_2      "Ieee/equiv.scm")
-	    (__r4_booleans_6_1         "Ieee/boolean.scm")
-	    (__r4_symbols_6_4          "Ieee/symbol.scm")
-	    (__r4_strings_6_7          "Ieee/string.scm")
-	    (__r4_pairs_and_lists_6_3  "Ieee/pair-list.scm")
-	    (__r4_input_6_10_2         "Ieee/input.scm")
-	    (__r4_control_features_6_9 "Ieee/control.scm")
-	    (__r4_vectors_6_8          "Ieee/vector.scm")
-	    (__r4_ports_6_10_1         "Ieee/port.scm")
-	    (__r4_output_6_10_3        "Ieee/output.scm")
-	    (__evenv                   "Eval/evenv.scm"))
+   (use     __type
+	    __bigloo
+	    __tvector
+	    __structure
+	    __tvector
+
+	    __r4_numbers_6_5
+	    __r4_numbers_6_5_fixnum
+	    __r4_numbers_6_5_flonum
+	    __r4_characters_6_6
+	    __r4_equivalence_6_2
+	    __r4_booleans_6_1
+	    __r4_symbols_6_4
+	    __r4_strings_6_7
+	    __r4_pairs_and_lists_6_3
+	    __r4_input_6_10_2
+	    __r4_control_features_6_9
+	    __r4_vectors_6_8
+	    __r4_ports_6_10_1
+	    __r4_output_6_10_3
+	    __evenv)
    
    (export  (normalize-pattern pat)
 	    (match-define-structure! exp)
 	    (extend.r.macro-env name fun)))
-
+ 
 ;;;===========================================================4
 ;;; The standardizer converts patterns with an 
 ;;; extended syntax into pattern within the reduced pattern set.
@@ -157,6 +131,7 @@
         ((term-variable? e) (standardize-term-variable e))
         ((hole-variable? e) (standardize-hole-variable e))
         ((vector? e) (standardize-vector e))
+        ((struct? e) (standardize-struct e))
         ((atom? e) (standardize-quote e))
         (else (standardize-patterns e)) ) )
 
@@ -374,19 +349,27 @@
   (and (pair? e)
        (lookup r.macro-pattern (car e)) ) )
 
-(defmacro-pattern atom (e . e*)
-   (lambda (r c)
-      (if (pair? e*)
-	  (match-wrong "Too many patterns provided for atom")
-	  ((standardize-pattern e)
-	   r
-	   (lambda (pattern rr)
-	      (c `(and (not (cons (any) (any)))
-		       ,pattern)
-		 rr))))))
+(defmacro-pattern atom l
+   (if (null? l)
+       (match-wrong "Illegal `atom' form" l)
+       (let ((e (car l))
+	     (e* (cdr l)))
+	  (lambda (r c)
+	     (if (pair? e*)
+		 (match-wrong "Too many patterns provided for atom" e*)
+		 ((standardize-pattern e)
+		  r
+		  (lambda (pattern rr)
+		     (c `(and (not (cons (any) (any)))
+			      ,pattern)
+			rr))))))))
 
-(defmacro-pattern or (e . e*)
-   (lambda (r c)
+(defmacro-pattern or l
+   (if (null? l)
+       (match-wrong "Illegal `or' form" l)
+       (let ((e (car l))
+	     (e* (cdr l)))
+	  (lambda (r c)
 	     (if (pair? e*)
 		 ((standardize-pattern e)
 		  r
@@ -397,8 +380,8 @@
 			 (if (and (coherent-environment? rr rrr)
 				  (coherent-environment? rrr rr))
 			     (c `(or ,pattern1 ,pattern2) rrr)
-			     (match-wrong "Incompatible alternative"))))))
-		 ((standardize-pattern e) r c))))
+			     (match-wrong "Incompatible alternative" l))))))
+		 ((standardize-pattern e) r c))))))
 
 ;;; (defmacro-pattern tagged-or (e l . e*)  */
 ;;;    (lambda (r c)  */
@@ -415,8 +398,12 @@
 ;;; 			     (match-wrong "Incompatible alternative"))))))  */
 ;;; 		 ((standardize-pattern e) r c) ) ) )  */
 
-(defmacro-pattern t-or (e . e*)
-   (lambda (r c)
+(defmacro-pattern t-or l
+   (if (null? l)
+       (match-wrong "Illegal `t-or' form" l)
+       (let ((e (car l))
+	     (e* (cdr l)))
+	  (lambda (r c)
 	     (if (pair? e*)
 		 ((standardize-pattern e)
 		  r
@@ -427,35 +414,46 @@
 			 (if (and (coherent-environment? rr rrr)
 				  (coherent-environment? rrr rr))
 			     (c `(t-or ,pattern1 ,pattern2) rrr)
-			     (match-wrong "Incompatible alternative"))))))
-		 ((standardize-pattern e) r c) ) ) )
+			     (match-wrong "Incompatible alternative" l))))))
+		 ((standardize-pattern e) r c) ) ) ) ) )
 
-(defmacro-pattern and (e . e*)
-   (lambda (r c)
-	   (if (pair? e*)
-	       ((standardize-pattern e)
-		r
-		(lambda (pattern1 rr)
-		   ((standardize-pattern `(and . ,e*))
-	 	    rr
-		    (lambda (pattern2 rrr)
-		       (c `(and ,pattern1 ,pattern2) rrr) ) ) ) )
-	       ((standardize-pattern e) r c))))
+(defmacro-pattern and l
+   (if (null? l)
+       (match-wrong "Illegal `and' form" l)
+       (let ((e (car l))
+	     (e* (cdr l)))
+	  (lambda (r c)
+	     (if (pair? e*)
+		 ((standardize-pattern e)
+		  r
+		  (lambda (pattern1 rr)
+		     ((standardize-pattern `(and . ,e*))
+		      rr
+		      (lambda (pattern2 rrr)
+			 (c `(and ,pattern1 ,pattern2) rrr) ) ) ) )
+		 ((standardize-pattern e) r c))))))
 
-(defmacro-pattern not (e)
-   (lambda (r c)
-	  ((standardize-pattern e)  
-	   r
-	   (lambda (pattern rr)
-	      (c `(not ,pattern) r) ) ) ) )
+(defmacro-pattern not l
+   (if (or (null? l) (pair? (cdr l)))
+       (match-wrong "Illegal `not'" l)
+       (let ((e (car l)))
+	  (lambda (r c)
+	     ((standardize-pattern e)  
+	      r
+	      (lambda (pattern rr)
+		 (c `(not ,pattern) r) ) ) ) ) ) )
 
-(defmacro-pattern ? (e)
-   (lambda (r c)
-      (c `(check ,e) r)))
+(defmacro-pattern ? l
+   (if (or (null? l) (pair? (cdr l)))
+       (match-wrong "Illegal `?' form" l)
+       (lambda (r c)
+	  (c `(check ,(car l)) r))))
 
-(defmacro-pattern kwote (e)
-   (lambda (r c)
-      (c `(quote ,e) r)))
+(defmacro-pattern kwote l
+   (if (or (null? l) (pair? (cdr l)))
+       (match-wrong "Illegal `kwote' form" l)
+       (lambda (r c)
+	  (c `(quote ,(car l)) r))))
 
 ;;; check coherency between arms of alternative patterns:
 ;;; For instance (match-lambda (or ?x ?y) t) is not coherent
@@ -475,8 +473,8 @@
 
 (define unbound-pattern '**Bad-Luck096561123523452**)
 
-(define (match-wrong . args)
-  (error 'Pattern-Matching "ERREUR: "args))
+(define (match-wrong msg arg)
+  (error 'Pattern-Matching msg arg))
 
 (define (r9.init n)
   unbound-pattern)
@@ -520,8 +518,7 @@
       ((eq? (car p) 'not) 1)
       ((tree-variable? (car p)) 0)
       ((memq (car p) '(??- ???-)) 0)
-      ((memq (car p) '(or and t-or tagged-or))
-       (pattern-length (cadr p)))
+      ((memq (car p) '(or and t-or tagged-or)) (pattern-length (cadr p)))
       (else (+ 1 (pattern-length (cdr p))))))
       
 ;*---------------------------------------------------------------------*/
@@ -540,11 +537,9 @@
 
 (defmacro-pattern struct-pat (name . e*)
    (lambda (r c)
-      (c `(struct-pat ,name
-                      ,@(map normalize-pattern e*))
-         r)))
+      (c `(struct-pat ,name ,@(map normalize-pattern e*)) r)))
 
-(defmacro-pattern _structure_ f              
+(define (standardize-struct e)
    (lambda (r c)
       (define (look-for-structure provided-fields)
          (let loop1 ((S *Match-Structures*))
@@ -558,13 +553,14 @@
                            (loop1 (cdr S))))))))
       (let* (;; On recupere le nom et la liste des champs
              ;; dans le bon ordre
+	     (f (struct->list e))
              (structure (if (pair? (car f))
-                            (look-for-structure
-                             (map car (cdr f)))
+                            (look-for-structure (map car (cdr f)))
                             (if (assoc (car f) *Match-Structures*)
                                 (assoc (car f) *Match-Structures*)
-                                (error "No such structure "
-                                       f *Match-Structures*))))
+                                (error 'match-case
+				       "No such structure "
+				       (car f)))))
              (name (car structure))
              (fields (cdr structure))
              (provided-fields (if (pair? (car f))
@@ -581,8 +577,7 @@
                 ,@(if (pair? (car f))
                       (map (lambda (field)
                               (if (assoc field provided-fields)
-                                  (cadr (assoc field
-                                               provided-fields))
+                                  (cadr (assoc field provided-fields))
                                   '?-))
                            fields)
                       (cdr f)))))
