@@ -1,36 +1,9 @@
-;* --------------------------------------------------------------------*/
-;*    Copyright (c) 1992-1998 by Manuel Serrano. All rights reserved.  */
-;*                                                                     */
-;*                                     ,--^,                           */
-;*                               _ ___/ /|/                            */
-;*                           ,;'( )__, ) '                             */
-;*                          ;;  //   L__.                              */
-;*                          '   \   /  '                               */
-;*                               ^   ^                                 */
-;*                                                                     */
-;*                                                                     */
-;*    This program is distributed in the hope that it will be useful.  */
-;*    Use and copying of this software and preparation of derivative   */
-;*    works based upon this software are permitted, so long as the     */
-;*    following conditions are met:                                    */
-;*           o credit to the authors is acknowledged following         */
-;*             current academic behaviour                              */
-;*           o no fees or compensation are charged for use, copies,    */
-;*             or access to this software                              */
-;*           o this copyright notice is included intact.               */
-;*      This software is made available AS IS, and no warranty is made */
-;*      about the software or its performance.                         */
-;*                                                                     */
-;*      Bug descriptions, use reports, comments or suggestions are     */
-;*      welcome. Send them to                                          */
-;*        Manuel Serrano -- Manuel.Serrano@unice.fr                    */
-;*-------------------------------------------------------------------- */
 ;*=====================================================================*/
-;*    serrano/prgm/project/bigloo/runtime1.9c/Pp/pp.scm                */
+;*    serrano/prgm/project/bigloo/runtime/Pp/pp.scm                    */
 ;*    -------------------------------------------------------------    */
 ;*    Author      :  Manuel Serrano                                    */
 ;*    Creation    :  Sat Nov 27 14:10:31 1993                          */
-;*    Last change :  Wed Dec 17 17:45:28 1997 (serrano)                */
+;*    Last change :  Fri Nov 26 11:27:29 2004 (serrano)                */
 ;*    -------------------------------------------------------------    */
 ;*    Le pretty-printer de Marc Feeley.                                */
 ;*=====================================================================*/
@@ -40,29 +13,30 @@
 ;*---------------------------------------------------------------------*/
 (module __pp
 
-   (import  (__error                   "Llib/error.scm")
-	    (__r4_output_6_10_3        "Ieee/output.scm"))
+   (import  __error
+	    __r4_output_6_10_3)
    
-   (use     (__type                    "Llib/type.scm")
-	    (__bigloo                  "Llib/bigloo.scm")
-	    (__tvector                 "Llib/tvector.scm")
-	    (__structure               "Llib/struct.scm")
-	    (__tvector                 "Llib/tvector.scm")
-	    (__rgc                     "Rgc/runtime.scm")
-	    (__r4_numbers_6_5          "Ieee/number.scm")
-	    (__r4_numbers_6_5_fixnum   "Ieee/fixnum.scm")
-	    (__r4_numbers_6_5_flonum   "Ieee/flonum.scm")
-	    (__r4_characters_6_6       "Ieee/char.scm")
-	    (__r4_equivalence_6_2      "Ieee/equiv.scm")
-	    (__r4_booleans_6_1         "Ieee/boolean.scm")
-	    (__r4_symbols_6_4          "Ieee/symbol.scm")
-	    (__r4_strings_6_7          "Ieee/string.scm")
-	    (__r4_pairs_and_lists_6_3  "Ieee/pair-list.scm")
-	    (__r4_input_6_10_2         "Ieee/input.scm")
-	    (__r4_control_features_6_9 "Ieee/control.scm")
-	    (__r4_vectors_6_8          "Ieee/vector.scm")
-	    (__r4_ports_6_10_1         "Ieee/port.scm")
-	    (__evenv                   "Eval/evenv.scm"))
+   (use     __type
+	    __bigloo
+	    __param
+	    __tvector
+	    __structure
+	    __tvector
+
+	    __r4_numbers_6_5
+	    __r4_numbers_6_5_fixnum
+	    __r4_numbers_6_5_flonum
+	    __r4_characters_6_6
+	    __r4_equivalence_6_2
+	    __r4_booleans_6_1
+	    __r4_symbols_6_4
+	    __r4_strings_6_7
+	    __r4_pairs_and_lists_6_3
+	    __r4_input_6_10_2
+	    __r4_control_features_6_9
+	    __r4_vectors_6_8
+	    __r4_ports_6_10_1
+	    __evenv)
 
    (export (pp::unspecified exp . port)
 	   *pp-width*
@@ -86,7 +60,7 @@
 		   (current-output-port)
 		   (let ((port (car port)))
 		      (if (not (output-port? port))
-			  (error "pp" "not an input-port" port)
+			  (error "pp" "not an output-port" port)
 			  port)))))
       (generic-write exp #f *pp-width* (lambda (s) (display s port) #t))
       #unspecified))
@@ -134,7 +108,7 @@
     (define (length1? l) (and (pair? l) (null? (cdr l))))
     (let ((head (car l)) (tail (cdr l)))
       (case head
-        ((QUOTE QUASIQUOTE UNQUOTE UNQUOTE-SPLICING) (length1? tail))
+        ((quote quasiquote unquote unquote-splicing) (length1? tail))
         (else                                        #f))))
 
   (define (read-macro-body l)
@@ -143,80 +117,106 @@
   (define (read-macro-prefix l)
     (let ((head (car l)) (tail (cdr l)))
       (case head
-        ((QUOTE)            "'")
-        ((QUASIQUOTE)       "`")
-        ((UNQUOTE)          ",")
-        ((UNQUOTE-SPLICING) ",@"))))
+        ((quote)            "'")
+        ((quasiquote)       "`")
+        ((unquote)          ",")
+        ((unquote-splicing) ",@"))))
 
   (define (out str col)
     (and col (output str) (+fx col (string-length str))))
 
   (define (wr obj col)
-
-    (define (wr-expr expr col)
-      (if (read-macro? expr)
-        (wr (read-macro-body expr) (out (read-macro-prefix expr) col))
-        (wr-lst expr col)))
-
-    (define (wr-lst l col)
-      (if (pair? l)
-        (let loop ((l (cdr l)) (col (wr (car l) (out "(" col))))
-          (and col
-               (cond ((pair? l) (loop (cdr l) (wr (car l) (out " " col))))
-                     ((null? l) (out ")" col))
-                     (else      (out ")" (wr l (out " . " col)))))))
-        (out "()" col)))
-
-    (cond ((match-case obj
-	      ((COMMENT (? integer?) (? string?))
-	       #t)
-	      (else
-	       #f))
-	   (let ((add (- *pp-width* (string-length (caddr obj)) 3)))
-	      (if (<=fx add 0)
-		  (out (caddr obj) col)
-		  (out (string-append (caddr obj) (make-string add #\space))
-		       col))))
-	  ((pair? obj)        (wr-expr obj col))
-          ((null? obj)        (wr-lst obj col))
-          ((vector? obj)      (wr-lst (vector->list obj)
-				      (out (vector-prefix obj) col)))
-          ((boolean? obj)     (out (if obj "#t" "#f") col))
-          ((number? obj)      (out (number->string obj) col))
-          ((symbol? obj)      (case *pp-case*
-				 ((respect)
-				  (out (symbol->string obj) col))
-				 ((upper)
-				  (out (string-upcase (symbol->string obj))
-				       col))
-				 (else
-				  (out (string-downcase (symbol->string obj))
-				       col))))
-          ((procedure? obj)   (out "#<procedure>" col))
-          ((string? obj)      (let ((obj (string-for-read obj)))
-				 (if display?
-				     (out obj col)
-				     (let loop ((i 0)
-						(j 0) 
-						(col (out "#\"" col)))
-					(if (and col (<fx j (string-length obj)))
-					    (let ((c (string-ref obj j)))
-					       (loop i (+fx j 1) col))
-					    (out "\""
-						 (out (substring obj i j)
-						      col)))))))
-          ((char? obj)        (if display?
-                                (out (make-string 1 obj) col)
-				(let ((p (open-output-string)))
-				   (write obj p)
-				   (out (close-output-port p) col))))
-          ((input-port? obj)  (out "#[input-port]" col))
-          ((output-port? obj) (out "#[output-port]" col))
-          ((eof-object? obj)  (out "#[eof-object]" col))
-          (else               (out (let ((p (open-output-string)))
-				      (write obj p)
-				      (close-output-port p))
-				   col))))
+     
+     (define (wr-expr expr col)
+	(if (read-macro? expr)
+	    (wr (read-macro-body expr) (out (read-macro-prefix expr) col))
+	    (wr-lst expr col)))
+     
+     (define (wr-lst l col)
+	(if (pair? l)
+	    (let loop ((l (cdr l)) (col (wr (car l) (out "(" col))))
+	       (and col
+		    (cond ((pair? l) (loop (cdr l) (wr (car l) (out " " col))))
+			  ((null? l) (out ")" col))
+			  (else      (out ")" (wr l (out " . " col)))))))
+	    (out "()" col)))
+     
+     (cond ((match-case obj
+	       ((comment (? fixnum?) (? string?))
+		#t)
+	       (else
+		#f))
+	    (let ((add (- *pp-width* (string-length (caddr obj)) 3)))
+	       (if (<=fx add 0)
+		   (out (caddr obj) col)
+		   (out (string-append (caddr obj) (make-string add #\space))
+			col))))
+	   ((pair? obj)
+	    (wr-expr obj col))
+	   ((null? obj)
+	    (wr-lst obj col))
+	   ((vector? obj)
+	    (wr-lst (vector->list obj)
+		    (out (vector-prefix obj) col)))
+	   ((boolean? obj)
+	    (out (if obj "#t" "#f") col))
+	   ((number? obj)
+	    (cond
+	       ((elong? obj)
+		(out (string-append "#e" (number->string obj)) col))
+	       ((llong? obj)
+		(out (string-append "#l" (number->string obj)) col))
+	       (else
+		(out (number->string obj) col))))
+	   ((symbol? obj)
+	    (let ((p (open-output-string)))
+	       (if display?
+		   (display obj p)
+		   (write obj p))
+	       (case *pp-case*
+		  ((respect)
+		   (out (close-output-port p) col))
+		  ((upper)
+		   (out (string-upcase (close-output-port p))
+			col))
+		  (else
+		   (out (string-downcase (close-output-port p))
+			col)))))
+	   ((procedure? obj)
+	    (out obj col))
+	   ((string? obj)
+	    (let ((obj (string-for-read obj)))
+	       (if display?
+		   (out obj col)
+		   (let loop ((i 0)
+			      (j 0) 
+			      (col (out (if (bigloo-strict-r5rs-strings)
+					    "#\""
+					    "\"")
+					col)))
+		      (if (and col (<fx j (string-length obj)))
+			  (let ((c (string-ref obj j)))
+			     (loop i (+fx j 1) col))
+			  (out "\""
+			       (out (substring obj i j)
+				    col)))))))
+	   ((char? obj)
+	    (if display?
+		(out (make-string 1 obj) col)
+		(let ((p (open-output-string)))
+		   (write obj p)
+		   (out (close-output-port p) col))))
+	   ((input-port? obj)
+	    (out "#[input-port]" col))
+	   ((output-port? obj)
+	    (out "#[output-port]" col))
+	   ((eof-object? obj)
+	    (out "#[eof-object]" col))
+	   (else
+	    (out (let ((p (open-output-string)))
+		    (write obj p)
+		    (close-output-port p))
+		 col))))
 
   (define (pp obj col)
 
@@ -345,20 +345,30 @@
     (define (pp-expr-defun l col extra)
       (pp-defun-form l col extra pp-expr))
     
-    (define (pp-DEFiNE expr col extra)
+    (define (pp-define expr col extra)
       (pp-general expr col extra #f pp-expr-list #f pp-expr)
       (out #"\n" 0))
 
-    (define (pp-DEFUN expr col extra)
+    (define (pp-defun expr col extra)
       (pp-general expr col extra #t pp-expr-defun #f pp-expr)
       (out #"\n" 0))
-      
-    (define (pp-LAMBDA expr col extra)
-      (pp-general expr col extra #f pp-expr-list #f pp-expr))
 
-    (define (pp-COMMENT expr col extra)
+    (define (pp-let expr col extra)
+       (pp-general expr col extra #f pp-expr-list #f pp-expr))
+    
+    ; (item1 item2
+    ;     item3
+    ;     item4)
+    (define (pp-lambda expr col extra)
+       (let* ((col (out "(" col))
+	      (col2 (wr (car expr) col))
+	      (col3 (out " " col2))
+	      (col4 (wr (cadr expr) col3)))
+	 (pp-down (cddr expr) col3 (+fx col 2) extra pp-expr)))
+   
+    (define (pp-comment expr col extra)
        (match-case expr
-	  ((COMMENT (and (? integer?) ?column) (and (? string?) ?string))
+	  ((comment (and (? fixnum?) ?column) (and (? string?) ?string))
 	   (let ((add (- *pp-width* (string-length string) 3)))
 	      (if (=fx column 0)
 		  (if (>fx add 0)
@@ -372,27 +382,27 @@
 	  (else
 	   (pp-general expr col extra #f pp-expr #f pp-expr))))
        
-    (define (pp-IF expr col extra)
+    (define (pp-if expr col extra)
       (pp-general expr col extra #f pp-expr #f pp-expr))
 
-    (define (pp-COND expr col extra)
+    (define (pp-cond expr col extra)
       (pp-call expr col extra pp-expr-list))
 
-    (define (pp-CASE expr col extra)
+    (define (pp-case expr col extra)
       (pp-general expr col extra #f pp-expr #f pp-expr-list))
-
-    (define (pp-AND expr col extra)
+    
+    (define (pp-and expr col extra)
       (pp-call expr col extra pp-expr))
 
-    (define (pp-LET expr col extra)
+    (define (pp-let expr col extra)
       (let* ((rest (cdr expr))
              (named? (and (pair? rest) (symbol? (car rest)))))
         (pp-general expr col extra named? pp-expr-list #f pp-expr)))
 
-    (define (pp-BEGIN expr col extra)
+    (define (pp-begin expr col extra)
       (pp-general expr col extra #f #f #f pp-expr))
 
-    (define (pp-DO expr col extra)
+    (define (pp-do expr col extra)
       (pp-general expr col extra #f pp-expr-list pp-expr-list pp-expr))
 
     ; define formatting style (change these to suit your style)
@@ -405,19 +415,20 @@
 
     (define (style head)
       (case (if (eq? *pp-case* 'respect)
-		(string->symbol (string-upcase (symbol->string head)))
+		(string->symbol (symbol->string head))
 		head)
-        ((LAMBDA LET* LETREC)                                pp-LAMBDA)
-	((DEFINE DEFINE-INLINE DEFINE-METHOD DEFINE-GENERIC) pp-DEFINE)
-	((DEFUN DE)                                          pp-DEFUN)
-        ((IF SET!)                                           pp-IF)
-        ((COND)                                              pp-COND)
-        ((CASE)                                              pp-CASE)
-        ((AND OR)                                            pp-AND)
-        ((LET)                                               pp-LET)
-        ((BEGIN)                                             pp-BEGIN)
-        ((DO)                                                pp-DO)
-	((COMMENT)                                           pp-COMMENT)
+        ((lambda)                                            pp-lambda)
+        ((let* letrec)                                       pp-let)
+	((define define-inline define-method define-generic) pp-define)
+	((defun de)                                          pp-defun)
+        ((if set!)                                           pp-if)
+        ((cond)                                              pp-cond)
+        ((case)                                              pp-case)
+        ((and or)                                            pp-and)
+        ((let)                                               pp-let)
+        ((begin)                                             pp-begin)
+        ((do)                                                pp-do)
+	((comment)                                           pp-comment)
         (else                                                #f)))
 
     (pr obj col 0 pp-expr))
@@ -448,7 +459,7 @@
 ; (object->string obj) returns the textual representation of 'obj' as a
 ; string.
 ;
-; Note: (write obj) = (display (object->string obj))
+; note: (write obj) = (display (object->string obj))
 
 (define (object->string obj)
   (let ((result '()))
@@ -474,7 +485,7 @@
               #t)))))
     (reverse-string-append result)))
 
-; (pretty-print obj port) pretty prints 'obj' on 'port'.  The current
+; (pretty-print obj port) pretty prints 'obj' on 'port'.  the current
 ; output port is used if 'port' is not specified.
 
 (define (pretty-print obj . opt)
@@ -488,7 +499,3 @@
   (let ((result '()))
     (generic-write obj #f 79 (lambda (str) (set! result (cons str result)) #t))
     (reverse-string-append result)))
-
-
-
-
